#include <sys/types.h>
#include <stdio.h>
#include <time.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <errno.h>
#include <pthread.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

/* User's header of BDB */
#include "db.h"
#include "TTDB.h"

char *homedir = NULL;
char *filename = NULL;
char *database = NULL;

DB_ENV *dbenv = NULL;
DB * dbp = NULL;

/*******************************************************************
 * Database operations
 *******************************************************************/
int dbStartup()
{
    int ret;

    printf("Open and join the environment.\n");

    /*Create the environment handle. */
    if ((ret = db_env_create(&dbenv, 0)) != 0) {
	printf("dbStartup: db_env_create: %s\n", 
	       db_strerror(ret));
	return (-1);
    }

    /* Set up error handling. */
    dbenv->set_errpfx(dbenv, "txnapp");

    /* Opening the Environment and setting its options */
    printf("[%s] Open env.\n");
    if ((ret = dbenv->open(dbenv, homedir,
			   DB_JOINENV | DB_THREAD,
			   S_IRUSR | S_IWUSR)) != 0) {
	printf("dbStartup: DB_ENV->open: %s\n", db_strerror(ret));
	dbenv->err(dbenv, ret, "dbenv->open: %s", homedir);
	return(-1);
    }

    printf("Environment joined\n");
    printf("Open the database.\n");

    /* Creating a db handle under the dbenvironment */
    if ((ret = db_create(&dbp, dbenv, 0)) != 0) {
	printf("dbStartup: db_create: %s\n", db_strerror(ret));
	return(-1);
    }

    /* Open the db  */
    printf("[%s] Open DB.\n");
    if ((ret = dbp->open(dbp, filename, database, DB_BTREE, 
			 DB_TEMPORAL, 
			 0660)) != 0) {
	printf("dbStartup: DB->open: %s\n", db_strerror(ret));
	dbp->err(dbp, ret, "db->open: %s", database);
	return(-1);
    }

    printf("Database opened\n");
    return (0);
}

int dbShutdown() 
{
    int ret;

    if ((ret = dbp->close(dbp, 0)) != 0) {
	dbp->err(dbp, ret, "db close");
	return(-1);
    }

    printf("Database closed\n");

    if ((ret = dbenv->close(dbenv,0)) != 0) {
	dbenv->err(dbenv, ret, "db env close");
	return(-1);
    }

    printf("Environment closed\n");
    return (0);
}

/*******************************************
 * Scan the entire database
 *******************************************/
int
scan()
{
    DBC *dbcp;
    DBT key, data;
    DB_TXN *tid;
    int ret, cnt = 0;

    if ((ret = txn_begin(dbenv, NULL, &tid, 0)) != 0) {
        dbenv->err(dbenv, ret, "DB_ENV->txn_begin");
	goto err;
    }

    /* Acquire a cursor for the database. */
    if ((ret = dbp->cursor(dbp, tid, &dbcp, 0)) != 0) {
        dbp->err(dbp, ret, "DB->cursor");
	goto err;
    }

    /* Initialize the key/data return pair. */
    memset(&key, 0, sizeof(key));
    memset(&data, 0, sizeof(data));

    printf("Scanning ...");
    /* Walk through the database and print out the key/data pairs. */
    while ((ret = dbcp->c_get(dbcp, &key, &data, DB_NEXT)) == 0)
        if ((cnt ++) % 100 == 0)
	    printf(".");
    printf(" Finished\n");

    if (ret != DB_NOTFOUND && ret != 0) {
        dbp->err(dbp, ret, "DBcursor->get");
	goto err;
    }
    if ((ret = dbcp->c_close(dbcp)) != 0) {
        dbp->err(dbp, ret, "db cursor close");
	goto err;
    }

    if ((ret = txn_commit(tid, 0)) != 0) {
	dbenv->err(dbenv, ret, "txn_commit");
	goto err;
    }

    return 0;
  err:
    exit(1);
}

void usage(char *cmd)
{
    printf("Usage: %s [-v] -h homedir -f filename -d database\n", cmd);
}

/* Parse the command line options and set the correspondent
 * flags and variables.
 */
int parse_opts(int argc, char *argv[])
{
    char c;

    /* Init getopt() */
    opterr = 0;

    /* Read all the options */
    while ( 1 ) {
        if ((c = getopt(argc, argv, "vh:f:d:")) == -1)
	    break;

        switch (c) {
        case 'v':
	    printf("%s for Temporal BerkeleyDB 3.2.9a\n", argv[0]);
	    exit (0);
            break;
        case 'h':
            homedir = optarg;
            break;
        case 'f':
	    filename = optarg;
            break;
        case 'd':
	    database = optarg;
            break;
	case ':':
	    printf("Missing arguemnt for option -%c\n", (char)optopt);
	    usage(argv[0]);
	    return 1;
	    break;
        case '?':
	    printf("Invalid option\n");
            usage(argv[0]);
            return 1;
        default:
            fprintf(stderr, "Unknown return char from getopt()\n");
            return 1;
        }
    }

    if ((homedir == NULL) || (filename == NULL) || (database == NULL)) {
	usage(argv[0]);
	printf("Please specify all the information for environment\n");
	printf("home directory, filename and database name.\n");
	return 1;
    }
    return 0;
}

int main (int argc, char *argv[]) {

    if (parse_opts(argc, argv) != 0)
	exit (1);

    dbStartup();
    scan();
    dbShutdown();

    return 0;
}
