/* STP data structure manipulation -
 * In STP, data structures are stored in the shared memory region.
 * They are mostly organized as linking lists in arrays in ascending order.
 * These are for their operations.
 * Because the FIND is different for differnet kinds of lists, but
 * INSERT and DEL are all the same, we define STP_XX_FIND for each different
 * list and a common STP_LIST_INSERT and STP_LIST_DEL.
 */

#define STP_TT_HASH(dbstp, id) ( id.txnid % dbstp->tt_htab_size )

/* Find Entry in TT  --
 * head - (STP_XX_ENTRY *) pointer to the head of the array.
 * begin - (int) index to the entry from where we do the search.
 * field - the key field we are doing the comparison.
 * key - the key value we are looking for.
 * result - (int) returns the index of the entry which is either immediately
 * before the entry which contains the key value or contains the
 * largest key value that is smaller than the target key.
 */
#ifdef EXPERIMENT
#define STP_TT_FIND(dbstp, tt_htab, id, tt_result) do { \
    u_int32_t tt_idx; \
    DB_HASHTAB *__bucket; \
    tt_idx = STP_TT_HASH(dbstp, id); \
    __bucket = &(tt_htab[tt_idx]); \
    for (tt_result = SH_TAILQ_FIRST(__bucket, __stp_tt_entry); \
        tt_result != NULL; \
        tt_result = SH_TAILQ_NEXT(tt_result, links, __stp_tt_entry)) { \
        count_ttscan(); \
        if (STPID_EQUAL(tt_result->txnid, id)) \
           break; \
    } \
} while(0)
#else
#define STP_TT_FIND(dbstp, tt_htab, id, tt_result) do { \
    u_int32_t tt_idx; \
    DB_HASHTAB *__bucket; \
    tt_idx = STP_TT_HASH(dbstp, id); \
    __bucket = &(tt_htab[tt_idx]); \
    for (tt_result = SH_TAILQ_FIRST(__bucket, __stp_tt_entry); \
        tt_result != NULL; \
        tt_result = SH_TAILQ_NEXT(tt_result, links, __stp_tt_entry)) { \
        if (STPID_EQUAL(tt_result->txnid, id)) \
           break; \
    } \
} while(0)
#endif

#define BT_COMPARE(b1, b2) ( \
    ((b1).pgno > (b2).pgno)? 1: \
        ( ((b1).pgno < (b2).pgno)? -1: memcmp((b1).fileid, (b2).fileid, DB_FILE_ID_LEN) ) \
)

#define STP_BT_HASH(dbstp, pgno, fileid) ( (pgno)  \
    % ((dbstp)->bt_htab_size) )

#define STP_BT_FIND(dbstp, pgno, fileid1, bt_result) do { \
    u_int32_t bt_idx; \
    DB_HASHTAB *__bucket; \
    STP_BT_ENTRY b; \
    b.pgno = pgno; \
    memcpy(b.fileid, fileid1, DB_FILE_ID_LEN); \
    bt_idx = STP_BT_HASH(dbstp, pgno, fileid1); \
    __bucket = &(dbstp->bt_htab[bt_idx]); \
    for (bt_result = SH_TAILQ_FIRST(__bucket, __stp_bt_entry); \
        bt_result != NULL; \
        bt_result = SH_TAILQ_NEXT(bt_result, links, __stp_bt_entry)) { \
        if (BT_COMPARE(*bt_result, b) == 0) \
           break; \
    } \
} while(0)

#define STP_BT_INSERT(dbstp, pgno, fileid, bt_result) do { \
    STP *stp; \
    u_int32_t bt_idx; \
    DB_HASHTAB *__bucket; \
    bt_idx = STP_BT_HASH(dbstp, pgno, fileid); \
    __bucket = &(dbstp->bt_htab[bt_idx]); \
    stp = dbstp->reginfo.primary; \
    if ((bt_result = SH_TAILQ_FIRST(&(stp->free_bt_ent), __stp_bt_entry)) != NULL) { \
        SH_TAILQ_REMOVE(&(stp->free_bt_ent), bt_result, links, __stp_bt_entry); \
        bt_result->pgno = pgno; \
        SH_TAILQ_INIT(&(bt_result->btn)); \
        memcpy(bt_result->fileid, fileid, DB_FILE_ID_LEN); \
        SH_TAILQ_INSERT_HEAD(__bucket, bt_result, links, __stp_bt_entry); \
    } \
} while(0)

#define STP_BT_DEL(dbstp, bt) do { \
    STP *stp; \
    u_int32_t bt_idx; \
    DB_HASHTAB *__bucket; \
    bt_idx = STP_BT_HASH(dbstp, pgno, fileid); \
    __bucket = &(dbstp->bt_htab[bt_idx]); \
    stp = dbstp->reginfo.primary; \
    SH_TAILQ_REMOVE(&(dbstp->bt_htab[bt_idx]), bt, links, __stp_bt_entry); \
    SH_TAILQ_INSERT_HEAD(&(stp->free_bt_ent), bt, links, __stp_bt_entry); \
} while(0)

/* The same thing for TL
#define STP_TL_FIND(head, start, id, result) do { \
result = start; \
while (head[result].next != IDX_INVALID) { \
if (head[head[result].next].txnid.envid > id.envid) break; \
if (head[head[result].next].txnid.envid == id.envid) \
        if (head[head[result].next].txnid.txnid >= id.txnid) break; \
	result = head[result].next; \
	} \
} while(0)
*/

/*
#define STP_BTN_FIND(head, start, tt_idx, result) do { \
result = start; \
while (head[result].next != IDX_INVALID) { \
if (head[head[result].next].tt >= tt_idx) break; \
result = head[result].next; \
} \
} while(0)
*/

/* Insert a new entry in the list --
 * after - (int) the index of the entry after which we insert the new node.
 * result - (int) returns the index of the newly inserted entry.
 */
#define STP_LIST_INSERT(head, after, result) do { \
result = head[1].next; \
if (result == IDX_INVALID) break; \
head[1].next = head[result].next; \
head[result].next = head[after].next; \
head[after].next = result; \
} while(0)

/* Delete an entry from the list --
 * del - (int) the index of the entry immediately before the entry are want
 * to delete.
 * no returned value.
 */
#define STP_LIST_DEL(head, del) do { \
list_index_tmp = head[1].next; \
head[1].next = head[del].next; \
head[del].next = head[head[del].next].next; \
head[head[1].next].next = list_index_tmp; \
} while(0)

#define STP_LIST_NEW(head, btn_new) do { \
        btn_new = head[1].next; \
        if (btn_new != IDX_INVALID) \
                head[1].next = head[btn_new].next; \
} while (0)

/*********************************************************-
 * STP module CPU time statistics for research experiments.
 *********************************************************/
