char const Heap_rcsid[] =
"$Id: heap.c,v 1.1 2012-02-16 00:42:19 ruizhang Exp $";

/********************************************************************

  General-purpose heap-based implementation of priority queue.

  Author:  Terence Kelly  <tpkelly@eecs.umich.edu>
           http://ai.eecs.umich.edu/~tpkelly/

  Copyright 1999, 2000
  The Regents of the University of Michigan
  All rights reserved.

  Permission is granted to use, copy, create derivative works and
  redistribute this software and such derivative works for
  noncommercial education and research purposes, so long as no fee is
  charged, and so long as the copyright notice above, this grant of
  permission, and the disclaimer below appear in all copies made; and
  so long as the name of the University of Michigan is not used in
  any advertising or publicity pertaining to the use or distribution
  of this software without specific, written prior authorization.

  This software is provided as is, without representation from the
  University of Michigan as to its fitness for any purpose, and
  without warranty by the University of Michigan of any kind, either
  express or implied, including without limitation the implied
  warranties of merchantability and fitness for a particular
  purpose.  The Regents of the University of Michigan shall not be
  liable for any damages, including special, indirect, incidental, or
  consequential damages, with respect to any claim arising out of or
  in connection with the use of the software, even if it has been or
  is hereafter advised of the possibility of such damages.

  ------------------------------------------------------------------

  Important: comparison function "cmp" must define a linear order
  on items!  Borrowed this from the LEDA priority queue docs at

      http://www.mpi-sb.mpg.de/LEDA/MANUAL/node85.html

  LEDA seems fairly complete, but it forces you to use C++ (it's
  a template class) and it offers only a decrease_key operation.
  LEDA has an "old-style" priority queue which offers change_key.

  LEDA uses Fibonacci heaps; might be interesting to compare
  performance.  LEDA PQs also use <key,value> paradigm.  Is there
  some advantage to this?  Can I claim to have a more general
  approach where comparisons are functions?

  Dimensions of comparison:  speed, memory consumption.

********************************************************************/

#include <assert.h>
#include <stdlib.h>
#include "heap.h"

#define T Heap_T

#define PARENT(i)      (i/2)
#define LKID(i)        (2*i)
#define RKID(i)      (2*i+1)

struct HeapItem_T {
  void *it;
};

struct T {
  int maxN, N;
  int (*cmp)(const void *a, const void *b);
  HI_T *(*self)(void *p);
  HI_T items;
};

/* note in C++ we'd use overloading */
#define HCMP(heap, i, op, j) \
((heap)->cmp((heap)->items[i].it, (heap)->items[j].it) op 0)

static int
ref_check(const T heap, void *p) {
  assert(heap && p);
  if (heap->self) {
    assert(heap->self(p));
    if (NULL == *heap->self(p) || (*heap->self(p))->it != p)
      return 0;
  }
  return 1;
}

/* heap->items[0].it must be NULL, and if items[] isn't full then
   so must item one beyond heap->N */
static int
boundary_check(const T heap) {
  assert(heap && 0 <= heap->N && heap->N <= heap->maxN);
  if (NULL != heap->items[0].it)
    return 0;
  if (heap->N < heap->maxN && NULL != heap->items[1 + heap->N].it)
    return 0;
  return 1;
}

static void
swap(T heap, const int a, const int b) {
  void *temp;
  assert(heap);
  assert(0 < a && a <= heap->N && 0 < b && b <= heap->N);
  temp = heap->items[a].it;
  heap->items[a].it = heap->items[b].it;
  heap->items[b].it = temp;
  /* maintain referential integrity */
  if (heap->self) {
    *heap->self(heap->items[a].it) = &heap->items[a];
    *heap->self(heap->items[b].it) = &heap->items[b];
  }
}

static void
sift_up(T heap, int idx) {
  int parent;
  assert(heap && 0 < idx && idx <= heap->N);

  /* given node must be less than children */
  assert(LKID(idx) > heap->N || HCMP(heap, idx, <=, LKID(idx)));
  assert(RKID(idx) > heap->N || HCMP(heap, idx, <=, RKID(idx)));

  while (1 < idx) {
    parent = PARENT(idx);
    if (HCMP(heap, idx, >=, parent))
      break;
    swap(heap, parent, idx);
    idx = parent;
  }
}

static void
sift_down(T heap, int idx) {
  int min;
  assert(heap && 0 < idx && idx <= heap->N);

  /* if given node isn't at top, it must be >= parent */
  assert(1 == idx || HCMP(heap, idx, >=, PARENT(idx)));

  for (;;) {
    if (LKID(idx) > heap->N)
      break;
    min = LKID(idx);
    if (RKID(idx) <= heap->N && HCMP(heap, RKID(idx), <, LKID(idx)))
      min = RKID(idx);
    if (HCMP(heap, idx, <=, min))
      break;
    swap(heap, idx, min);
    idx = min;
  }
}

T
Heap_new(const int max_items, int cmp(const void *a, const void *b),
         HI_T *(*self)(void *p)) {
  T heap;
  int i;
  assert(cmp);
  assert(0 < max_items);
  /* why "1 + init_size"? element zero is unused */
  heap = (T)malloc(sizeof *heap +
                   (1 + max_items) * sizeof(heap->items[0]));
  if (heap) {
    heap->maxN      = max_items;
    heap->N         = 0;
    heap->cmp       = cmp;
    heap->self      = self;
    heap->items     = (HI_T)(heap + 1);
    for (i = 0; i <= heap->maxN; i++)
      heap->items[i].it = NULL;
    assert(boundary_check(heap));
  }
  return heap;
}

/* Must maintain invariant that all slots in heap beyond N are NULL
   pointers; also heap->items[0].it == NULL.  Note also that you
   can't store pointers to things that move, e.g. by realloc(), in a
   heap!  I made that mistake.  Also can't use realloc() to grow the
   heap's capacity if things are pointing into it via self.  Duh.
   (Actually, one could imagine a grow-able heap, but I haven't
   implemented it.)
*/
HeapErr_T
Heap_insert(T heap, void *p) {
  assert(heap && p);
  assert(boundary_check(heap));
  assert(NULL == heap->self || NULL == *heap->self(p));

  if (heap->N == heap->maxN)
    return HEAP_ERR_OVFL;

  heap->N++;
  heap->items[heap->N].it = p;
  if (heap->self)
    *heap->self(p) = &(heap->items[heap->N]);

  sift_up(heap, heap->N);

  assert(ref_check(heap, p));
  assert(boundary_check(heap));
  return HEAP_OKAY;
}

void *
Heap_top(const T heap) {
  assert(heap && boundary_check(heap));
  return heap->items[1].it;
}

void *
Heap_extract_min(T heap) {
  void *p;
  assert(heap && boundary_check(heap));

  if (0 == heap->N)
    return NULL;

  p = heap->items[1].it;
  heap->items[1].it = heap->items[heap->N].it;
  heap->items[heap->N].it = NULL;
  heap->N--;

  if (heap->self) {
    assert(heap->self(p));
    *heap->self(p) = NULL;
    if (0 < heap->N)
      *heap->self(heap->items[1].it) = &(heap->items[1]);
  }

  if (0 < heap->N)
    sift_down(heap, 1);

  return p;
}

void
Heap_check(T heap, void func(HeapErr_T err)) {
  int i;
  assert(heap && func);
  if (!boundary_check(heap))
    func(HEAP_ERR_BOUNDARY);
  for (i = 1; i <= heap->N; i++) {
    if (LKID(i) <= heap->N && HCMP(heap, i, >, LKID(i)))
      func(HEAP_ERR_NOTHEAP);
    if (RKID(i) <= heap->N && HCMP(heap, i, >, RKID(i)))
      func(HEAP_ERR_NOTHEAP);
    if (heap->self && !ref_check(heap, heap->items[i].it))
      func(HEAP_ERR_REF);
  }
  for (i = 1 + heap->N; i <= heap->maxN; i++)
    if (NULL != heap->items[i].it)
      func(HEAP_ERR_BOUNDARY);
  func(HEAP_OKAY);
}

int
Heap_size(T heap) {
  assert(heap && boundary_check(heap));
  return heap->N;
}

void
Heap_free(T *pheap) {
  int i;
  T heap;
  void *p;
  assert(pheap && *pheap && boundary_check(*pheap));
  heap = *pheap;
  for (i = 0; i <= heap->maxN; i++) {
    if (heap->self) {
      p = heap->items[i].it;
      if (p) {
        assert(ref_check(heap, p));
        *heap->self(p) = NULL;
      }
    }
    heap->items[i].it = NULL;
  }
  heap->maxN = heap->N = 0;
  heap->cmp  = NULL;
  heap->self = NULL;
  free(heap);
  *pheap = NULL;
}

/* The following functions can't be implemented efficiently without
   self.  Note that self() function may never return NULL */

#define SELF_FTNS_ASSERTIONS()                        \
  assert(heap && heap->self && p && heap->self(p) &&  \
         boundary_check(heap))

int
Heap_member(const T heap, void *p) {
  SELF_FTNS_ASSERTIONS();
  if (*heap->self(p)) {
    assert(ref_check(heap, p));
    return 1;
  }
  return 0;
}

void *
Heap_remove(T heap, void *p) {   /* generalizes Heap_extract_min() */
  int idx;
  SELF_FTNS_ASSERTIONS();
  if (*heap->self(p)) {
    assert(ref_check(heap, p));
    idx = *heap->self(p) - heap->items;
    assert(0 < idx && idx <= heap->N);
    *heap->self(p) = NULL;
    if (idx < heap->N) {
      heap->items[idx].it = heap->items[heap->N].it;
      *heap->self(heap->items[idx].it) = &heap->items[idx];
    }
    heap->items[heap->N].it = NULL;
    heap->N--;
    if (idx <= heap->N)
      Heap_update(heap, heap->items[idx].it);
    return p;
  }
  return NULL;
}

void
Heap_update(T heap, void *p) {
  int idx;
  SELF_FTNS_ASSERTIONS();
  assert(ref_check(heap, p));
  idx = *heap->self(p) - heap->items;
  assert(0 < idx && idx <= heap->N);
  if (1 < idx && heap->cmp(p, heap->items[PARENT(idx)].it) < 0)
    sift_up(heap, idx);
  else
    sift_down(heap, idx);
}
