/*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*-
 * Transaction Time Support to BDB - Stamper Subsystem
 *
 * By     : Shilong Yao           (yao@CS.Arizona.EDU)
 *          Yuhong Liu            (yuhongl@CS.Ariozna.EDU)
 * Advisor: Dr. Richard Snodgrass (rts@CS.Arizona.EDU)
 * Computer Science Dept., The University of Arizona
 *
 * 04/08/02: Create
 *     Only BTree is implemented. Hooks for other page layout are provided.
 * 06/20/02: Version 2 Begin Coding
 * 12/27/02: Version 2 Finished
 * 04/09/05: version 3 Finished
 */

  /***************************************************************************
   *                                                                         *
   * STP (Transaction Time Stamper) module is a lower level module embedded  *
   * in the BerkeleyDB system to timestamp the data records of the           *
   * transaction-time supported database. This file allocats and initializes *
   * all data structures for STP when BerkeleyDB is opened.                  *
   *                                                                         *
   ***************************************************************************/

#include "db_config.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Original BDB system */
#include "db_int.h"
#include "db_shash.h"
#include "lock.h"
#include "db_page.h"
#include "mp.h"
#include "qam.h"
#include "log.h"

/* TT support */
#include "db.h"
#include "TTDB.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "stp_tt.h"
/* Huilong 7/16/06*/
#ifdef TBDB_AUDIT
#include "aud_int.h"
#endif
/* end */
#include "temporal.h"

extern DB_ENV	*gdbenv;
extern int stp_pause;


/*******************************
 *     External Routines       *
 *******************************/

/*
 * __stp_open --
 *	Open and initialize the STP module.
 *	Called in __dbenv_open (env/env_open.c).
 */
int
__stp_open(dbenv)
    DB_ENV 	*dbenv;
{
    DB_STP		*dbstp;
    STP		*stp;
    int		ret;



    STP_STOP_WORKING();

    STP_DEBUG( printf("\tEnter __stp_open()\n") );

    gdbenv = dbenv;

/* modified by Huilong at 15:12 7/11/2006 */
//    ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid = time(NULL);
{
    struct timeval tv;

    gettimeofday(&tv, NULL);
    ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid.tv_sec  = tv.tv_sec;
    ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid.tv_msec =(u_int16_t)(tv.tv_usec / 1000);
}
/* end of modification */
    /* Create/initialize the DB_STP structure. */
    if ((ret = __os_calloc(dbenv, 1, sizeof(DB_STP), &dbstp)) != 0)
	return (ret);
    dbstp->dbenv = dbenv;

    /* Join/create the log region. */
    dbstp->reginfo.type = REGION_TYPE_STP;
    dbstp->reginfo.id = INVALID_REGION_ID;
    dbstp->reginfo.mode = dbenv->db_mode;
    dbstp->reginfo.flags = REGION_JOIN_OK;
    if (F_ISSET(dbenv, DB_ENV_CREATE))
	F_SET(&dbstp->reginfo, REGION_CREATE_OK);
    if ((ret = __db_r_attach(
	dbenv, &dbstp->reginfo, STP_REG_SIZE)) != 0)
	goto err;

    /* If we created the region, initialize it. */
    if (F_ISSET(&dbstp->reginfo, REGION_CREATE))
	if ((ret = __stp_init(dbenv, dbstp)) != 0)
	    goto err;


    /* Set the local addresses. */
    stp = dbstp->reginfo.primary =
	R_ADDR(&dbstp->reginfo, dbstp->reginfo.rp->primary);

    /* in case we are joining another env, we don't enter __stp_init,
     *  we should hook the tables to the dbstp's pointer here.
     * - 01/30/03
     */
    dbstp->tt_htab = R_ADDR(&dbstp->reginfo, stp->tt_htab_off);
    dbstp->bt_htab = R_ADDR(&dbstp->reginfo, stp->bt_htab_off);
    dbstp->max_tt_ent = stp->max_tt_ent;
    dbstp->max_bt_node = stp->max_bt_node;
    dbstp->max_bt_ent = stp->max_bt_ent;
    dbstp->tt_htab_size = stp->tt_htab_size;
    dbstp->bt_htab_size = stp->bt_htab_size;

/* Huilong 7/27/06 */
/*    dbstp->ov_table = R_ADDR(&dbstp->reginfo, stp->ov_table_off);
    dbstp->ov_count = stp->ov_count;
/* Huilong */

    /*
     * If the region is threaded, then we have to lock both the handles
     * and the region, and we need to allocate a mutex for that purpose.
     */
    if (F_ISSET(dbenv, DB_ENV_THREAD)) {
#ifdef BDB_4
	if ((ret = __db_mutex_setup(dbenv, &dbstp->reginfo, &dbstp->mutexp,
				    MUTEX_ALLOC | MUTEX_NO_RLOCK | MUTEX_THREAD)) != 0)
	    goto err;
#endif
#ifdef BDB_3
	if ((ret = __db_mutex_alloc(dbenv, &dbstp->reginfo,
				    &dbstp->mutexp)) != 0)
	    goto err;
	if ((ret = __db_mutex_init(dbenv, dbstp->mutexp,
				   0, MUTEX_THREAD)) != 0)
	    goto err;
#endif
	/* Shilong, 05/27/03
	if ((ret = __db_shmutex_init(dbenv, dbstp->mutexp, 0,
				     MUTEX_THREAD, &dbstp->reginfo,
				     (REGMAINT *)R_ADDR(&dbstp->reginfo, stp->maint_off))) != 0)
	    goto err;
	*/
    }

    /* Initialize the rest of the structure. */

    /*
     * Set the handle -- we may be about to run recovery, which allocates
     * log cursors.  Log cursors require logging be already configured,
     * and the handle being set is what demonstrates that.
     */
    ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle = dbstp;

    R_UNLOCK(dbenv, &(dbstp->reginfo));
 /*printf("******************exit  stp_open     *******************\n"); */
    return (0);

  err:
    if (dbstp->reginfo.addr != NULL) {
	if (F_ISSET(&dbstp->reginfo, REGION_CREATE))
	    ret = __db_panic(dbenv, ret);
	R_UNLOCK(dbenv, &dbstp->reginfo);
	(void)__db_r_detach(dbenv, &dbstp->reginfo, 0);
    }

    if (dbstp->mutexp != NULL)
	__db_mutex_free(dbenv, &dbstp->reginfo, dbstp->mutexp);

#ifdef BDB_4
	__os_free(dbenv, dbstp);
#endif
#ifdef BDB_3
	__os_free(dbstp, sizeof(*dbstp));
#endif
STP_DEBUG( printf("\tExit __stp_open()\n") );

    return (ret);
}


/*
 * __stp_close --
 *	Close and free the STP module.
 *	Called in __dbenv_refresh (env/env_open.c).
 */
int
__stp_close(dbenv)
    DB_ENV 	*dbenv;
{
    DB_STP		*dbstp;
    STP		*stp;
    int		ret;

	if(dbenv->temporal_handle == NULL) return 0;

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;
    if( dbstp == NULL)
    	return 0;
	/* free stp region in shared memory --> __db_shalloc_free */

    /* free stp --> __os_free(dbstp, sizeof(DB_STP)) */
    if (dbstp->reginfo.addr != NULL) {
	if (F_ISSET(&dbstp->reginfo, REGION_CREATE))
	    ret = __db_panic(dbenv, ret);
	R_UNLOCK(dbenv, &dbstp->reginfo);
	(void)__db_r_detach(dbenv, &dbstp->reginfo, 0);
    }

    if (dbstp->mutexp != NULL)
	__db_mutex_free(dbenv, &dbstp->reginfo, dbstp->mutexp);

#ifdef BDB_4
	__os_free(dbenv, dbstp);
#endif
#ifdef BDB_3
	__os_free(dbstp, sizeof(*dbstp));
#endif
STP_DEBUG( printf("\tExit __stp_close()\n") );

    return ret;
}

/*******************************
 *     Internal Routines       *
 *******************************/


/*
 * __stp_init --
 *	Initialize a stp region in shared memory.
 *	Called by __stp_open.
 */
int
__stp_init(dbenv, dbstp)
    DB_ENV	*dbenv;
    DB_STP 	*dbstp;
{
    DB_HASHTAB *htab;
    STP			*region;
    int 		ret, i;
    STP_TT_ENTRY	*tt;
    STP_BT_ENTRY	*bt;
    STP_BTN             *btnp;
#ifdef BDB_4
#ifdef HAVE_MUTEX_SYSTEM_RESOURCES
    u_int8_t *addr;
#endif
#endif

#ifdef BDB_3
    u_int8_t *addr;
#ifdef MUTEX_SYSTEM_RESOURCES
    size_t maint_size;
#endif
#endif

    /* Initialize the table size */
    dbstp->tt_htab_size = STP_TT_HTAB_SIZE;
    dbstp->bt_htab_size = STP_BT_HTAB_SIZE;
    dbstp->max_tt_ent = STP_MAX_TT_ENT;
    dbstp->max_bt_node = STP_MAX_BT_NODE;
    dbstp->max_bt_ent = STP_MAX_BT_ENT;

    if ((ret = __db_shalloc(dbstp->reginfo.addr,
			    sizeof(*region), 0, &dbstp->reginfo.primary)) != 0)
	goto mem_err;
    dbstp->reginfo.rp->primary =
	R_OFFSET(&dbstp->reginfo, dbstp->reginfo.primary);
    region = dbstp->reginfo.primary;
    memset(region, 0, sizeof(*region));

#ifdef BDB_4
#ifdef HAVE_MUTEX_SYSTEM_RESOURCES
    /* Allocate room for the txn maintenance info and initialize it. */
    if ((ret = __db_shalloc(dbstp->reginfo.addr,
			    sizeof(REGMAINT) + TXN_MAINT_SIZE, 0, &addr)) != 0) {
	__db_err(dbenv,
		 "Unable to allocate memory for mutex maintenance");
	return (ret);
    }
    __db_maintinit(&dbstp->reginfo, addr, TXN_MAINT_SIZE);
    region->maint_off = R_OFFSET(&dbstp->reginfo, addr);
#endif
#endif

#ifdef BDB_3
#ifdef  MUTEX_SYSTEM_RESOURCES
    /* Allocate room for the txn maintenance info and initialize it. */
    if ((ret = __db_shalloc(dbstp->reginfo.addr,
			    sizeof(REGMAINT) + STP_MAINT_SIZE, 0, &addr)) != 0)
	goto mem_err;
    __db_maintinit(&dbstp->reginfo, addr, STP_MAINT_SIZE);
    region->maint_off = R_OFFSET(&dbstp->reginfo, addr);
#endif
    if ((ret = __db_shmutex_init(dbenv, &region->flush, 0,
				 0, &dbstp->reginfo,
				 (REGMAINT *)R_ADDR(&dbstp->reginfo, region->maint_off))) != 0)
	return (ret);
#endif

    region->tt_htab_size = dbstp->tt_htab_size;
    region->bt_htab_size = dbstp->bt_htab_size;
    region->max_tt_ent = dbstp->max_tt_ent;
    region->max_bt_node = dbstp->max_bt_node;
    region->max_bt_ent = dbstp->max_bt_ent;


    /* Allocate room for the TT hash table and initialize it. */
    if ((ret =  __db_shalloc(dbstp->reginfo.addr,
			     (dbstp->tt_htab_size)*sizeof(DB_HASHTAB),
			     0,
			     &htab)) != 0)
	goto mem_err;
    __db_hashinit(htab, dbstp->tt_htab_size);
    region->tt_htab_off = R_OFFSET(&dbstp->reginfo, htab);
    dbstp->tt_htab = htab;

    /* Allocate room for the BT hash table and initialize it. */
    if ((ret =  __db_shalloc(dbstp->reginfo.addr,
			     (dbstp->bt_htab_size)*sizeof(DB_HASHTAB),
			     0,
			     &htab)) != 0)
	goto mem_err;
    __db_hashinit(htab, dbstp->bt_htab_size);
    region->bt_htab_off = R_OFFSET(&dbstp->reginfo, htab);
    dbstp->bt_htab = htab;

    /* Initialize TT entries onto a free list. */
    SH_TAILQ_INIT(&region->free_tt_ent);
    for (i = 0; i < region->max_tt_ent; ++i) {
	if ((ret = __db_shalloc(dbstp->reginfo.addr,
				sizeof(STP_TT_ENTRY),
				0,
				&tt)) != 0)
	    goto mem_err;
	SH_TAILQ_INSERT_HEAD(&(region->free_tt_ent), tt, links, __stp_tt_entry);
    }

    /* Initialize BT entries onto a free list. */
    SH_TAILQ_INIT(&region->free_bt_ent);
    for (i = 0; i < region->max_bt_ent; ++i) {
	if ((ret = __db_shalloc(dbstp->reginfo.addr,
				sizeof(STP_BT_ENTRY),
				0,
				&bt)) != 0)
	    goto mem_err;
	SH_TAILQ_INSERT_HEAD(&(region->free_bt_ent), bt, links, __stp_bt_entry);
    }

    /* Initialize BTN entries onto a free list. */
    SH_TAILQ_INIT(&region->free_bt_node);
    for (i = 0; i < region->max_bt_node; ++i) {
	if ((ret = __db_shalloc(dbstp->reginfo.addr,
				sizeof(STP_BTN),
				0,
				&btnp)) != 0) {
	  mem_err:
	    __db_err(dbenv, "Unable to allocate memory for the lock table");
	    return (ret);
	}
	SH_TAILQ_INSERT_HEAD(&(region->free_bt_node), btnp, hlinks, __stp_bt_node);
    }

/* Huilnog 7/27/06 */
/*    if ((ret = __db_shalloc(dbstp->reginfo.addr,
                            sizeof(STP_OVN)*MAX_NUM_OVN,
                            0,
                            &(dbstp->ov_table))) != 0) {
            __db_err(dbenv, "Unable to allocate memory for the lock table");
            return (ret);
   }
   region->ov_table_off = R_OFFSET(&dbstp->reginfo, dbstp->ov_table);    
/* Huilong */

    return (0);
}
