/*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*-
 * Transaction Time Support to BDB - Stamper Subsystem
 *
 * By     : Shilong Yao           (yao@CS.Arizona.EDU)
 *          Yuhong Liu            (yuhongl@CS.Ariozna.EDU)
 * Advisor: Dr. Richard Snodgrass (rts@CS.Arizona.EDU)
 * Computer Science Dept., The University of Arizona
 *
 * 04/08/02: Create
 *     Only BTree is implemented. Hooks for other page layout are provided.
 * 06/20/02: Version 2 Begin Coding
 * 12/27/02: Version 2 Finished
 * 04/09/05: version 3 Finished
 *     Rewrite recover functions for each type of stp logs
 */



 /***************************************************************************
  *                                                                         *
  * STP (Transaction Time Stamper) module is a lower level module embedded  *
  * in the BerkeleyDB system to timestamp the data records of the           *
  * transaction-time supported database. Three types of log entry have been *
  * introduced along with associated recovery functions. This file          *
  * implements all following functions needed in STP logging and recover    *
  * process:                                                                *
  *                                                                         *
  * Log Entry Operations:                                                   *
  *    int __stp_ckp_log(DB_ENV *dbenv, DB_TXN *txnid, DB_LSN *ret_lsnp,    *
  *             u_int32_t flags)                                            *
  *    int _stp_txn_commit_log(DB_ENV *dbenv, STP_ID txnid, TV *timep,      *
  *             u_int32_t *od_pgcnt, DB_LSN *ret_lsnp, u_int32_t flags)     *
  *    int _stp_decr_od_pgcnt_log(DB_ENV *dbenv, STP_ID txnid, TV *timep,      *
  *             u_int32_t *od_pgcnt, DB_LSN *ret_lsnp, u_int32_t flags)     *
  *                                                                         *
  * Recovery Actions:                                                       *
  *    int __stp_recover_ckp(DB_ENV *dbenv, DBT *dbtp, DB_LSN *lsnp,        *
  *             db_recops op, void *info)                                   *
  *    int _stp_recover_txn_commit(DB_ENV *dbenv, DBT *dbtp, DB_LSN *lsnp,  *
  *             db_recops op, void *info)                                   *
  *    int _stp_recover_decr_od_pgcnt(DB_ENV *dbenv, DBT *dbtp, DB_LSN *lsnp,  *
  *             db_recops op, void *info)                                   *
  *                                                                         *
  * Register Recovery operation:                                            *
  *    int __stp_init_recover(DB_ENV *dbenv)                                *
  *                                                                         *
  ***************************************************************************/
#include "db_config.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Original BDB system */
#include "db_int.h"
#include "db_shash.h"
#include "lock.h"
#include "db_page.h"
#include "mp.h"
#include "qam.h"
#include "log.h"

/* TT support */
#include "db.h"
#include "TTDB.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "stp_tt.h"

/* Modified by Huilong at 7/14/2006 */
#ifdef TBDB_AUDIT
#include "aud_int.h"
#endif
/* end of modification */

#include "temporal.h"
#include "stp_auto.h"

extern DB_ENV *gdbenv;
extern int stp_pause;
static STP_INCR_CHUNK *stp_incr_set_head;

#ifdef TBDB_AUDIT
int __aud_tol_recover_log();
#endif

/* Huilong 10/6/2006 */
void __stp_insert_incr_set(STP_ID txnid, int32_t fileid, db_pgno_t pgno)
{
  STP_INCR_CHUNK *tmp_p;

  if(stp_incr_set_head==NULL
     || stp_incr_set_head->count==INCR_CHUNK_SIZE){
     // allocate a new chunk
     tmp_p = (STP_INCR_CHUNK *)malloc(sizeof(STP_INCR_CHUNK));
     if(tmp_p == NULL){
        printf("error in allocate memory!\n");
        exit(1);
     }
     // initialize the chunk
     tmp_p->count = 0;
     // insert the chunk into list
     tmp_p->next  = stp_incr_set_head;
     stp_incr_set_head = tmp_p;
  }
  // put the record into the set
  stp_incr_set_head->incrs[stp_incr_set_head->count].txnid = txnid;
  stp_incr_set_head->incrs[stp_incr_set_head->count].fileid = fileid;
  stp_incr_set_head->incrs[stp_incr_set_head->count].pgno = pgno;
  stp_incr_set_head->count++;
}

int __stp_lookup_incr_set(STP_ID txnid, int32_t fileid, db_pgno_t pgno)
{
  STP_INCR_CHUNK *tmp_p;
  int i;

  tmp_p = stp_incr_set_head;
  while(tmp_p!=NULL){
    for(i=0;i<tmp_p->count;i++){
      if(txnid.envid.tv_sec==tmp_p->incrs[i].txnid.envid.tv_sec
         && txnid.envid.tv_msec==tmp_p->incrs[i].txnid.envid.tv_msec
         && txnid.txnid==tmp_p->incrs[i].txnid.txnid
         && fileid==tmp_p->incrs[i].fileid
         && pgno==tmp_p->incrs[i].pgno)
       return 1;
    }
    tmp_p = tmp_p->next;
  }
  return 0;
}

int __stp_clean_incr_set()
{
  STP_INCR_CHUNK *tmp_p1, *tmp_p2;

  tmp_p1 = stp_incr_set_head;
  while(tmp_p1!=NULL){
    tmp_p2 = tmp_p1->next;
    free(tmp_p1);
    tmp_p1 = tmp_p2;
  }
  return 1;
}

int __stp_recovery_finanlize(DB_ENV *dbenv)
{
  // clean up the INCR set
   __stp_clean_incr_set();

  // no need to Write all pages in memory to disk,
  // because at the end of recovery, these will be flushed
  return 1;
}

int __stp_recovery_stamp_page(DB_ENV *dbenv, STP_ID txnid,
                               int32_t fileid, db_pgno_t pgno)
{
  db_pgno_t my_pgno;
  PAGE *pagep;
  DB *file_dbp;
  DB_MPOOLFILE *mpf;
  int ret;

  if ((ret = __db_fileid_to_db(dbenv,&file_dbp,
                 fileid, 1)) != 0) {
      printf("error in find the fileid!\n");
      exit(1);
  }
  // if the page is not in memory, read it in
  my_pgno = pgno;
  mpf = file_dbp->mpf;
  if ((ret = memp_fget(mpf, &pgno, 0, &pagep)) != 0) {
       printf("error in get the page %d\n", pgno);
       exit(1);
  }

  // stamp the page
  __stp_page_recover(dbenv, mpf, pagep, PAGE_READ, txnid);
  if ((ret = memp_fput(mpf, pagep, DB_MPOOL_DIRTY)) != 0){
       printf("error in put the page %d\n", pgno);
       exit(1);
  }
  return 1;
}

/* return  0-unchanged; 1-changed.
 */
int __stp_recovery_stamp_id(DB_STP *dbstp, IdOrTime *idtimep)
{
  STP *stp;
  STP_TT_ENTRY	*tt;
  STP_ID txnid;
  int ret;

  stp = dbstp->reginfo.primary;
  txnid = idtimep->stamperId;
  ret = 0;
  stp = dbstp->reginfo.primary;

  STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);
  if(tt==NULL) return 0;
  if(IS_INVALID_TIME(tt->time))  return 0;
  memcpy(&idtimep->curr, &(tt->time), sizeof(TV));
  return 1;

}

int __stp_recovery_stamp_tuple(DB_ENV *dbenv, DBT *tuple)
{
  DB_STP       *dbstp;
  STP          *stp;
  DB_HASHTAB   *tt_htab;
  TTDATA       ttdata;
  int          changed;

  dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;
  stp = (dbstp->reginfo).primary;

  // get tt info
  memcpy (&ttdata, (char *)tuple->data, sizeof(TTDATA));

  /* Stamp the "transaction start time" of this data record */
  switch (ttdata.state) {
    case ID_ID:
               if(__stp_recovery_stamp_id(dbstp, &(ttdata.start)) == 1) {
                  ttdata.state = TIME_ID;
               }
               break;
    case ID_UC:
               if (__stp_recovery_stamp_id(dbstp, &(ttdata.start)) == 1) {
                  ttdata.state = TIME_UC;
               }
               break;
    case ID_TIME:
               if (__stp_recovery_stamp_id(dbstp, &(ttdata.start)) == 1) {
                  ttdata.state = TIME_TIME;
               }
               break;
  }

  /* Stamp the "transaction stop time" of this data record */
  switch (ttdata.state) {
     case ID_ID:
               if (__stp_recovery_stamp_id(dbstp, &(ttdata.stop)) == 1) {
                  ttdata.state = ID_TIME;
               }
               break;
     case TIME_ID:
               if (__stp_recovery_stamp_id(dbstp, &(ttdata.stop)) == 1) {
                  ttdata.state = TIME_TIME;
               }
               break;
  }

  return 1;
}
/* Huilong */

/*************************************
 *      Log Entry Operations         *
 *************************************/

/*
 * __stp_ckp_log --
 *	Log the TT table, set of (txnid, time, od_pgcnt), before the the Berkeley DB ckp log.
 *  It is called by txn_checkpoint(), which is called by __db_apprec().
 *
 */
int
__stp_ckp_log(dbenv, txnid, ret_lsnp, flags)
    DB_ENV *dbenv;
    DB_TXN *txnid;
    DB_LSN *ret_lsnp;
    u_int32_t flags;
{
    DBT logrec;
    DB_LSN *lsnp, null_lsn;
    u_int32_t uinttmp, i, tt_index;
    u_int32_t rectype, txn_num;
    int ret;
    u_int8_t *bp;
    STP_TT_ENTRY	*tt, *tmp;
    DB_STP		*dbstp;
	STP_BTN	        *btnp;

/* Modified by Huilong on 7/14/2006 */
#ifdef TBDB_AUDIT
    u_int32_t j;
    DB_AUD *dbaud;
    AUD *aud;
    AUD_TOL_ENTRY *tolep, *new_tolep, *prev_tolep;
#endif
/* End of modification */

	/*
	if( F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
		return 0;
	}
	*/
    STP_STOP_WORKING();

    STP_DEBUG( printf("\tenter __stp_ckp_log\n") );

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    rectype = DB_stp_ckp;
    txn_num = 0; /*txnid == NULL ? 0 : txnid->txnid;*/

/* Added by Huilong at 7/14/2006 from Stanley's code */
#ifdef TBDB_AUDIT
   dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
    if (dbaud == NULL)
        return 0;
    aud = dbaud->reginfo.primary;
#endif
/* End of modification */

    /* Count how many entries are there in TT */
    i = 0; tmp = NULL;
    for (tt_index=0; tt_index < dbstp->tt_htab_size; tt_index++) {
	for (tt= SH_TAILQ_FIRST(&(dbstp->tt_htab[tt_index]), __stp_tt_entry);
	     tt != NULL;
	     tt = SH_TAILQ_NEXT(tt, links, __stp_tt_entry)){
			if(tmp != NULL){
				__tt_reclaim(dbstp, tmp);
				tmp = NULL;
			}
		    if (tt->od_pgcnt > 0)
				i ++;
			else
				tmp = tt;
		}
    }

	/* check if the last tt entry needs to be reclaimed */
	if(tmp != NULL){
		__tt_reclaim(dbstp, tmp);
	}

    STP_DEBUG( printf("\t  nr of entries in TT is i=%ld\n", i) );

/* Added by Huilong, from Stanley's code on 7/14/2006 */
#ifdef TBDB_AUDIT
    /* Count how many entries are there in TOL */
    j = 0;
/*    for (tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
         tolep != NULL;
         tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry)) {
        j ++;
    }*/
#endif
/* End of modification */

/* Huilong 10/11/2006 */
   /* Initialized the INCR set */
   stp_incr_set_head = NULL;
/* Huilong */

    /* Allocate memory for storing TT */
/* Modified by Huilong at 3:32 7/12/2006 */
//    logrec.size = sizeof(rectype) + sizeof(txn_num) + sizeof(i)
//	+ i * (sizeof(time_t) + sizeof(u_int32_t) + sizeof(TV) + sizeof(u_int32_t));
      logrec.size = sizeof(rectype) + sizeof(txn_num) + sizeof(i)
        + i * (sizeof(ENV_ID_TYPE) + sizeof(u_int32_t) + sizeof(TV) + sizeof(u_int32_t))
#ifdef TBDB_AUDIT
        /* Number of TOL entries */
        + sizeof(j) + AUD_HASH_LEN
        + j * (sizeof(TV) + AUD_HASH_LEN)
#endif
;
/* end of modification */
    if ((ret = __os_malloc(dbenv, logrec.size,
#ifdef BDB_3
			   NULL,
#endif
			   &logrec.data)) != 0)
	return (ret);

    /* Fill in the log record */
    bp = logrec.data;

    memcpy(bp, &rectype, sizeof(rectype));
    bp += sizeof(rectype);

    memcpy(bp, &txn_num, sizeof(txn_num));
    bp += sizeof(txn_num);

    memcpy(bp, &i, sizeof(i));
    bp += sizeof(i);

    for (tt_index=0; tt_index < dbstp->tt_htab_size; tt_index++) {
	for (tt = SH_TAILQ_FIRST(&(dbstp->tt_htab[tt_index]), __stp_tt_entry);
	     tt != NULL;
	     tt = SH_TAILQ_NEXT(tt, links, __stp_tt_entry))
	    if (tt->od_pgcnt > 0) {
		STP_DEBUG( printf("\t  write log: txnid=%ld\n", tt[tt_index].txnid) );

/*Modified by Huilong at 2:40 7/12/2006 */
                memcpy(bp, &(tt->txnid.envid), sizeof(ENV_ID_TYPE));
                bp += sizeof(ENV_ID_TYPE);
		//memcpy(bp, &(tt->txnid.envid), sizeof(time_t));
		//bp += sizeof(time_t);
/* end of modification */
		memcpy(bp, &(tt->txnid.txnid), sizeof(u_int32_t));
		bp += sizeof(u_int32_t);
		memcpy(bp, &(tt->time), sizeof(TV));
		bp += sizeof(TV);
		memcpy(bp, &(tt->od_pgcnt), sizeof(u_int32_t));
		bp += sizeof(u_int32_t);
	    }
    }

/* Added by Huilong, from Stanley's code, on 7/14/2006 */
#ifdef TBDB_AUDIT
    memcpy(bp, &j, sizeof(j));
    bp += sizeof(j);

    memcpy(bp, aud->hash_value, AUD_HASH_LEN);
    bp += AUD_HASH_LEN;

    /* Store TOL entries */
/*    for (tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
         tolep != NULL;
         tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry)) {
        memcpy(bp, &(tolep->time), sizeof(TV));
        bp += sizeof(TV);
        memcpy(bp, tolep->hash_value, AUD_HASH_LEN);
        bp += AUD_HASH_LEN;
    }*/
#endif
/* End of modification */

    DB_ASSERT((u_int32_t)(bp - (u_int8_t *)logrec.data) == logrec.size);
#ifdef BDB_4
    ret = dbenv->log_put(dbenv, ret_lsnp, (DBT *)&logrec, flags|DB_FLUSH);
#endif
#ifdef BDB_3
    ret = log_put(dbenv, ret_lsnp, (DBT *)&logrec, DB_FLUSH);
#endif

#ifdef BDB_4
    __os_free(dbenv, logrec.data);
#endif
#ifdef BDB_3
    __os_free(logrec.data, logrec.size);
#endif
STP_DEBUG( printf("\texit __stp_ckp_log\n") );
    return (ret);
}

/* Added by Huilong (from Stanley's code) on 7/14/2006 */
#ifdef TBDB_AUDIT
int
__aud_tol_insert_log(dbenv, timep, hash_value, ret_lsnp, flags)
    DB_ENV *dbenv;
    TV *timep;
    char *hash_value;
    DB_LSN *ret_lsnp;
    u_int32_t flags;
{
    DBT         logrec; 
    DB_LSN      *lsnp, null_lsn;
    u_int32_t   uinttmp, i, tt_index;
    u_int32_t   rectype;
    int         ret;
    u_int8_t    *bp;
    STP_TT_ENTRY        *tt;
    DB_STP              *dbstp;

#ifdef STP_STOP_WORKING
    return 0;
#endif

#ifdef STAMPER_DEBUG
    printf("\tenter __stp_tt_insert_log\n");
#endif

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL) return 0;

    rectype = DB_aud_tol_insert;

    /* Allocate memory for storing TT */
    logrec.size = sizeof(rectype) + sizeof(TV) + AUD_HASH_LEN;
    if ((ret = __os_malloc(dbenv, logrec.size,
#ifdef BDB_3
                           NULL,
#endif
                           &logrec.data)) != 0)
        return (ret);

    /* Fill in the log record */
    bp = logrec.data;

    memcpy(bp, &rectype, sizeof(rectype));
    bp += sizeof(rectype);

    memcpy(bp, timep, sizeof(TV));
    bp += sizeof(TV);

    memcpy(bp, hash_value, AUD_HASH_LEN);
    bp += AUD_HASH_LEN;

    DB_ASSERT((u_int32_t)(bp - (u_int8_t *)logrec.data) == logrec.size);
#ifdef BDB_4
    ret = dbenv->log_put(dbenv, ret_lsnp, (DBT *)&logrec, flags /*| DB_FLUSH*/);
#endif
#ifdef BDB_3
    ret = log_put(dbenv, ret_lsnp, (DBT *)&logrec, flags /*| DB_FLUSH*/);
#endif

#ifdef BDB_4
    __os_free(dbenv, logrec.data);
#endif
#ifdef BDB_3
    __os_free(logrec.data, logrec.size);
#endif
    return (ret);
}
#endif
/* end of modification */

/*
 * __stp_insert_log --
 *	At commit, log the committed txn and its TT info, (txnid, txn_time, od_pgcnt).
 *      But we don't have to flush here, because at the end of the committing,
 *      there is originally a log flush in the original BDB system :)
 *  It is called by __stp_txn_commit()
 */
int
_stp_txn_commit_log(dbenv, txnid, timep, od_pgcnt, ret_lsnp, flags)
    DB_ENV *dbenv;
    STP_ID txnid;
    TV *timep;
    u_int32_t *od_pgcnt;
    DB_LSN *ret_lsnp;
    u_int32_t flags;
{
    DBT		logrec;
    DB_LSN		*lsnp, null_lsn;
    u_int32_t	uinttmp, i, tt_index;
    u_int32_t	rectype;
    int		ret;
    u_int8_t	*bp;
    STP_TT_ENTRY	*tt;
    DB_STP		*dbstp;

	/* Don't insert the log entry if it is during recovery */
	if( F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
		return 0;
	}

    STP_STOP_WORKING();
    STP_DEBUG( printf("\tenter _stp_txn_commit_log\n") );

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    rectype = DB_stp_txn_commit;

    /* Allocate memory for storing TT */
/* Modified by Huilong at 3:35pm 7/12/2006 */
//    logrec.size = sizeof(rectype) + sizeof(time_t) + sizeof(u_int32_t)
//	+ sizeof(TV) + sizeof(u_int32_t);
      logrec.size = sizeof(rectype) + sizeof(ENV_ID_TYPE) + sizeof(u_int32_t)
        + sizeof(TV) + sizeof(u_int32_t);
/* end of modification */
    if ((ret = __os_malloc(dbenv, logrec.size,
#ifdef BDB_3
			   NULL,
#endif
			   &logrec.data)) != 0)
	return (ret);

    /* Fill in the log record */
    bp = logrec.data;

    memcpy(bp, &rectype, sizeof(rectype));
    bp += sizeof(rectype);

/* Modified by Huilong at 7/12/2006 */
    memcpy(bp, &(txnid.envid), sizeof(ENV_ID_TYPE));
    bp += sizeof(ENV_ID_TYPE);
//    memcpy(bp, &(txnid.envid), sizeof(time_t));
//    bp += sizeof(time_t);
/* end of modification */

    memcpy(bp, &(txnid.txnid), sizeof(u_int32_t));
    bp += sizeof(u_int32_t);

    memcpy(bp, timep, sizeof(TV));
    bp += sizeof(TV);

    memcpy(bp, &od_pgcnt, sizeof(od_pgcnt));
    bp += sizeof(od_pgcnt);

    DB_ASSERT((u_int32_t)(bp - (u_int8_t *)logrec.data) == logrec.size);
#ifdef BDB_4
    ret = dbenv->log_put(dbenv, ret_lsnp, (DBT *)&logrec, flags);
#endif
#ifdef BDB_3
    ret = log_put(dbenv, ret_lsnp, (DBT *)&logrec, flags);
#endif

#ifdef BDB_4
    __os_free(dbenv, logrec.data);
#endif
#ifdef BDB_3
    __os_free(logrec.data, logrec.size);
#endif
STP_DEBUG( printf("\texit _stp_txn_commit_log\n") );
    return (ret);
}


/* Huilong 10/9/2006 */
/*
 * _stp_incr_od_pgcnt_log --
 *	Log the TT entry for the txn, (fileid, pgno, txnid), whenever the od_pgcnt is incremented.
 *  It is called by __stp_page() when a page-write increments the od_pgcnt.
 */
int
_stp_incr_od_pgcnt_log(dbenv, txnid, fileid, pgno, ret_lsnp, flags)
    DB_ENV *dbenv;
    STP_ID txnid;
    int32_t fileid;
    u_int32_t pgno;
    DB_LSN *ret_lsnp;
    u_int32_t flags;
{
    DBT logrec;
    DB_LSN *lsnp, null_lsn;
    u_int32_t uinttmp, i, tt_index;
    u_int32_t rectype, txn_num;
    int ret;
    u_int8_t *bp;
    STP_TT_ENTRY	*tt;
    DB_STP			*dbstp;

	/* Don't insert the log entry if it is during recovery */
	if( F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
		return 0;
	}

    STP_STOP_WORKING();

    STP_DEBUG( printf("\tenter _stp_incr_od_pgcnt_log\n") );

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    rectype = DB_stp_incr_od_pgcnt;

    /* Allocate memory for storing TT */
/* Modified by Huilong at 7/12/2006 */
//    logrec.size = sizeof(rectype) + sizeof(time_t) + sizeof(u_int32_t)
//	+ sizeof(fileid) + sizeof(pgno);
      logrec.size = sizeof(rectype) + sizeof(ENV_ID_TYPE) + sizeof(u_int32_t)
        + sizeof(fileid) + sizeof(pgno);
/* End of modification */
    if ((ret = __os_malloc(dbenv, logrec.size,
#ifdef BDB_3
			   NULL,
#endif
			   &logrec.data)) != 0)
	return (ret);

    /* Fill in the log record */
    bp = logrec.data;

    memcpy(bp, &rectype, sizeof(rectype));
    bp += sizeof(rectype);

/* Modified by Huilong at 2:43 7/12/2006 */
    memcpy(bp, &(txnid.envid), sizeof(ENV_ID_TYPE));
    bp += sizeof(ENV_ID_TYPE);
//    memcpy(bp, &(txnid.envid), sizeof(time_t));
//    bp += sizeof(time_t);
/* end of modification */
    memcpy(bp, &(txnid.txnid), sizeof(u_int32_t));
    bp += sizeof(u_int32_t);

    memcpy(bp, &fileid, sizeof(fileid));
    bp += sizeof(fileid);

    memcpy(bp, &pgno, sizeof(pgno));
    bp += sizeof(pgno);

    DB_ASSERT((u_int32_t)(bp - (u_int8_t *)logrec.data) == logrec.size);
#ifdef BDB_4
    ret = dbenv->log_put(dbenv, ret_lsnp, (DBT *)&logrec, flags | DB_FLUSH);
#endif
#ifdef BDB_3
    ret = log_put(dbenv, ret_lsnp, (DBT *)&logrec, flags | DB_FLUSH);
#endif

#ifdef BDB_4
	__os_free(dbenv, logrec.data);
#endif
#ifdef BDB_3
	__os_free(logrec.data, logrec.size);
#endif

STP_DEBUG( printf("\texit _stp_decr_od_pgcnt_log\n") );

    return (ret);
}
/* Huilong */

/*
 * _stp_decr_od_pgcnt_log --
 *	Log the TT entry for the txn, (fileid, pgno, txnid), whenever the od_pgcnt is decremented.
 *  It is called by __stp_page() when a page-write decrements the od_pgcnt.
 */
int
_stp_decr_od_pgcnt_log(dbenv, txnid, fileid, pgno, ret_lsnp, flags)
    DB_ENV *dbenv;
    STP_ID txnid;
    int32_t fileid;
    u_int32_t pgno;
    DB_LSN *ret_lsnp;
    u_int32_t flags;
{
    DBT logrec;
    DB_LSN *lsnp, null_lsn;
    u_int32_t uinttmp, i, tt_index;
    u_int32_t rectype, txn_num;
    int ret;
    u_int8_t *bp;
    STP_TT_ENTRY	*tt;
    DB_STP			*dbstp;

	/* Don't insert the log entry if it is during recovery */
	if( F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
		return 0;
	}

    STP_STOP_WORKING();

    STP_DEBUG( printf("\tenter _stp_decr_od_pgcnt_log\n") );

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    rectype = DB_stp_decr_od_pgcnt;

    /* Allocate memory for storing TT */
/* Modified by Huilong at 7/12/2006 */
//    logrec.size = sizeof(rectype) + sizeof(time_t) + sizeof(u_int32_t)
//	+ sizeof(fileid) + sizeof(pgno);
      logrec.size = sizeof(rectype) + sizeof(ENV_ID_TYPE) + sizeof(u_int32_t)
        + sizeof(fileid) + sizeof(pgno);
/* End of modification */
    if ((ret = __os_malloc(dbenv, logrec.size,
#ifdef BDB_3
			   NULL,
#endif
			   &logrec.data)) != 0)
	return (ret);

    /* Fill in the log record */
    bp = logrec.data;

    memcpy(bp, &rectype, sizeof(rectype));
    bp += sizeof(rectype);

/* Modified by Huilong at 2:43 7/12/2006 */
    memcpy(bp, &(txnid.envid), sizeof(ENV_ID_TYPE));
    bp += sizeof(ENV_ID_TYPE);
//    memcpy(bp, &(txnid.envid), sizeof(time_t));
//    bp += sizeof(time_t);
/* end of modification */
    memcpy(bp, &(txnid.txnid), sizeof(u_int32_t));
    bp += sizeof(u_int32_t);

    memcpy(bp, &fileid, sizeof(fileid));
    bp += sizeof(fileid);

    memcpy(bp, &pgno, sizeof(pgno));
    bp += sizeof(pgno);

    DB_ASSERT((u_int32_t)(bp - (u_int8_t *)logrec.data) == logrec.size);
#ifdef BDB_4
    ret = dbenv->log_put(dbenv, ret_lsnp, (DBT *)&logrec, flags | DB_FLUSH);
#endif
#ifdef BDB_3
    ret = log_put(dbenv, ret_lsnp, (DBT *)&logrec, flags | DB_FLUSH);
#endif

#ifdef BDB_4
	__os_free(dbenv, logrec.data);
#endif
#ifdef BDB_3
	__os_free(logrec.data, logrec.size);
#endif

STP_DEBUG( printf("\texit _stp_decr_od_pgcnt_log\n") );

    return (ret);
}




/*********************************
 *      Recovery actions         *
 *********************************/

 /*
  * __stp_recover_ckp(dbenv, dbtp, lsnp, op, info) --
  *   function to recover the __stp_ckp_log. It builds up the
  *   TT table for stamper.
  *   It is called by __db_apprec (env_recover.c) during REDO Pass
  */
int
__stp_recover_ckp_help(dbenv, dbtp, lsnp, op, info)
    DB_ENV *dbenv;
    DBT *dbtp;
    DB_LSN *lsnp;
    db_recops op;
    void *info;
{
	STP *stp;
    u_int32_t	rectype, nr_entries;
    u_int8_t	*bp;
    STP_TT_ENTRY	*tt;
    STP_BTN 	*btn;
    DB_STP		*dbstp;
#ifdef BDB_4
    DB_LOGC		*logc;
#endif
    STP_ID		txnid;
    TV		time;
    u_int32_t	od_pgcnt, tt_index;
    int		ret;

	STP_STOP_WORKING();

	STP_DEBUG( printf("\tEnter __stp_recover_ckp ...\n") );

	dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

	/* If the environment is NON-TEMPORAL, the stamper do nothing. */
	if (dbstp == NULL)
	return 0;
	stp = dbstp->reginfo.primary;

	/* During undo pass, mark the ckp lsn first seen the last_ckp_lsn.
	 * The stamper recovery is done from this point to end of the log.
	 */
 	memcpy(&(rectype), (u_int8_t *)dbtp->data, sizeof(rectype));
	if(op == DB_TXN_FORWARD_ROLL && rectype == DB_stp_ckp){
		/* Recovery the TT table from the checkpoint TT log record */
		bp = (u_int8_t *)dbtp->data;
		memcpy(&nr_entries, bp + sizeof(rectype) + sizeof(u_int32_t), sizeof(nr_entries));

		bp += sizeof(rectype) + sizeof(u_int32_t) + sizeof(nr_entries);

		STP_DEBUG( printf("\tnr_entries = %d\n", nr_entries) );

		while (nr_entries >= 1) {
/* Modified by Huilong at 2:45 7/12/2006 */
//			memcpy(&(txnid.envid), bp, sizeof(time_t));
//			bp += sizeof(time_t);
                        memcpy(&(txnid.envid), bp, sizeof(ENV_ID_TYPE));
                        bp += sizeof(ENV_ID_TYPE);
/* end of modification */
			memcpy(&(txnid.txnid), bp, sizeof(u_int32_t));
			bp += sizeof(u_int32_t);
			memcpy(&time, bp, sizeof(TV));
			bp += sizeof(TV);
			memcpy(&od_pgcnt, bp, sizeof(od_pgcnt));
			bp += sizeof(od_pgcnt);

			/* Check if the txn is committed. If so, add it into TT table
			 * If the txn is not committed yet, do nothing because it might
			 * be added to TT later by txn_commit, or it is aborted
			 */


			if(!IS_INVALID_TIME(time)){
				STP_DEBUG( printf("\tFind the commited txnid ---------- %d\n", txnid.txnid) );
				STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);
				if (tt != NULL)
					STP_ERROR();
				if ((tt = SH_TAILQ_FIRST(&(stp->free_tt_ent), __stp_tt_entry)) != NULL)
					SH_TAILQ_REMOVE(&(stp->free_tt_ent), tt, links, __stp_tt_entry);
				else
					STP_ERROR();
				tt->txnid = txnid;
				memcpy(&(tt->time), &time, sizeof(TV));
				SH_TAILQ_INIT(&(tt->btn));
				tt->od_pgcnt = od_pgcnt;
				tt_index = STP_TT_HASH(dbstp, txnid);
				SH_TAILQ_INSERT_HEAD(&(dbstp->tt_htab[tt_index]), tt, links, __stp_tt_entry);
			}
			else {
				STP_DEBUG( printf("\tFind the uncommited txnid ---------- %d\n", txnid.txnid) );
				STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);
				if (tt != NULL){

/* Huilong 10/9/2006: comment this line. The time should not be changed as invalid time */
/* */
					//memcpy(&(tt->time), &time, sizeof(TV));
/* Huilong */
					tt->od_pgcnt = od_pgcnt;
				}
			}
			 /*
			if( !IS_INVALID_TIME(time) )
			{
				/* Insert a new TT entry for txnid *
			}
			*/
			nr_entries --;
		}
		STP_DEBUG(print_tt(dbstp) );
		return (0);

		err:
	    	printf("ERROR in __stp_recover_ckp.\n");
    		exit(1);
	}
	return (0);
}

int
__stp_recover_ckp(dbenv, dbtp, lsnp, op, info)
    DB_ENV *dbenv;
    DBT *dbtp;
    DB_LSN *lsnp;
    db_recops op;
    void *info;
{
	int ret = 0;

	 if( *((u_int32_t *)dbtp->data) == 11){

		 if((ret = log_get(dbenv, lsnp, dbtp, DB_PREV)) == 0){
		   /*__stp_recover_ckp_help(dbenv, dbtp, lsnp, DB_TXN_FORWARD_ROLL, info); */
			 __stp_recover_ckp_help(dbenv, dbtp, lsnp, op, info);
			 log_get(dbenv, lsnp, dbtp, DB_NEXT);
		 }

	 }
	 return ret;
}

/*
 * _stp_recover_txn_commit(dbenv, dbtp, lsnp, op, info) --
 *   function to recover the stp_tt_insert_log. It inserts a
 *   TT entry during UNDO to keep track on the list of committed
 *   txns and updates the txn time and od_pgcnt during REDO
 *   It is called by __db_dispatch (db_dispatch.c)
 */
int
_stp_recover_txn_commit(dbenv, dbtp, lsnp, op, info)
    DB_ENV *dbenv;  /* The environment. */
    DBT *dbtp;      /* The log record upon which to recovery. */
    DB_LSN *lsnp;   /* The lsn of the record being recoveried */
    db_recops op;   /* Redo this op (or undo it ). */
    void *info;
{
	STP *stp;
    u_int32_t	rectype;
    u_int8_t	*bp;
    STP_TT_ENTRY	*tt;
    STP_BTN 	*btn;
    DB_STP		*dbstp;
#ifdef BDB_4
    DB_LOGC		*logc;
#endif
    STP_ID		txnid;
    TV		time;
    u_int32_t	od_pgcnt;
    u_int32_t	tt_prev, tt_index, tt_new, tt_tmp;
    DB_LSN		first_lsn;
    struct __db	*file_dbp;
    DB_MPOOLFILE	*mpf;
    PAGE		*pagep;
    int32_t		fileid;
    int		ret;
    DBC *dbc;

	/* undo this log */
    if(op == DB_TXN_BACKWARD_ROLL || op == DB_TXN_FORWARD_ROLL){
		STP_STOP_WORKING();

		STP_DEBUG( printf("\tEnter _stp_recover_txn_commit ...\n") );
		dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

		/* If the environment is NON-TEMPORAL, the stamper do nothing. */
		if (dbstp == NULL)
		return 0;
		stp = dbstp->reginfo.primary;

		bp = (u_int8_t *)dbtp->data;
		bp += sizeof(rectype);
/* Modified by Huilong at 2:53pm 7/12/2006 */
//		memcpy(&(txnid.envid), bp, sizeof(time_t));
//		bp += sizeof(time_t);
                memcpy(&(txnid.envid), bp, sizeof(ENV_ID_TYPE));
                bp += sizeof(ENV_ID_TYPE);
/* end of modification */
                memcpy(&(txnid.txnid), bp, sizeof(u_int32_t));
		bp += sizeof(u_int32_t);
		memcpy(&time, bp, sizeof(TV));
		bp += sizeof(TV);
		memcpy(&od_pgcnt, bp, sizeof(od_pgcnt));
		bp += sizeof(od_pgcnt);

		/* Find the entry to TT */
		STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);
		/* Insert the entry into TT if not exist */
		if (tt == NULL){
			if(op == DB_TXN_BACKWARD_ROLL){
				STP_DEBUG( printf("\t         UNDO _stp_recover_txn_commit ...\n") );
				if ((tt = SH_TAILQ_FIRST(&(stp->free_tt_ent), __stp_tt_entry)) != NULL)
					SH_TAILQ_REMOVE(&(stp->free_tt_ent), tt, links, __stp_tt_entry);
				else
					STP_ERROR();
				tt->txnid = txnid;
				INVALID_TIME(tt->time);
				tt->od_pgcnt = 0;

				SH_TAILQ_INIT(&(tt->btn));

				tt_index = STP_TT_HASH(dbstp, txnid);
				SH_TAILQ_INSERT_HEAD(&(dbstp->tt_htab[tt_index]), tt, links, __stp_tt_entry);
			}
			else
			STP_ERROR();
		}
		else {
		  /*tt->txnid = txnid;
			memcpy(&(tt->time), &time, sizeof(TV));
			tt->od_pgcnt = od_pgcnt;
		  */
			if(op == DB_TXN_BACKWARD_ROLL)
				STP_ERROR();
			else{
				STP_DEBUG( printf("\t         REDO _stp_recover_txn_commit ...\n") );
/* Huilong 10/10/2006 */
				memcpy(&(tt->time), &time, sizeof(TV));
//				tt->od_pgcnt = od_pgcnt;
/* Huilong */
			}
		}
		return (0);

		err:
	    	printf("ERROR in _stp_recover_txn_commit.\n");
    		exit(1);
	}
    return (0);
}


/*
 * _stp_recover_decr_od_pgcnt(dbenv, dbtp, lsnp, op, info) --
 *    function to recover the stp_tt_modify_log. It decrements
 *    the od_pgcnt by 1 during REDO
 *    It is called by __db_dispatch (db_dispatch.c)
 */
int
_stp_recover_decr_od_pgcnt(dbenv, dbtp, lsnp, op, info)
    DB_ENV *dbenv;
    DBT *dbtp;
    DB_LSN *lsnp;
    db_recops op;
    void *info;
{
    STP *stp;
    u_int32_t	rectype;
    u_int8_t	*bp;
    STP_TT_ENTRY	*tt;
    STP_BTN 	*btn;
    DB_STP		*dbstp;
#ifdef BDB_4
    DB_LOGC		*logc;
#endif
    STP_ID		txnid;
    TV		time;
    u_int32_t	od_pgcnt;
    u_int32_t	tt_prev, tt_index, tt_new, tt_tmp;
    DB_LSN		first_lsn;
    struct __db	*file_dbp;
    DB_MPOOLFILE	*mpf;
    PAGE		*pagep;
    int32_t		fileid;
    int		ret;
    DBC *dbc;
/* Huilong */
    u_int32_t pgno;
/* Huilong */

	/* undo this log */
    if(op == DB_TXN_FORWARD_ROLL){
		STP_STOP_WORKING();
		STP_DEBUG( printf("\tEnter _stp_recover_decr_od_pgcnt ...\n") );

		dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

		/* If the environment is NON-TEMPORAL, the stamper do nothing. */
		if (dbstp == NULL)
		return 0;
		stp = dbstp->reginfo.primary;

		bp = (u_int8_t *)dbtp->data;
		bp += sizeof(rectype);
/* Modified by Huilong at 7/12/2006 */
//		memcpy(&(txnid.envid), bp, sizeof(time_t));
//		bp += sizeof(time_t);
                memcpy(&(txnid.envid), bp, sizeof(ENV_ID_TYPE));
                bp += sizeof(ENV_ID_TYPE);
/* end of modification */
		memcpy(&(txnid.txnid), bp, sizeof(u_int32_t));
		bp += sizeof(u_int32_t);

/* Huilong 10/12/2006 */
/*		memcpy(&time, bp, sizeof(TV));
		bp += sizeof(TV);
		memcpy(&od_pgcnt, bp, sizeof(od_pgcnt));
		bp += sizeof(od_pgcnt);*/
                memcpy(&fileid, bp, sizeof(fileid));
                bp += sizeof(fileid);
                memcpy(&pgno, bp, sizeof(pgno));
                bp += sizeof(pgno);
/* Huilong */

		/* find txn in TT */
		STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);
		if (tt == NULL)
			STP_ERROR();

/* Huilong 10/12/2006 */
       // stamped the page, to be continued .....
       __stp_recovery_stamp_page(dbenv, txnid, fileid, pgno);
      if(!__stp_lookup_incr_set(txnid, fileid, pgno)){
/* Huilong */
 		/* decrement the od_pgcnt by 1 */
 		tt->od_pgcnt --;
 		if(tt->od_pgcnt < 0)
 			STP_ERROR();
 		/* reclaim the tt entry */
		if(tt->od_pgcnt == 0){
			__tt_reclaim(dbstp, tt);
		}
/* Huilong 10/12/2006 */
      }
/* Huilong */
		return (0);


		err:
	    	printf("ERROR in _stp_recover_decr_od_pgcnt.\n");
    		exit(1);
	}
    return (0);
}

/* Huilong 10/10/2006 */

/*
 * _stp_recover_incr_od_pgcnt(dbenv, dbtp, lsnp, op, info) --
 *    function to recover the stp_tt_modify_log. It decrements
 *    the od_pgcnt by 1 during REDO
 *    It is called by __db_dispatch (db_dispatch.c)
 */
int
_stp_recover_incr_od_pgcnt(dbenv, dbtp, lsnp, op, info)
    DB_ENV *dbenv;
    DBT *dbtp;
    DB_LSN *lsnp;
    db_recops op;
    void *info;
{
    STP *stp;
    u_int32_t	rectype;
    u_int8_t	*bp;
    STP_TT_ENTRY	*tt;
    STP_BTN 	*btn;
    DB_STP		*dbstp;
#ifdef BDB_4
    DB_LOGC		*logc;
#endif
    STP_ID		txnid;
    TV		time;
    u_int32_t	od_pgcnt;
    u_int32_t	tt_prev, tt_index, tt_new, tt_tmp;
    DB_LSN		first_lsn;
    struct __db	*file_dbp;
    DB_MPOOLFILE	*mpf;
    PAGE		*pagep;
    int32_t		fileid;
    int		ret;
    DBC *dbc;
    u_int32_t pgno;

	/* undo this log */
    if(op == DB_TXN_FORWARD_ROLL){
		STP_STOP_WORKING();
		STP_DEBUG( printf("\tEnter _stp_recover_incr_od_pgcnt ...\n") );

		dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

		/* If the environment is NON-TEMPORAL, the stamper do nothing. */
		if (dbstp == NULL)
		return 0;
		stp = dbstp->reginfo.primary;

		bp = (u_int8_t *)dbtp->data;
		bp += sizeof(rectype);
                memcpy(&(txnid.envid), bp, sizeof(ENV_ID_TYPE));
                bp += sizeof(ENV_ID_TYPE);
		memcpy(&(txnid.txnid), bp, sizeof(u_int32_t));
		bp += sizeof(u_int32_t);

                memcpy(&fileid, bp, sizeof(fileid));
                bp += sizeof(fileid);
                memcpy(&pgno, bp, sizeof(pgno));
                bp += sizeof(pgno);

		/* find txn in TT */
		STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);
		if (tt == NULL)
			return 0;

// to be continued ...
        	// stamped the page
        	__stp_recovery_stamp_page(dbenv, txnid, fileid, pgno);
        	__stp_insert_incr_set(txnid, fileid, pgno);
//		printf("_stp_recover_incr_od_pgcnt is called: pgno=%d\n", pgno);
		return (0);


		err:
	    	printf("ERROR in _stp_recover_incr_od_pgcnt.\n");
    		exit(1);
	}
    return (0);
}

/* Huilong */

/*
 * __stp_process_log_entry(dbenv, dbtp, lsnp, op, info) --
 *  Null function does nothing. It is registered for __stp_ckp_log since
 *  since recovery for such log is called directly with another function.
 *	But we need to define these null functions to make a fool of the BDB
 *	recovery system, because calling a corresponding recovery function
 *	for each log record is the way BDB working.
 */
int
__stp_process_log_entry(dbenv, dbtp, lsnp, op, info)
    DB_ENV *dbenv;
    DBT *dbtp;
    DB_LSN *lsnp;
    db_recops op;
    void *info;
{
	STP_DEBUG( printf("  ENTER __stp_process_log_entry\n") );
    return (0);
}



/*********************************************
 *      Register Recovery Operations         *
 *********************************************/

/*
 * __stp_init_recover(dbenv) --
 *  Register the recovery functions for the newly added log records
 *	for STP module. This is called in __dbenv_open (env/env_open.c).
 *	See detailed information at __stp_process_log_entry.
 */
int
__stp_init_recover(dbenv)
    DB_ENV *dbenv;
{
    int ret;

    STP_STOP_WORKING();

    STP_DEBUG( printf("\tENTER __stp_init_recover()\n"); fflush(stdout) );

	/* register recovery function __stp_process_log_entry for stp_ckp log */
    if ((ret = __db_add_recovery(dbenv,
#ifdef BDB_4
				 &dbenv->recover_dtab, &dbenv->recover_dtab_size,
#endif
				 __stp_process_log_entry, DB_stp_ckp)) != 0)
	return (ret);

	/* register recovery function _stp_recover_txn_commit for stp_txn_commit log */
    if ((ret = __db_add_recovery(dbenv,
#ifdef BDB_4
				 &dbenv->recover_dtab, &dbenv->recover_dtab_size,
#endif
				 _stp_recover_txn_commit, DB_stp_txn_commit)) != 0)
	return (ret);

	/* register recovery function _stp_recover_decr_od_pgcnt for stp_decr_od_pgcnt log */
    if ((ret = __db_add_recovery(dbenv,
#ifdef BDB_4
				 &dbenv->recover_dtab, &dbenv->recover_dtab_size,
#endif
				 _stp_recover_decr_od_pgcnt, DB_stp_decr_od_pgcnt)) != 0)
	return (ret);

/* Huilong 10/10/2006 */
	/* register recovery function _stp_recover_incr_od_pgcnt for stp_incr_od_pgcnt log */
    if ((ret = __db_add_recovery(dbenv,
#ifdef BDB_4
				 &dbenv->recover_dtab, &dbenv->recover_dtab_size,
#endif
				 _stp_recover_incr_od_pgcnt, DB_stp_incr_od_pgcnt)) != 0)
	return (ret);
/* Huilong */

/* Modified by Huilong at 7/14/2006*/
#ifdef TBDB_AUDIT
    /********************************************
     * For AUD module
     ********************************************/
    if ((ret = __db_add_recovery(dbenv,
#ifdef BDB_4
                                 &dbenv->recover_dtab, &dbenv->recover_dtab_size,
#endif
                                 __aud_tol_recover_log, DB_aud_tol_insert)) != 0)
        return (ret);
#endif
/* End of modification */

    STP_DEBUG( printf("\tEXIT __stp_init_recover()\n"); fflush(stdout) );

#ifdef STAMPER_DEBUG
    printf("\tEXIT __stp_init_recover()\n");fflush(stdout);
#endif

/* Huilong 10/11/2006 */
   stp_incr_set_head = NULL;
/* Huilong */
    return (0);
}



/***********************************
 *     Debug: Log Printout         *
 ***********************************/


/*
 * __stp_ckp_recover --
 *	print out all the log entries from the most recent log
 *  to the last ckp.
 *	It is called in __db_apprec(), after the pass#0.
 */
int
__stp_ckp_recover(dbenv)
    DB_ENV *dbenv;
{
    STP *stp;
    u_int32_t	rectype, nr_entries, tid;
    u_int8_t	*bp;
    STP_TT_ENTRY	*tt;
    STP_BTN 	*btn;
    DB_STP		*dbstp;
#ifdef BDB_4
    DB_LOGC		*logc;
#endif
    DBT		data;
    DB_LSN		lsn;
    STP_ID		txnid;
    TV		time;
    u_int32_t	od_pgcnt, od_pgcnt_tmp;
    u_int32_t	tt_prev, tt_index, tt_new, tt_tmp;
    DB_LSN		first_lsn;
    struct __db	*file_dbp;
    DB_MPOOLFILE	*mpf;
    PAGE		*pagep;
    int32_t		fileid;
    u_int32_t	pgno;
    int		ret, found;

#ifdef TBDB_AUDIT
    u_int32_t j;
    TV t;
    DB_AUD *dbaud;
    AUD *aud;
    AUD_TOL_ENTRY *tolep, *new_tolep, *next_tolep;
#endif

    DBC *dbc;

    STP_STOP_WORKING();

    /* STP_DEBUG( printf("\tEnter __stp_ckp_recover ...\n") ); */

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    stp = dbstp->reginfo.primary;

/* Huilong */
#ifdef TBDB_AUDIT
   dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
    if (dbaud == NULL)
        return 0;
    aud = dbaud->reginfo.primary;
#endif
/* End */

    /* Allocate a cursor for the log. */
#ifdef BDB_4
    if ((ret = dbenv->log_cursor(dbenv, &logc, 0)) != 0)
	STP_ERROR();
#endif

    memset(&data, 0, sizeof(data));
    found = 0;
#ifdef BDB_4
    if ((ret = logc->get(logc, &first_lsn, &data, DB_LAST)) != 0)
#endif
#ifdef BDB_3
    if ((ret = log_get(dbenv, &first_lsn, &data, DB_LAST)) != 0)
#endif
	STP_ERROR();

    /* From the end of the log, search backwards for the first TT
     *	ckp log.
     */

     printf("\n Print out all the log entries: \n");
    do {
	/* rectype is at the beginning. */
	memcpy(&rectype, (u_int8_t *)data.data, sizeof(rectype));
	memcpy(&tid, (u_int8_t *)data.data + sizeof(rectype), sizeof(tid));

	if(rectype == DB_stp_decr_od_pgcnt ){
	    bp = (u_int8_t *)data.data;
	    bp += sizeof(rectype);
/* Modified by Huilong at 2:56pm 7/12/2006 */
//	    memcpy(&(txnid.envid), bp, sizeof(time_t));
//	    bp += sizeof(time_t);
            memcpy(&(txnid.envid), bp, sizeof(ENV_ID_TYPE));
            bp += sizeof(ENV_ID_TYPE);
/* end of modification */
	    memcpy(&(txnid.txnid), bp, sizeof(u_int32_t));
	    bp += sizeof(u_int32_t);
	    memcpy(&fileid, bp, sizeof(fileid));
	    bp += sizeof(fileid);
	    memcpy(&pgno, bp, sizeof(pgno));
	    bp += sizeof(pgno);
		printf("log record: rectype=%d txnid=(%ld, %ld) fileid=%d pgno=%d\n",
		   rectype, txnid.envid, txnid.txnid, fileid, pgno);

	}
	else if(rectype == DB_stp_txn_commit){
	    bp = (u_int8_t *)data.data;
	    bp += sizeof(rectype);
/* Modified by Huilong at 7/12/2006 */
//	    memcpy(&(txnid.envid), bp, sizeof(time_t));
//	    bp += sizeof(time_t);
            memcpy(&(txnid.envid), bp, sizeof(ENV_ID_TYPE));
            bp += sizeof(ENV_ID_TYPE);
/* end of modification */
	    memcpy(&(txnid.txnid), bp, sizeof(u_int32_t));
	    bp += sizeof(u_int32_t);
	    memcpy(&time, bp, sizeof(TV));
	    bp += sizeof(TV);
	    memcpy(&od_pgcnt, bp, sizeof(od_pgcnt));
	    bp += sizeof(od_pgcnt);
		printf("log record: rectype=%d txnid=(%ld, %ld) time=(%d, %d) od_pgcnt=%d\n",
		   rectype, txnid.envid, txnid.txnid, time.tv_sec, time.tv_usec, od_pgcnt);

	}
	else if(rectype == DB_stp_ckp){
		printf("log record: rectype=%d  txnid=%d\n", rectype, tid);
		break;
	}
	else
		printf("log record: rectype=%d  txnid=%d\n", rectype, tid);
	/*
	if (rectype == DB_stp_ckp) {
	    found = 1;

	    STP_DEBUG( printf("\tfound!!!!TT\n") );

	    break;
	}
	*/
#ifdef BDB_4
    } while ((ret = logc->get(logc, &lsn, &data, DB_PREV)) == 0);
#endif
#ifdef BDB_3
    } while ((ret = log_get(dbenv, &lsn, &data, DB_PREV)) == 0);
#endif

#ifdef TBDB_AUDIT
    /* Recovery the TOL from the checkpoint TOL log record */
    memcpy(&j, bp, sizeof(j));
    bp += sizeof(j);
    memcpy(aud->hash_value, bp, AUD_HASH_LEN);
    bp += AUD_HASH_LEN;

#ifdef STAMPER_DEBUG
    printf("\tj = %d\n", j);
#endif

/*    while (j >= 1) {
        memcpy(&t, bp, sizeof(TV));
        bp += sizeof(TV);

    
        for (tolep = SH_TAILQ_FIRST(&aud->tol, __aud_tol_entry);
             tolep != NULL;
             tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry)) {
            next_tolep = SH_TAILQ_NEXT(tolep, links, __aud_tol_entry);
            if ((next_tolep == NULL)
                || (next_tolep->time.tv_sec > t.tv_sec)
                || ((next_tolep->time.tv_sec == t.tv_sec)
                    && (next_tolep->time.tv_usec >= t.tv_usec)))
                break;
        }

       
        if ((tolep != NULL)
            && (tolep->time.tv_sec == t.tv_sec)
            && (tolep->time.tv_usec == t.tv_usec)) {
            printf("50\n");
            goto err;
        }

        
        new_tolep = SH_TAILQ_FIRST(&aud->free_tol_entries, __aud_tol_entry);
        if (new_tolep != NULL)
            SH_TAILQ_REMOVE(&aud->free_tol_entries, new_tolep,
                            links, __aud_tol_entry);
        else {
            printf("60\n");
            goto err;
        }

        
        memset(new_tolep, 0, sizeof(AUD_TOL_ENTRY));
        new_tolep->time = t;
        new_tolep->ready = 1;
        memcpy(new_tolep->hash_value, bp, AUD_HASH_LEN);
        bp += AUD_HASH_LEN;
        if (tolep)
            SH_TAILQ_INSERT_AFTER(&aud->tol, tolep, new_tolep,
                                  links, __aud_tol_entry);
        else
            SH_TAILQ_INSERT_HEAD(&aud->tol, new_tolep, links, __aud_tol_entry);

        j --;
    }*/

#endif

	STP_DEBUG( printf("\tExit __stp_ckp_recover ...\n") );
    return (0);

  err:	__db_err(dbenv, "__stp_ckp_recover\n");
    exit(1);
    return (STP_ERROR_TXN_ABORT);
}



/* __stp_record_recover --
 *	Process a record in a page and stamp it if necessary.
 *  It is called from __stp_page()
 */
int
__stp_record_recover(dbstp, bt, idtimep, action, tid)
    DB_STP              *dbstp;
    STP_BT_ENTRY	*bt;
    IdOrTime		*idtimep;
    u_int32_t		action;
    STP_ID	        tid;
{
    STP *stp;
    STP_TT_ENTRY	*tt;
    STP_BTN     *btnp;
    int		type;
    int		getstampped;
    u_int32_t	after, tt_index, i;
    STP_ID	txnid;
    u_int32_t	flags;
    int ret;

    STP_STOP_WORKING();
    if (stp_pause)
	return 0;
    STP_DEBUG( printf("  ENTER __stp_record_recover action %d\n", action) );
    txnid = idtimep->stamperId;
    ret = UNCHANGED;
    stp = dbstp->reginfo.primary;

    if(tid.envid.tv_sec!=txnid.envid.tv_sec ||
       tid.envid.tv_msec!=txnid.envid.tv_msec ||
       tid.txnid!=txnid.txnid) return UNCHANGED;

    STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);
    /* Un-begun (no entry for this txn in TT) */
    if (tt == NULL) {
        return UNCHANGED;
    }

	btnp = NULL;
	flags = 0;

    /* Create a BTN node if not exist when action is PAGE_READ or PAGE_WRITE */
    if((action == PAGE_READ || action == PAGE_WRITE) && bt != NULL){
		/* Find BTN node in BT entry. */
		for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
			 btnp != NULL;
			 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node)) {
			if ( STPID_EQUAL(btnp->tt->txnid, txnid) )
			break;
		}

		/* If BTN not found, add one */
		if (btnp == NULL) {
			if ((btnp = SH_TAILQ_FIRST(&(stp->free_bt_node), __stp_bt_node)) != NULL)
			SH_TAILQ_REMOVE(&(stp->free_bt_node), btnp, hlinks, __stp_bt_node);
			else {
			fprintf(stderr, "STP (__stp_record_recover): run out of BTN\n");
			STP_ERROR();
			}

			btnp->tt = tt;
			flags = 0;

			/* Set read_state to UNSTAMPED if page read, NONE if page write */
			if(action == PAGE_READ)
				LF_CLR(STP_READ_NONE);
			else
				LF_SET(STP_READ_NONE);

			/* There is no need to set up the write_state yet since page_write will set it after stamping */
			/*LF_CLR(STP_WRITE_STAMPED);*/
			btnp->flags = flags;

			/* Vertical & horizontal insert of the newly allocated BTN node */
			SH_TAILQ_INSERT_HEAD(&(bt->btn), btnp, hlinks, __stp_bt_node);
			SH_TAILQ_INSERT_HEAD(&(tt->btn), btnp, vlinks, __stp_bt_node);
		}
    }

    if (IS_INVALID_TIME(tt->time)) {
		/* Uncommitted -- nothing to do */

    } else { /* Commmitted */
		STP_DEBUG( printf("\t\tSTAMPing!!!\n") );

		/* Now we STAMP this record */
		memcpy(&idtimep->curr, &(tt->time), sizeof(TV));
		ret = CHANGED;
    }

	STP_DEBUG( printf("  Exit __stp_record_recover action %d\n", action) );
    /* If this record got stampped, we return 1, otherwise return 0. */
    return (ret);

  err:
    printf("ERROR in __stp_record_recover: \n");
    exit(1);
}


/*
 * __stp_page --
 *	Scan a page and give the stamps.
 *  It is called by: __stp_pgread(), __stp_pgwrite(), __stp_mpfput(), and __stp_dbret()
 *  with action as PAGE_READ, PAGE_WRITE, MPF_PUT, and MPF_PUT, accordingly.
 */
int
__stp_page_recover(dbenv, dbmfp, addrp, action, tid)
    DB_ENV			*dbenv;
    DB_MPOOLFILE		*dbmfp;
    void			*addrp;
    u_int32_t		action;
    STP_ID          tid;
{
    DB_MPOOL		*dbmp;
    u_int32_t		index;
    PAGE		*h;
    void		*test;
    db_pgno_t		pgno;
    db_indx_t		indx;
    BKEYDATA		*bk;
    DBT			pg_dbt;
    char		*filename;
    TTDATA		ttdata;
    DB_STP		*dbstp;
    STP			*stp;
    STP_TT_ENTRY	*tt;
    DB_HASHTAB          *tt_htab;
    STP_BT_ENTRY	*bt;
    STP_BTN 		*btn, *btnp, *temp;
    BH			*bhp;
    u_int32_t		result, tl_file_head, tmp;
    u_int32_t           tl_index, tl_head_index, bt_prev, bt_index;
    u_int32_t		bt_head_index1, bt_tmp1, tt_prev, tt_index, tt_index1, tt_tmp1;
    u_int32_t           bt_index1, tt_tmp;
    u_int8_t 		fileid[DB_FILE_ID_LEN];
    int			unstamped, ret, i, bt_found, bt_empty, fileid_equal, changed, increment;
    MPOOL		*c_mp;
    DB_LSN		lsn;
    struct __db		*dbp;
    int32_t		ndx;
    DB_LOG		*logp;
    DB_ENTRY		*dbe;
    db_indx_t		step;
    MPOOLFILE   *mfp;

/* Huilong 7/29/06 */
    int         has_ovflow = 0;
/* Huilong */

    int key_value;
    BOVERFLOW *bo;

    STP_STOP_WORKING();
    if (stp_pause)
	return 0;

	STP_DEBUG( printf("  ENTER __stp_page action %d\n", action) );

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    /* If the database is NON-TEMPORAL, the stamper do nothing. */
    if (!__stp_istemporal(dbenv, dbmfp))
	return 0;

    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

    dbmp = dbmfp->dbmp;
    stp = (dbstp->reginfo).primary;
    logp = dbenv->lg_handle;

    /* Convert the page address to a buffer header. */
    bhp = (BH *)((u_int8_t *)addrp - SSZA(BH, buf));

    mfp = R_ADDR(dbmp->reginfo, bhp->mf_offset);

    /* Convert the buffer header to a cache. */
#ifdef BDB_4
    c_mp = dbmp->reginfo[NCACHE(dbmp->reginfo[0].primary,
				bhp->mf_offset,
				bhp->pgno)].primary;
#endif
#ifdef BDB_3
    c_mp = BH_TO_CACHE(dbmp, bhp);
#endif

/* Huilong 7/24/06 */
//    R_LOCK(dbenv, &dbstp->reginfo);
/* Huilong */
    unstamped = 0; /* This page has unstamped record(s) */
    h = (PAGE *)bhp->buf;
    pgno = bhp->pgno;
    changed = 0;

/* Huilong 7/24/06*/
//    tt_htab = dbstp->tt_htab;
/* Huilong*/

    /* Get the fileid */
#ifdef BDB_4
    memcpy(fileid, dbmfp->fileid, DB_FILE_ID_LEN);
#endif
#ifdef BDB_3
    memcpy(fileid, R_ADDR(dbmp->reginfo, dbmfp->mfp->fileid_off), DB_FILE_ID_LEN);
#endif

	bt = NULL;

    /* Walk through the data DBT on this page doing the stamp */
    switch (TYPE(h)) {
    case P_INVALID:		/* Invalid page type. */
	goto done;
    case __P_DUPLICATE:	/* Duplicate. DEPRECATED in 3.1 */
	goto done;
    case P_HASH:		/* Hash. */
	goto done;
    case P_IBTREE:		/* Btree internal. */

/* Huilong 7/24/06 */
        R_LOCK(dbenv, &dbstp->reginfo);
        tt_htab = dbstp->tt_htab;
/* Huilong */

	/* This is possible. When doing a root page split, if the
	   original root is a leaf page containing some unstamped
	   records, we should decrease those transactions' od_pgcnt
	   by one when pgwrite the page onto disk.
	*/
	if (action == PAGE_WRITE) {
	    STP_DEBUG( printf("\tIN stp_page<P_IBTREE>(action=%d) pgno = %d, type = %d, num_entry = %d\n", \
			      action, pgno, TYPE(h), NUM_ENT(h)) );

	    /* Find the page(pgno+fileid) in BT */
	    STP_BT_FIND(dbstp, pgno, fileid, bt);

	    /* If no BT entry for this page or TL is empty, this
	     *  page must not be in the situation described above.
	     *  And we do nothing about it.
	     */
	    if (bt == NULL){
/* Huilong 7/24/2006 */
                R_UNLOCK(dbenv, &dbstp->reginfo);
/* Huilong */
		goto done;
            }
	}
	goto update;

    case P_IRECNO:		/* Recno internal. */
	goto done;
    case P_LDUP:		/* Off-page duplicate leaf. */
    case P_LRECNO:		/* Recno leaf. */
    case P_LBTREE:		/* Btree leaf. */

/* Huilong 7/24/06 */
        R_LOCK(dbenv, &dbstp->reginfo);
        tt_htab = dbstp->tt_htab;
/* Huilong */

	/* If called in __memp_pgread or __memp_pgwrite, we have
	 *	to deal with BT table.
	 */
	if ((action == PAGE_READ) || (action == PAGE_WRITE)) {
	    /* Find the page(pgno+fileid) in BT */
	    STP_BT_FIND(dbstp, pgno, fileid, bt);

	    /* If there is no entry for this page in BT ...
	     * In "PAGE_READ": it's because this is the first time to process
	     *	this page. We simply add an entry for it into the BT.
	     * In "PAGE_WRITE": it's because this page is newly created in mem.
	     *	We simply add anentry for it into the BT.
	     */
	    if (bt == NULL) {
			/* Allocate a new BT entry for this page */
			STP_BT_INSERT(dbstp, pgno, fileid, bt);

			STP_DEBUG( printf("  *** new BT inserted!! bt=%d\n", bt) );
			/*STP_DEBUG( print_bt(dbstp) );*/

			if (bt == NULL) {
				fprintf(stderr, "STP (__stp_page): Run out of BT entries\b");
				STP_ERROR();
			}

			/* Initialize the new BT entry */
			bt->pgno = pgno;
			SH_TAILQ_INIT(&(bt->btn));
			memcpy(bt->fileid, fileid, DB_FILE_ID_LEN);
		}
	}

	STP_DEBUG( printf("\tIN stp_page(action=%d) pgno = %d, type = %d, num_entry = %d\n", \
			  action, pgno, TYPE(h), NUM_ENT(h)) );

	changed = 0; unstamped = 0;
	/* Walk through this page.
	 * For each key/data pair of this page, we do the stamp
	 */
	for (indx = 0;
	     indx < NUM_ENT(h);
	     indx += 1) {

	    if (TYPE(h)==P_LBTREE)
		indx++;
	    if (indx >= NUM_ENT(h))
		break;
#ifdef BDB_4
	    bk = GET_BKEYDATA(dbp, h, indx);
#endif
#ifdef BDB_3
	    bk = GET_BKEYDATA(h, indx);
#endif

	    pg_dbt.data = bk->data;
	    pg_dbt.size = bk->len;

/* Modification by Huilong for overflow records */
	    /* Get the start/stop time from data DBT */
//	    memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA)); // original code
if (B_TYPE(bk->type) != B_OVERFLOW) {
            memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA));
}
else{
            bo = (BOVERFLOW *)bk;
            memcpy(&ttdata, (char *)(&(bo->ttdata)), sizeof(TTDATA));
}
/* end of modification by Huilong */

	    /* If the data DBT is shorter than TTDATA the
	     *   memory referencing will cause error !
	     *   So we check here. If so, it means this data
	     *   doesn't contain TTDATA at all since it is
	     *   not long enough to contain one.
	     * An on-page record pointing to the OPD page is like
	     *   this.
	     * - Shilong
	     */
	    if (pg_dbt.size < sizeof(TTDATA)) {
		STP_DEBUG( printf("\t\tDBT(%d):(State=%d, INVALID\n", indx, ttdata.state) );
		continue;
	    }
/* Huilong, 7/14/06*/
#ifdef TBDB_AUDIT
            if ((action == PAGE_WRITE) || (action == MPF_PUT)) {
                if (__aud_tuple_hash(dbenv, pg_dbt) == 1)
                    changed = 1;
                memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA));
            }
#endif
/* end 7/14/06 */
	    /* Stamp the "transaction start time" of this data record */
	    switch (ttdata.state) {
	    case ID_ID:
		if (__stp_record_recover(dbstp, bt, &(ttdata.start), action, tid) == CHANGED) {
		    ttdata.state = TIME_ID;
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    case ID_UC:
		if (__stp_record_recover(dbstp, bt, &(ttdata.start), action, tid) == CHANGED) {
		    ttdata.state = TIME_UC;
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    case ID_TIME:
		if (__stp_record_recover(dbstp, bt, &(ttdata.start), action, tid) == CHANGED) {
		    ttdata.state = TIME_TIME;
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    }

	    /* Stamp the "transaction stop time" of this data record */
	    switch (ttdata.state) {
	    case ID_ID:
		if (__stp_record_recover(dbstp, bt, &(ttdata.stop), action, tid) == CHANGED) {
		    ttdata.state = ID_TIME;
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    case TIME_ID:
		if (__stp_record_recover(dbstp, bt, &(ttdata.stop), action, tid) == CHANGED) {
		    ttdata.state = TIME_TIME;
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    }

#ifdef _STAMPER_DEBUG_
	    printf("    DBT(%d):(State=%d, ", indx, ttdata.state);
	    if ((ttdata.state==ID_ID)||(ttdata.state==ID_UC)||(ttdata.state==ID_TIME)){
		printf("StartID=(%ld, %ld), ", ttdata.start.stamperId.envid,
		       ttdata.start.stamperId.txnid);
	    }else
		printf("StartTime=(%ld, %ld), ", ttdata.start.curr.tv_sec,
		       ttdata.start.curr.tv_usec);

	    if ((ttdata.state==ID_ID)||(ttdata.state==TIME_ID)){
		printf("StopID=%ld)", ttdata.stop.stamperId);
	    }else if((ttdata.state==ID_UC)||(ttdata.state==TIME_UC)) {
		printf("StopTime=UC");
	    }else
		printf("StopTime=(%ld, %ld)", ttdata.stop.curr.tv_sec,
		       ttdata.stop.curr.tv_usec);
	    printf(":%s\n",GET_BKEYDATA(h, indx-1)->data);
#endif


/* Modification by Huilong for overflow records */
	    /* Copy the stampped ttdata back to the data record. */
	    //memcpy(pg_dbt.data, &ttdata, sizeof(TTDATA));
if (B_TYPE(bk->type) != B_OVERFLOW) {
	    /* Copy the stampped ttdata back to the data record. */
	    memcpy(pg_dbt.data, &ttdata, sizeof(TTDATA));
}
else{
	    bo = (BOVERFLOW *)bk;
            memcpy(&(bo->ttdata), &ttdata, sizeof(TTDATA));
}
/* end of modification by Huilong */

	} /* Finished walking through this page */

/* Huilong 7/30/06 */
/*      if(has_ovflow){
           R_UNLOCK(dbenv, dbmfp->dbmp->reginfo);
           R_UNLOCK(dbenv, &dbstp->reginfo);
           goto done;
      }
/* Huilong */

#ifdef _STAMPER_DEBUG_
printf("	Page content: (key/data)\n");
	for (indx = 0;
	     indx < NUM_ENT(h);
	     indx += 1) {
	    bk = GET_BKEYDATA(h, indx);
	    if(TYPE(h) == P_LBTREE){
	    /*printf("	DBT(%d): key=%d\n", indx, *(int *)bk->data);*/
	    memcpy(&key_value, bk->data, bk->len);
	    printf("    DBT(%d): len=%d type=%d key=%d\n", indx, bk->len, B_TYPE(bk->type), key_value);
	    indx++;
	    bk = GET_BKEYDATA(h, indx);
		}
		if(B_TYPE(bk->type) == B_KEYDATA){
			pg_dbt.data = bk->data;
			pg_dbt.size = bk->len;
			memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA));
			printf("    DBT(%d): len=%d type=%d (State=%d, ", indx, bk->len, B_TYPE(bk->type), ttdata.state);

			printf("StartTime=(%ld, %ld), ", ttdata.start.curr.tv_sec,
				   ttdata.start.curr.tv_usec);
			if((ttdata.state==ID_UC)||(ttdata.state==TIME_UC)) {
			printf("StopTime=UC");
			}else
			printf("StopTime=(%ld, %ld)", ttdata.stop.curr.tv_sec,
				   ttdata.stop.curr.tv_usec);
			printf(":%s\n",(char *)(bk->data + sizeof(TTDATA)));
		}
		else{ /* duplicates or overflow */
			bo = (BOVERFLOW *)bk;
			printf("    DBT(%d): pgno=%d type=%d len=%d\n", indx, bo->pgno, B_TYPE(bo->type), bo->tlen);
		}
	}
#endif


	/* If this page is stamped, we mark it as DIRTY */
	if (changed && !F_ISSET(bhp, BH_DIRTY)) {
	    DB_ASSERT(c_mp->stat.st_page_clean != 0);
	    --c_mp->stat.st_page_clean;
	    ++c_mp->stat.st_page_dirty;
	    F_SET(bhp, BH_DIRTY);
	}

/* Huilong 7/29/06 */
/*    if (has_ovflow) {
		R_UNLOCK(dbenv, &dbstp->reginfo);
		goto done;
	}
/* Huilong */

	/* If we have just written a page to the disk, we should
	 *  Update TT according to TL.
	 */

update:
	if ((action == PAGE_WRITE)) {
	    /* Traverse the TL to update TT. For each TL entry, there
	     *  are several casess: some requires to increase the od_pccnt
	     *  in the TT, some requires to decrease it, some requires doing
	     *  nothing and the rest are impossible cases.
	     */
	    for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
		 btnp != NULL;
		 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node)) {

			/* Set up the write_state for each btn node */

			if(!IS_INVALID_TIME(btnp->tt->time) || unstamped == 0 || TYPE(h) == P_IBTREE)
				F_SET(btnp, STP_WRITE_STAMPED);

			else
				F_CLR(btnp, STP_WRITE_STAMPED);

			/* In this case we should increase the od_pgcnt */
			if (F_ISSET(btnp, STP_READ_NONE) && !F_ISSET(btnp, STP_WRITE_STAMPED)) {
				/* Do not change the od_pgcnt if it is during recovery */
				if( !F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
					STP_DEBUG( printf("\t(=======) We need to increase this entry's od_pgcnt\n") );
					btnp->tt->od_pgcnt ++;
				/* Huilong 11/24/2006 */
					MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
					/* find the dbmfp's DB */
					for (dbp = LIST_FIRST(&dbenv->dblist);
						 dbp != NULL;
						 dbp = LIST_NEXT(dbp, dblistlinks))
						if (dbp->mpf == dbmfp)
							break;
					MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
					if (dbp == NULL) {
						fprintf(stderr, "STP (__stp_page): DB not found\b");
						STP_ERROR();
					}

					/* find dbp's int32_t fileid */
					#ifdef BDB_4
					ndx = dbp->log_filename->id;
					#endif
					#ifdef BDB_3
					ndx = dbp->log_fileid;
					#endif
					/* log the "od_pgcnt" modification */
					_stp_incr_od_pgcnt_log(dbenv,
								btnp->tt->txnid, ndx,
								pgno, &lsn, 0);
				/* Huilong */

				}

				/* update the read_state and write_state --
				   set read_state to UNSTAMPED and write_state to UNSTAMEPD
				 */
				F_CLR(btnp, STP_READ_NONE);
				F_CLR(btnp, STP_WRITE_STAMPED);

				/* We don't have to log the od_pgcnt update here, because
				 *  this txn hasn't committed yet. Even if the system
				 *  crashes, we will undo this txnid, which means getting
				 *  rid of all the id of this txn in the database data
				 *  records. So we do not need the TT entry of this txn
				 *  when we next restart the system.
				 */

			}

			/* In this case we should decrease the od_pgcnt */
			else if (!F_ISSET(btnp, STP_READ_NONE) && F_ISSET(btnp, STP_WRITE_STAMPED) ){
				/* Do not change the od_pgcnt if it is during recovery */
				if( !F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
					STP_DEBUG( printf("\t(========)      We need to decrease this entry's od_pgcnt\n") );
					if (btnp->tt->od_pgcnt > 0) {
						btnp->tt->od_pgcnt --;
						MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
						/* find the dbmfp's DB */
						for (dbp = LIST_FIRST(&dbenv->dblist);
							 dbp != NULL;
							 dbp = LIST_NEXT(dbp, dblistlinks))
							if (dbp->mpf == dbmfp)
								break;
						MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
						if (dbp == NULL) {
							fprintf(stderr, "STP (__stp_page): DB not found\b");
							STP_ERROR();
						}

						/* find dbp's int32_t fileid */
			#ifdef BDB_4
						ndx = dbp->log_filename->id;
			#endif
			#ifdef BDB_3
						ndx = dbp->log_fileid;
			#endif

						/* log the "od_pgcnt" modification */
						_stp_decr_od_pgcnt_log(dbenv,
								btnp->tt->txnid, ndx,
								pgno, &lsn, 0);
					}
				}

				/* update the read_state and write_state so the tt entry can be reclaimed */
				F_SET(btnp, STP_READ_NONE);
				F_SET(btnp, STP_WRITE_STAMPED);
			}
			/* In the last case we do nothing -- read_state=UNSTAMPED and write_state=UNSTAMPED */
			else {

			}
	    }

/* Huilong 7/29/06 */

/* Huilong*/

		/* Scan through BTN nodes one more time, remove all the BTN nodes
		   with read_state=NONE and write_state=STAMPED, so the corresponding
		   TT entry can be reclaimed after the page write
		 */
		temp = NULL;
		for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
			 btnp != NULL;
			 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node))
		{
			if(temp != NULL){
				/* remove temp from the BT list and TT list */
				SH_TAILQ_REMOVE(&(bt->btn), temp, hlinks, __stp_bt_node);
				SH_TAILQ_REMOVE(&(temp->tt->btn), temp, vlinks, __stp_bt_node);
				/* insert it into the free_bt_node list */
				SH_TAILQ_INSERT_HEAD(&(stp->free_bt_node), temp, hlinks, __stp_bt_node);
				/* __tt_reclaim(dbstp, temp->tt); */
				temp = NULL;
			}
			if(F_ISSET(btnp, STP_READ_NONE) && F_ISSET(btnp, STP_WRITE_STAMPED)){
				temp = btnp;
			}
			/* also remove the btn node for a committed txn with some other states -- might be added
			   during the overflow page insertion -- needed?
			 */
			else if(!IS_INVALID_TIME(btnp->tt->time)){
				temp = btnp;
			}
		}

		if(temp != NULL){
			/* remove temp from the BT list and TT list */
			SH_TAILQ_REMOVE(&(bt->btn), temp, hlinks, __stp_bt_node);
			SH_TAILQ_REMOVE(&(temp->tt->btn), temp, vlinks, __stp_bt_node);
			/* insert it into the gree_bt_node list */
			SH_TAILQ_INSERT_HEAD(&(stp->free_bt_node), temp, hlinks, __stp_bt_node);
			__tt_reclaim(dbstp, temp->tt);
			temp = NULL;
		}

		/* remove BT entry if it has no btn node */
		if((btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node)) == NULL){
			STP_DEBUG( printf("  *** BT removed!! bt=%d\n", bt) );
			STP_BT_DEL(dbstp, bt);
		}

	}
/* Huilong 7/24/2006 */
    R_UNLOCK(dbenv, &dbstp->reginfo);
}
  done:
/* Huilong 7/24/06*/
//    R_UNLOCK(dbenv, &dbstp->reginfo);
/* Huilong */
    STP_EXP( __stp_timestop() );
    STP_DEBUG( printf("  Exit __stp_page action %d\n", action) );
    return (0);

  err:
    printf("ERROR in __stp_page: ret=%d\n", ret);
    R_UNLOCK(dbenv, &dbstp->reginfo);
    STP_EXP( __stp_timestop() );
    __db_err(dbenv, "__stp_page\n");
    exit(1);

}

/* Added by Huilong for audit recovery on 7/17/2007 */
#ifdef TBDB_AUDIT
int
__aud_tol_recover_log(dbenv, dbtp, lsnp, op, info)
    DB_ENV *dbenv;
    DBT *dbtp;
    DB_LSN *lsnp;
    db_recops op;
    void *info;
{
    DB_AUD *dbaud;
    AUD *aud;
    AUD_TOL_ENTRY *tolep, *new_tolep, *prev_tolep;
    u_int32_t   rectype;
    int         ret;
    u_int8_t    *bp;
    TV          time;
    unsigned char hash_value[AUD_HASH_LEN];

    if(op == DB_TXN_FORWARD_ROLL){
          dbaud = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->aud_handler;
          if (dbaud == NULL)
              return 0;
          aud = dbaud->reginfo.primary;
          bp = (u_int8_t *)dbtp->data;
          bp += sizeof(rectype);
          memcpy(&rectype, bp, sizeof(rectype));
          bp += sizeof(rectype);
          memcpy(&time, bp, sizeof(TV));
          bp += sizeof(TV);
          memcpy(hash_value, bp, AUD_HASH_LEN);
          bp += AUD_HASH_LEN;
          if (__aud_hash(hash_value, sizeof(AUD_TXN_HASH_DATA),
                       aud->hash_value) != 0)
            goto err;
          aud->hash_tail_time = time;
          return (0);

          err:
                printf("ERROR in _stp_recover_incr_od_pgcnt.\n");
                exit(1);
        }
    return (0);

}
#endif
