/*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

/*-
 * Transaction Time Support to BDB - Stamper Subsystem
 *
 * By     : Shilong Yao           (yao@CS.Arizona.EDU)
 *          Yuhong Liu            (yuhongl@CS.Ariozna.EDU)
 * Advisor: Dr. Richard Snodgrass (rts@CS.Arizona.EDU)
 * Computer Science., The University of Arizona
 *
 * 04/08/02: Create
 * 06/20/02: Version 2 Begin Coding
 * 12/27/02: Version 2 Finished
 * 04/09/05: version 3 Finished
 */

/***************************************************************************
 *                                                                         *
 * STP (Transaction Time Stamper) module is a lower level module embedded  *
 * in the BerkeleyDB system to timestamp the data records of the           *
 * transaction-time supported database. Before a transaction is commited,  *
 * the transaction-time is not available, we write the thansaction ID to   *
 * the record. Later when the transaction is committed, STP stamps the     *
 * record by substituting the transaction ID in the data record with the   *
 * real transaction time. This file implements all following functions     *
 * needed in the stamping process:                                         *
 *                                                                         *
 * External:                                                               *
 *    int __stp_pgread(DB_MPOOLFILE *dbmfp, void *addrp)                   *
 *    int __stp_pgwrite(DB_MPOOLFILE *dbmfp, void *addrp)                  *
 *    int __stp_mpfput(DB_MPOOLFILE *dbmfp, void *addrp)                   *
 *    int __stp_dbret(DB_MPOOLFILE *dbmfp, void *addrp, u_int32_t len,     *
 *                    void *data)                                          *
 *    //int __stp_bhfree(DB_MPOOL *dbmp, BH *bhp)                          *
 *    int __stp_scanpg(struct __dbc *dbc, PAGE *p, PAGE *lp, PAGE *rp)     *
 *    int __stp_addbtn(struct __dbc *dbc, PAGE *h)                         *
 *                                                                         *
 * Internal:                                                               *
 *    int __stp_istemporal(DB_ENV *dbenv, DB_MPOOLFILE *dbmfp)             *
 *    struct timeval __stp_get_ctime(DB_ENV *dbenv, STP_ID txnid)          *
 *    int __stp_record(DB_STP *dbstp, STP_BT_ENTRY *bt, IdOrTime *idtimep, *
 *                     u_int32_t action)                                   *
 *    int __stp_page(DB_ENV *dbenv, DB_MPOOLFILE *dbmfp, void *addrp,      *
 *                     u_int32_t action)                                   *
 *    void __tt_reclaim(DB_STP *dbstp, STP_TT_ENTRY *tt)                   *
 *    int __tt_addbtn(DB_ENV *dbenv, STP_ID txnid, db_pgno_t pgno,         *
 *                     u_int8_t fileid[DB_FILE_ID_LEN])                    *
 *                                                                         *
 ***************************************************************************/

#include "db_config.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Original BDB system */
#include "db_int.h"
#include "db_shash.h"
#include "lock.h"
#include "db_page.h"
#include "mp.h"
#include "qam.h"
#include "log.h"
#include "txn.h"

/* TT support */
#include "db.h"
#include "TTDB.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "stp_tt.h"

/* Huilong 7/14/06*/
#ifdef TBDB_AUDIT
#include "aud_int.h"
#endif
/* end 7/14/06*/

#include "temporal.h"

DB_ENV	*gdbenv;
/* During the undo phase of the recovery, we should pause the STP module.
 * This variable is defined "extern" in other STP .c files.
 */
int stp_pause = 0;
void __tt_reclaim(DB_STP *dbstp, STP_TT_ENTRY *tt);
int __stp_istemporal(DB_ENV *dbenv, DB_MPOOLFILE *dbmfp);
int __tt_addbtn(DB_ENV *dbenv, STP_ID txnid, db_pgno_t pgno, u_int8_t fileid[DB_FILE_ID_LEN]);

/***************************
 *    External Routines    *
 ***************************/

/*
 * __stp_pgread --
 *  Read a page from disk. If the page contains any transaction ID,
 *  stampe the page whenever possible.
 *	Called by __memp_pgread.
 */
int
__stp_pgread(dbmfp, addrp)
    DB_MPOOLFILE	*dbmfp;
    void		*addrp;
{
    STP_STOP_WORKING();
    if (stp_pause)
	return 0;
	STP_DEBUG( printf("  ENTER __stp_pgread\n") );
    return (__stp_page(dbmfp->dbmp->dbenv, dbmfp, addrp, PAGE_READ));
}

/*
 * __stp_pgwrite --
 *  Write a page into disk. If the page contains any transaction ID,
 * stamp the page before writing.
 *	Called by __memp_pgwrite.
 */
int
__stp_pgwrite(dbmfp, addrp)
    DB_MPOOLFILE	*dbmfp;
    void			*addrp;
{
    STP_STOP_WORKING();
    if (stp_pause)
	return 0;
	STP_DEBUG( printf("  ENTER __stp_pgwrite\n") );
    return (__stp_page(dbmfp->dbmp->dbenv, dbmfp, addrp, PAGE_WRITE));
}

/*
 * __stp_mpfput --
 *  Put an in-memory page. If the page contains any transaction ID,
 *  stamp the page.
 *	Called by __memp_put when (bh->ref == 0).
 */
int
__stp_mpfput(dbmfp, addrp)
    DB_MPOOLFILE	*dbmfp;
    void		*addrp;
{
    STP_STOP_WORKING();
    if (stp_pause){
	return 0;
	STP_DEBUG( printf("  ENTER __stp_mpfput\n") );
	//return 0;
    }
    return (__stp_page(dbmfp->dbmp->dbenv, dbmfp, addrp, MPF_PUT));
	

}


/*
 * __stp_dbret --
 *  Retrive a data record from a in-memory page. If the record contains transaction
 *  Id, stamp the page.
 *	Called by __db_dbret().
 */
int
__stp_dbret(dbmfp, addrp, len,  data)
    DB_MPOOLFILE	*dbmfp;
    void		*addrp;
    u_int32_t           len;
    void                *data; /* data.data of the pair (key, data) */
{
    TTDATA ttdata;

    STP_STOP_WORKING();
    if (stp_pause)
	return 0;
	STP_DEBUG( printf("  ENTER __stp_dbret\n") );
    memcpy (&ttdata, (char *)data, sizeof (TTDATA));
    if (len < sizeof(TTDATA))
	return 0;

    if ( (TTDB_DBT_STATE(ttdata.state) == ID_ID) ||
         (TTDB_DBT_STATE(ttdata.state) == ID_UC) ||
         (TTDB_DBT_STATE(ttdata.state) == ID_TIME) ||
         (TTDB_DBT_STATE(ttdata.state) == TIME_ID) )
	return (__stp_page(dbmfp->dbmp->dbenv, dbmfp, addrp, MPF_PUT));
    else
	return 0;
}

/*
 * __stp_scanpg --
 *	When data move from one page to another page, there is
 *	possibility that the moving records are not stamped. Because
 *	the BTN list of each TT entry keeps track of which in-memory page
 *	needs to be stamped, this data moving needs to be reflected
 *	in BTN list.
 *	This function will reset the write_state of the page to be split to STAMPED
 *  and put the new pages(left and right) that begins to contain
 *	unstamped data into the corresponding BTN list of a certain TT entry.
 *	This function is called by btree split functions __bam_root & __bam_page (btree/bt_split.c).
 *	We pass "dbc" just to get the dbenv, then DB_STP, then TT.
 */
int
__stp_scanpg(dbc, p, lp, rp)
    struct __dbc	*dbc;
    PAGE		*p;      /* original page to split */
    PAGE		*lp;     /* left page after split */
    PAGE		*rp;     /* right page after split */
{
    DB_ENV		*dbenv;
    struct __db	*dbp;
    DB_STP		*dbstp;
    STP_TT_ENTRY	*tt;
    STP_BT_ENTRY	*bt;
    STP_BTN   	*btnp;
    u_int32_t	tt_index, tt_prev, txnid_prev;
    u_int32_t	bt_index, bt_prev;
    TTDATA		ttdata;
    BH		*bhp;
    db_pgno_t	pgno;
    db_indx_t	indx;
    BKEYDATA	*bk;
    DBT		pg_dbt;
    PAGE		*h;

    STP_STOP_WORKING();
    if (stp_pause)
	return 0;

    STP_DEBUG( printf("\tENTER __stp_scanpg: pgno=%d lp=%d  rp=%d\n", PGNO(p), lp->pgno, rp->pgno) );

    dbp = dbc->dbp;
    dbenv = dbp->dbenv;
    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    /* If the database is NON-TEMPORAL, the stamper does nothing. */
    if ( (dbc->dbp->flags & DB_TEMPORAL) == 0 )
	return 0;

    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

	/* reset the write-state of btn node for page p */

	/* -- don't need to reset write_state here, page-write will do it later */
	/*
	pgno = p->pgno;
	STP_BT_FIND(dbstp, pgno, dbp->fileid, bt);
	if(bt != NULL){
		for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
		 btnp != NULL;
		 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node)) {
			F_SET(btnp, STP_WRITE_STAMPED);
		}
	}
	*/
    /* Scan through the left page if it is not the same as p, for each txnid
       encountered, create a btn node. If lp == p, then there is no need to
       do the scan and btn insertion since the necessary BT entry and btn node
       for p have already been created.
     */
    if(p->pgno != lp->pgno){
		h = lp;

		/* Get the dbp, dbenv, DB_STP, TT and BT */
		pgno = PGNO(h);

		/* Convert the page address to a buffer header. */
		bhp = (BH *)((u_int8_t *)h - SSZA(BH, buf));
		R_LOCK(dbenv, &dbstp->reginfo);

		switch (TYPE(h)) {
			case P_LDUP:		/* Off-page duplicate leaf. */
			case P_LRECNO:		/* Recno leaf. */
			case P_LBTREE:		/* Btree leaf. */
			/* Support for Btree AM
			 * Walk through all records on the page.
			 */

			txnid_prev = 0;
			for (indx = 0;
				 indx < NUM_ENT(h);
				 indx += 1) {

				if (TYPE(h)==P_LBTREE) indx++;
				if (indx >= NUM_ENT(h)) break;
		#ifdef BDB_4
				bk = GET_BKEYDATA(dbp, h, indx);
		#endif
		#ifdef BDB_3
				bk = GET_BKEYDATA(h, indx);
		#endif

				pg_dbt.data = bk->data;
				pg_dbt.size = bk->len;

				/* !! get the start/stop time from data DBT */
				memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA));

				if (pg_dbt.size < sizeof(TTDATA))
				continue;

				/* Check the start time
				 * Add (bhp+fileid) to BT of txnid if haven't done so
				 */
				if ( (TTDB_DBT_STATE(ttdata.state) == ID_UC) ||
				 (TTDB_DBT_STATE(ttdata.state) == ID_ID) ||
				 (TTDB_DBT_STATE(ttdata.state) == ID_TIME) ) {
					 if(txnid_prev == 0 || ttdata.start.stamperId.txnid != txnid_prev){
						__tt_addbtn(dbenv, ttdata.start.stamperId, pgno, dbp->fileid);
						txnid_prev = ttdata.start.stamperId.txnid;
					}
				}

				/* Check the stop time
				 * Add (bhp+fileid) to BT of txnid if haven't done so
				 */
				if ( (TTDB_DBT_STATE(ttdata.state) == ID_ID) ||
				 (TTDB_DBT_STATE(ttdata.state) == TIME_ID) ) {
					 if(txnid_prev == 0 || ttdata.stop.stamperId.txnid != txnid_prev){
						__tt_addbtn(dbenv, ttdata.stop.stamperId, pgno, dbp->fileid);
						txnid_prev = ttdata.stop.stamperId.txnid;
					}
					 /*__tt_addbtn(dbenv, ttdata.stop.stamperId, pgno, dbp->fileid); */
				}
			}
			break;
		}
	}

    /* Scan through the right page, for each txnid encountered, create a btn node */
    h = rp;

    /* Get the dbp, dbenv, DB_STP, TT and BT */
    pgno = PGNO(h);

    /* Convert the page address to a buffer header. */
    bhp = (BH *)((u_int8_t *)h - SSZA(BH, buf));
    /* R_LOCK(dbenv, &dbstp->reginfo); */

    switch (TYPE(h)) {
		case P_LDUP:		/* Off-page duplicate leaf. */
		case P_LRECNO:		/* Recno leaf. */
		case P_LBTREE:		/* Btree leaf. */
		/* Support for Btree AM
		 * Walk through all records on the page.
		 */
		txnid_prev = 0;
		for (indx = 0;
			 indx < NUM_ENT(h);
			 indx += 1) {

			if (TYPE(h)==P_LBTREE) indx++;
			if (indx >= NUM_ENT(h)) break;
	#ifdef BDB_4
			bk = GET_BKEYDATA(dbp, h, indx);
	#endif
	#ifdef BDB_3
			bk = GET_BKEYDATA(h, indx);
	#endif

			pg_dbt.data = bk->data;
			pg_dbt.size = bk->len;

			/* !! get the start/stop time from data DBT */
			memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA));

			if (pg_dbt.size < sizeof(TTDATA))
			continue;

			/* Check the start time
			 * Add (bhp+fileid) to BT of txnid if haven't done so
			 */
			if ( (TTDB_DBT_STATE(ttdata.state) == ID_UC) ||
			 (TTDB_DBT_STATE(ttdata.state) == ID_ID) ||
			 (TTDB_DBT_STATE(ttdata.state) == ID_TIME) ) {
				 if(txnid_prev == 0 || ttdata.start.stamperId.txnid != txnid_prev){
					__tt_addbtn(dbenv, ttdata.start.stamperId, pgno, dbp->fileid);
					txnid_prev = ttdata.start.stamperId.txnid;
				}
			}

			/* Check the stop time
			 * Add (bhp+fileid) to BT of txnid if haven't done so
			 */
			if ( (TTDB_DBT_STATE(ttdata.state) == ID_ID) ||
			 (TTDB_DBT_STATE(ttdata.state) == TIME_ID) ) {
				 if(txnid_prev == 0 || ttdata.stop.stamperId.txnid != txnid_prev){
					__tt_addbtn(dbenv, ttdata.stop.stamperId, pgno, dbp->fileid);
					txnid_prev = ttdata.stop.stamperId.txnid;
				}
				 /*__tt_addbtn(dbenv, ttdata.stop.stamperId, pgno, dbp->fileid); */
			}
		}
		break;
	}

    R_UNLOCK(dbenv, &dbstp->reginfo);

    STP_DEBUG( printf("\tEXIT __stp_scanpg\n") );
    STP_EXP( __stp_timestop() );

    return (0);
}

/*
 * __stp_addbtn --
 *  When a tuple is inserted into a page, we know for sure that the page must
 *  contain a txnid (the current txn is not committed yet) and needs to be stamped.
 *  Because the BTN list of each TT entry keeps track of which in-memory page
 *	needs to be stamped, this page needs to be reflected in BTN list.
 *	This function will put the new page that begins to contain
 *	unstamped data into the corresponding BTN list of a certain TT entry.
 *	This function is called by _bam_iitem & __bam_dup_convert (btree/bt_put.c).
 *	We pass "dbc" just to get the dbenv, then DB_STP, then TT.
 */
int
__stp_addbtn(dbc, h)
    struct __dbc	*dbc;
    PAGE		*h;
{
    DB_ENV		*dbenv;
    struct __db	*dbp;
    DB_STP		*dbstp;
    STP_TT_ENTRY	*tt;
    STP_BT_ENTRY	*bt;
    STP_BTN   	*btn;
    u_int32_t	tt_index, tt_prev;
    u_int32_t	bt_index, bt_prev;
    TTDATA		ttdata;
    BH		*bhp;
    db_pgno_t	pgno;
    db_indx_t	indx;
    BKEYDATA	*bk;
    DBT		pg_dbt;
    STP_ID		txnid;


    STP_STOP_WORKING();
    if (stp_pause)
	return 0;

    STP_DEBUG( printf("\tENTER __stp_addbtn: pgno=%d\n", PGNO(h)) );

    dbp = dbc->dbp;
    dbenv = dbp->dbenv;
    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    /* If the database is NON-TEMPORAL, the stamper does nothing. */
    if ( (dbc->dbp->flags & DB_TEMPORAL) == 0 )
	return 0;

    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

    /* Get the dbp, dbenv, DB_STP, TT and BT */
    pgno = PGNO(h);

    R_LOCK(dbenv, &dbstp->reginfo);

    txnid.envid = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid;
    txnid.txnid = dbc->txn->txnid;

    switch (TYPE(h)) {
    case P_LDUP:		/* Off-page duplicate leaf. */
    case P_LRECNO:		/* Recno leaf. */
    case P_LBTREE:		/* Btree leaf. */
		__tt_addbtn(dbenv, txnid, pgno, dbp->fileid);
 		goto done;
    default:
		goto done;
    }
  done:
    R_UNLOCK(dbenv, &dbstp->reginfo);

    STP_DEBUG( printf("\tEXIT __stp_addbtn\n") );
    STP_EXP( __stp_timestop() );

    return (0);
}



/***************************
 *    Internal Routines    *
 ***************************/


/* __stp_istemporal --
 *      Find out if the underlying/overlying DB is temporal or not.
 *      This is called at the beginning of __stp_page. If the DB is not
 *      temporal, the __stp_page will not do any stamping work. This is
 *      for the TUC application operating on a regular database.
 *      It is called by __memp_fget(), which is then called by: __bam_read_root(),
 *      __bam_search(), __db_new(), __bam_c_search(), __bam_c_first(), and
 *      stp_ckp_recover()
 */
int
__stp_istemporal(dbenv, dbmfp)
    DB_ENV *dbenv;
    DB_MPOOLFILE *dbmfp;
{
    struct __db *dbp;

    STP_STOP_WORKING();
    if (stp_pause)
	return 0;

    MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
STP_DEBUG( printf("  ENTER __stp_istemporal\n") );
    /* find the dbmfp's DB */
    for (dbp = LIST_FIRST(&dbenv->dblist);
	 dbp != NULL;
	 dbp = LIST_NEXT(dbp, dblistlinks))
	if (dbp->mpf == dbmfp)
	    break;

    MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);

    if (dbp == NULL)
	STP_ERROR();

    STP_DEBUG( if (!(dbp->flags & DB_TEMPORAL)) \
	       printf("\tDB is non-temporal, STP do nothing!\n") );
STP_DEBUG( printf("  Exit __stp_istemporal\n") );
    return (dbp->flags & DB_TEMPORAL);

  err:
    printf("ERROR in __stp_istemporal: DB not found\n");
    exit(1);
}



/*
 * __stp_get_ctime(dbenv, txnid) --
 *   Return the commit time for a given txn if it
 *   is committed, invalid time if not
 */
struct timeval
__stp_get_ctime(dbenv, txnid)
    DB_ENV *dbenv;
    STP_ID txnid;
{
    DB_STP *dbstp;
    STP_TT_ENTRY    *tt;
    u_int32_t       tt_prev, tt_index;
    struct timeval	rettime;

    INVALID_TIME(rettime);
#ifdef _STP_STOP_WORKING_
    return rettime;
#endif
    if (stp_pause)
	return rettime;

STP_DEBUG( printf("  ENTER __stp_get_ctime txnid=%d\n", txnid) );
    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return rettime;

    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

    R_LOCK(dbenv, &dbstp->reginfo);

    STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);

    if ( (tt != NULL) &&
	 (IS_INVALID_TIME(tt->time)) ) { /* Uncommitted */
	INVALID_TIME(rettime);
    } else if (tt != NULL)  { /* Commmitted */
	rettime = tt->time;
    } else /* Un-begun (no entry for this txn in TT) */
	STP_ERROR();

    STP_EXP( __stp_timestop() );

    R_UNLOCK(dbenv, &dbstp->reginfo);
   STP_DEBUG( printf("  Exit __stp_get_ctime txnid=%d\n", txnid) );
    return rettime;

  err:
/* The code below is changed by Huilong at 10:54am 6/23/2006 */
    printf("ERROR in __stp_get_ctime: Can't find the txnid (%d, %d) in TT.\n", txnid.envid, txnid.txnid);
/* end of change by Huilong */
    exit(1);
}

/* __stp_record --
 *	Process a record in a page and stamp it if necessary.
 *  It is called from __stp_page()
 */
int
__stp_record(dbstp, bt, idtimep, action)
    DB_STP              *dbstp;
    STP_BT_ENTRY	*bt;
    IdOrTime		*idtimep;
    u_int32_t		action;
{
    STP *stp;
    STP_TT_ENTRY	*tt;
    STP_BTN     *btnp;
    int		type;
    int		getstampped;
    u_int32_t	after, tt_index, i;
    STP_ID	txnid;
    u_int32_t	flags;
    int ret;

    STP_STOP_WORKING();
    if (stp_pause)
	return 0;
    STP_DEBUG( printf("  ENTER __stp_record action %d\n", action) );
    txnid = idtimep->stamperId;
    ret = UNCHANGED;
    stp = dbstp->reginfo.primary;

    STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);
    /* Un-begun (no entry for this txn in TT) */
    if (tt == NULL) {
/* Huilong 11/24/2006 */
/*	print_tt(dbstp);
	fprintf(stderr, "STP (__stp_record): Un-begun transaction ((%d, %d), %d)\n",
		txnid.envid.tv_sec, txnid.envid.tv_msec, txnid.txnid);
	STP_ERROR();*/
        return UNCHANGED;
/* Huilong */
    }

	btnp = NULL;
	flags = 0;

    /* Create a BTN node if not exist when action is PAGE_READ or PAGE_WRITE */
    if((action == PAGE_READ || action == PAGE_WRITE) && bt != NULL){
		/* Find BTN node in BT entry. */
		for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
			 btnp != NULL;
			 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node)) {
			if ( STPID_EQUAL(btnp->tt->txnid, txnid) )
			break;
		}

		/* If BTN not found, add one */
		if (btnp == NULL) {
			if ((btnp = SH_TAILQ_FIRST(&(stp->free_bt_node), __stp_bt_node)) != NULL)
			SH_TAILQ_REMOVE(&(stp->free_bt_node), btnp, hlinks, __stp_bt_node);
			else {
			fprintf(stderr, "STP (__stp_record): run out of BTN\n");
			STP_ERROR();
			}

			btnp->tt = tt;
			flags = 0;

			/* Set read_state to UNSTAMPED if page read, NONE if page write */
			if(action == PAGE_READ)
				LF_CLR(STP_READ_NONE);
			else
				LF_SET(STP_READ_NONE);

			/* There is no need to set up the write_state yet since page_write will set it after stamping */
			/*LF_CLR(STP_WRITE_STAMPED);*/
			btnp->flags = flags;

			/* Vertical & horizontal insert of the newly allocated BTN node */
			SH_TAILQ_INSERT_HEAD(&(bt->btn), btnp, hlinks, __stp_bt_node);
			SH_TAILQ_INSERT_HEAD(&(tt->btn), btnp, vlinks, __stp_bt_node);
		}
    }

    if (IS_INVALID_TIME(tt->time)) {
		/* Uncommitted -- nothing to do */
    } else { /* Commmitted */
		STP_DEBUG( printf("\t\tSTAMPing!!!\n") );

		/* Now we STAMP this record */
		memcpy(&idtimep->curr, &(tt->time), sizeof(TV));
		ret = CHANGED;
    }

	STP_DEBUG( printf("  Exit __stp_record action %d\n", action) );
    /* If this record got stampped, we return 1, otherwise return 0. */
    return (ret);

  err:
    printf("ERROR in __stp_record: \n");
    exit(1);
}


/*
 * __stp_page --
 *	Scan a page and give the stamps.
 *  It is called by: __stp_pgread(), __stp_pgwrite(), __stp_mpfput(), and __stp_dbret()
 *  with action as PAGE_READ, PAGE_WRITE, MPF_PUT, and MPF_PUT, accordingly.
 */
int
__stp_page(dbenv, dbmfp, addrp, action)
    DB_ENV			*dbenv;
    DB_MPOOLFILE		*dbmfp;
    void			*addrp;
    u_int32_t		action;
{
    DB_MPOOL		*dbmp;
    u_int32_t		index;
    PAGE		*h;
    void		*test;
    db_pgno_t		pgno;
    db_indx_t		indx;
    BKEYDATA		*bk;
    DBT			pg_dbt;
    char		*filename;
    TTDATA		ttdata;
    DB_STP		*dbstp;
    STP			*stp;
    STP_TT_ENTRY	*tt;
    DB_HASHTAB          *tt_htab;
    STP_BT_ENTRY	*bt;
    STP_BTN 		*btn, *btnp, *temp;
    BH			*bhp;
    u_int32_t		result, tl_file_head, tmp;
    u_int32_t           tl_index, tl_head_index, bt_prev, bt_index;
    u_int32_t		bt_head_index1, bt_tmp1, tt_prev, tt_index, tt_index1, tt_tmp1;
    u_int32_t           bt_index1, tt_tmp;
    u_int8_t 		fileid[DB_FILE_ID_LEN];
    int			unstamped, ret, i, bt_found, bt_empty, fileid_equal, changed, increment;
    MPOOL		*c_mp;
    DB_LSN		lsn;
    struct __db		*dbp;
    int32_t		ndx;
    DB_LOG		*logp;
    DB_ENTRY		*dbe;
    db_indx_t		step;
    MPOOLFILE   *mfp;

/* Huilong 7/29/06 */
    int         has_ovflow = 0;
/* Huilong */

    int key_value;
    BOVERFLOW *bo;

    STP_STOP_WORKING();
    if (stp_pause)
	return 0;

	STP_DEBUG( printf("  ENTER __stp_page action %d\n", action) );

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    /* If the database is NON-TEMPORAL, the stamper do nothing. */
    if (!__stp_istemporal(dbenv, dbmfp))
	return 0;

    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

    dbmp = dbmfp->dbmp;
    stp = (dbstp->reginfo).primary;
    logp = dbenv->lg_handle;

    /* Convert the page address to a buffer header. */
    bhp = (BH *)((u_int8_t *)addrp - SSZA(BH, buf));

    mfp = R_ADDR(dbmp->reginfo, bhp->mf_offset);

    /* Convert the buffer header to a cache. */
#ifdef BDB_4
    c_mp = dbmp->reginfo[NCACHE(dbmp->reginfo[0].primary,
				bhp->mf_offset,
				bhp->pgno)].primary;
#endif
#ifdef BDB_3
    c_mp = BH_TO_CACHE(dbmp, bhp);
#endif

/* Huilong 7/24/06 */
//    R_LOCK(dbenv, &dbstp->reginfo);
/* Huilong */
    unstamped = 0; /* This page has unstamped record(s) */
    h = (PAGE *)bhp->buf;
    pgno = bhp->pgno;
    changed = 0;

/* Huilong 7/24/06*/
//    tt_htab = dbstp->tt_htab;
/* Huilong*/

    /* Get the fileid */
#ifdef BDB_4
    memcpy(fileid, dbmfp->fileid, DB_FILE_ID_LEN);
#endif
#ifdef BDB_3
    memcpy(fileid, R_ADDR(dbmp->reginfo, dbmfp->mfp->fileid_off), DB_FILE_ID_LEN);
#endif

	bt = NULL;

    /* Walk through the data DBT on this page doing the stamp */
    switch (TYPE(h)) {
    case P_INVALID:		/* Invalid page type. */
	goto done;
    case __P_DUPLICATE:	/* Duplicate. DEPRECATED in 3.1 */
	goto done;
    case P_HASH:		/* Hash. */
	goto done;
    case P_IBTREE:		/* Btree internal. */

/* Huilong 7/24/06 */
        R_LOCK(dbenv, &dbstp->reginfo);
        tt_htab = dbstp->tt_htab;
/* Huilong */

	/* This is possible. When doing a root page split, if the
	   original root is a leaf page containing some unstamped
	   records, we should decrease those transactions' od_pgcnt
	   by one when pgwrite the page onto disk.
	*/
	if (action == PAGE_WRITE) {
	    STP_DEBUG( printf("\tIN stp_page<P_IBTREE>(action=%d) pgno = %d, type = %d, num_entry = %d\n", \
			      action, pgno, TYPE(h), NUM_ENT(h)) );

	    /* Find the page(pgno+fileid) in BT */
	    STP_BT_FIND(dbstp, pgno, fileid, bt);

	    /* If no BT entry for this page or TL is empty, this
	     *  page must not be in the situation described above.
	     *  And we do nothing about it.
	     */
	    if (bt == NULL){
/* Huilong 7/24/2006 */
                R_UNLOCK(dbenv, &dbstp->reginfo);
/* Huilong */
		goto done;
            }
	}
	goto update;

    case P_IRECNO:		/* Recno internal. */
	goto done;
    case P_LDUP:		/* Off-page duplicate leaf. */
    case P_LRECNO:		/* Recno leaf. */
    case P_LBTREE:		/* Btree leaf. */

/* Huilong 7/24/06 */
        R_LOCK(dbenv, &dbstp->reginfo);
        tt_htab = dbstp->tt_htab;
/* Huilong */

	/* If called in __memp_pgread or __memp_pgwrite, we have
	 *	to deal with BT table.
	 */
	if ((action == PAGE_READ) || (action == PAGE_WRITE)) {
	    /* Find the page(pgno+fileid) in BT */
	    STP_BT_FIND(dbstp, pgno, fileid, bt);

	    /* If there is no entry for this page in BT ...
	     * In "PAGE_READ": it's because this is the first time to process
	     *	this page. We simply add an entry for it into the BT.
	     * In "PAGE_WRITE": it's because this page is newly created in mem.
	     *	We simply add anentry for it into the BT.
	     */
	    if (bt == NULL) {
			/* Allocate a new BT entry for this page */
			STP_BT_INSERT(dbstp, pgno, fileid, bt);

			STP_DEBUG( printf("  *** new BT inserted!! bt=%d\n", bt) );
			/*STP_DEBUG( print_bt(dbstp) );*/

			if (bt == NULL) {
				fprintf(stderr, "STP (__stp_page): Run out of BT entries\b");
				STP_ERROR();
			}

			/* Initialize the new BT entry */
			bt->pgno = pgno;
			SH_TAILQ_INIT(&(bt->btn));
			memcpy(bt->fileid, fileid, DB_FILE_ID_LEN);
		}
	}

	STP_DEBUG( printf("\tIN stp_page(action=%d) pgno = %d, type = %d, num_entry = %d\n", \
			  action, pgno, TYPE(h), NUM_ENT(h)) );

	changed = 0; unstamped = 0;
	/* Walk through this page.
	 * For each key/data pair of this page, we do the stamp
	 */
	for (indx = 0;
	     indx < NUM_ENT(h);
	     indx += 1) {

	    if (TYPE(h)==P_LBTREE)
		indx++;
	    if (indx >= NUM_ENT(h))
		break;
#ifdef BDB_4
	    bk = GET_BKEYDATA(dbp, h, indx);
#endif
#ifdef BDB_3
	    bk = GET_BKEYDATA(h, indx);
#endif

	    pg_dbt.data = bk->data;
/* Modified by Huilong for overflow records on 4/18/2007 */
if(B_TYPE(bk->type) != B_OVERFLOW) 
	    pg_dbt.size = bk->len;
else{
            bo = (BOVERFLOW *)bk;
            pg_dbt.size = bo->tlen;            
}

/* end of modification */

/* Modification by Huilong for overflow records on 4/13/2007 */
	    /* Get the start/stop time from data DBT */
//	    memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA)); // original code
if (B_TYPE(bk->type) != B_OVERFLOW) {
            memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA));
}
else{
            has_ovflow = 1;
            bo = (BOVERFLOW *)bk;
            memcpy(&ttdata, (char *)(&(bo->ttdata)), sizeof(TTDATA));
}
/* end of modification by Huilong */

	    /* If the data DBT is shorter than TTDATA the
	     *   memory referencing will cause error !
	     *   So we check here. If so, it means this data
	     *   doesn't contain TTDATA at all since it is
	     *   not long enough to contain one.
	     * An on-page record pointing to the OPD page is like
	     *   this.
	     * - Shilong
	     */
	    if (pg_dbt.size < sizeof(TTDATA)) {
		STP_DEBUG( printf("\t\tDBT(%d):(State=%d, INVALID\n", indx, ttdata.state) );
		continue;
	    }
/* Huilong, 7/14/06*/
#ifdef TBDB_AUDIT
            if ((action == PAGE_WRITE) || (action == MPF_PUT)) {
                if (__aud_tuple_hash(dbenv, pg_dbt) == 1)
                    changed = 1;
                if (B_TYPE(bk->type) != B_OVERFLOW) 
                      memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA));
                else{
                      has_ovflow = 1;
                      bo = (BOVERFLOW *)bk;
                      memcpy(&ttdata, (char *)(&(bo->ttdata)), sizeof(TTDATA));                         
                }
            }
#endif
/* end 7/14/06 */
	    /* Stamp the "transaction start time" of this data record */
	    switch (TTDB_DBT_STATE(ttdata.state)) {
	    case ID_ID:
		if (__stp_record(dbstp, bt, &(ttdata.start), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, TIME_ID);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    case ID_UC:
		if (__stp_record(dbstp, bt, &(ttdata.start), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, TIME_UC);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    case ID_TIME:
		if (__stp_record(dbstp, bt, &(ttdata.start), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, TIME_TIME);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    }

	    /* Stamp the "transaction stop time" of this data record */
	    switch (TTDB_DBT_STATE(ttdata.state)) {
	    case ID_ID:
		if (__stp_record(dbstp, bt, &(ttdata.stop), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, ID_TIME);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    case TIME_ID:
		if (__stp_record(dbstp, bt, &(ttdata.stop), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, TIME_TIME);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    }

#ifdef _STAMPER_DEBUG_
	    printf("    DBT(%d):(State=%d, ", indx, ttdata.state);
	    if ((ttdata.state==ID_ID)||(ttdata.state==ID_UC)||(ttdata.state==ID_TIME)){
		printf("StartID=(%ld, %ld), ", ttdata.start.stamperId.envid,
		       ttdata.start.stamperId.txnid);
	    }else
		printf("StartTime=(%ld, %ld), ", ttdata.start.curr.tv_sec,
		       ttdata.start.curr.tv_usec);

	    if ((ttdata.state==ID_ID)||(ttdata.state==TIME_ID)){
		printf("StopID=%ld)", ttdata.stop.stamperId);
	    }else if((ttdata.state==ID_UC)||(ttdata.state==TIME_UC)) {
		printf("StopTime=UC");
	    }else
		printf("StopTime=(%ld, %ld)", ttdata.stop.curr.tv_sec,
		       ttdata.stop.curr.tv_usec);
	    printf(":%s\n",GET_BKEYDATA(h, indx-1)->data);
#endif


/* Modification by Huilong for overflow records */
	    /* Copy the stampped ttdata back to the data record. */
	    //memcpy(pg_dbt.data, &ttdata, sizeof(TTDATA));
if (B_TYPE(bk->type) != B_OVERFLOW) {
	    /* Copy the stampped ttdata back to the data record. */
	    memcpy(pg_dbt.data, &ttdata, sizeof(TTDATA));
}
else{
            bo = (BOVERFLOW *)bk;
            memcpy((char *)(&(bo->ttdata)), &ttdata, sizeof(TTDATA));
}
/* end of modification by Huilong */

	} /* Finished walking through this page */

/* Huilong 7/30/06 */
/*      if(has_ovflow){
           R_UNLOCK(dbenv, dbmfp->dbmp->reginfo);
           R_UNLOCK(dbenv, &dbstp->reginfo);
           goto done;
      }
/* Huilong */

#ifdef _STAMPER_DEBUG_
printf("	Page content: (key/data)\n");
	for (indx = 0;
	     indx < NUM_ENT(h);
	     indx += 1) {
	    bk = GET_BKEYDATA(h, indx);
	    if(TYPE(h) == P_LBTREE){
	    /*printf("	DBT(%d): key=%d\n", indx, *(int *)bk->data);*/
	    memcpy(&key_value, bk->data, bk->len);
	    printf("    DBT(%d): len=%d type=%d key=%d\n", indx, bk->len, B_TYPE(bk->type), key_value);
	    indx++;
	    bk = GET_BKEYDATA(h, indx);
		}
		if(B_TYPE(bk->type) == B_KEYDATA){
			pg_dbt.data = bk->data;
			pg_dbt.size = bk->len;
			memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA));
			printf("    DBT(%d): len=%d type=%d (State=%d, ", indx, bk->len, B_TYPE(bk->type), ttdata.state);

			printf("StartTime=(%ld, %ld), ", ttdata.start.curr.tv_sec,
				   ttdata.start.curr.tv_usec);
			if((ttdata.state==ID_UC)||(ttdata.state==TIME_UC)) {
			printf("StopTime=UC");
			}else
			printf("StopTime=(%ld, %ld)", ttdata.stop.curr.tv_sec,
				   ttdata.stop.curr.tv_usec);
			printf(":%s\n",(char *)(bk->data + sizeof(TTDATA)));
		}
		else{ /* duplicates or overflow */
			bo = (BOVERFLOW *)bk;
			printf("    DBT(%d): pgno=%d type=%d len=%d\n", indx, bo->pgno, B_TYPE(bo->type), bo->tlen);
		}
	}
#endif


	/* If this page is stamped, we mark it as DIRTY */
	if (changed && !F_ISSET(bhp, BH_DIRTY)) {
	    DB_ASSERT(c_mp->stat.st_page_clean != 0);
	    --c_mp->stat.st_page_clean;
	    ++c_mp->stat.st_page_dirty;
	    F_SET(bhp, BH_DIRTY);
	}

/* Huilong 7/29/06 */
/*    if (has_ovflow) {
		R_UNLOCK(dbenv, &dbstp->reginfo);
		goto done;
	}
/* Huilong */

	/* If we have just written a page to the disk, we should
	 *  Update TT according to TL.
	 */

update:
	if ((action == PAGE_WRITE)) {
	    /* Traverse the TL to update TT. For each TL entry, there
	     *  are several casess: some requires to increase the od_pccnt
	     *  in the TT, some requires to decrease it, some requires doing
	     *  nothing and the rest are impossible cases.
	     */
	    for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
		 btnp != NULL;
		 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node)) {

			/* Set up the write_state for each btn node */

			if(!IS_INVALID_TIME(btnp->tt->time) || unstamped == 0 || TYPE(h) == P_IBTREE)
				F_SET(btnp, STP_WRITE_STAMPED);

			else
				F_CLR(btnp, STP_WRITE_STAMPED);

			/* In this case we should increase the od_pgcnt */
			if (F_ISSET(btnp, STP_READ_NONE) && !F_ISSET(btnp, STP_WRITE_STAMPED)) {
				/* Do not change the od_pgcnt if it is during recovery */
				if( !F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
					STP_DEBUG( printf("\t(=======) We need to increase this entry's od_pgcnt\n") );
					btnp->tt->od_pgcnt ++;
                                /* Huilong 11/24/2006 */
                                        MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
                                        /* find the dbmfp's DB */
                                        for (dbp = LIST_FIRST(&dbenv->dblist);
                                                 dbp != NULL;
                                                 dbp = LIST_NEXT(dbp, dblistlinks))
                                                if (dbp->mpf == dbmfp)
                                                        break;
                                        MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
                                        if (dbp == NULL) {
                                                fprintf(stderr, "STP (__stp_page): DB not found\b");
                                                STP_ERROR();
                                        }

                                        /* find dbp's int32_t fileid */
                                        #ifdef BDB_4
                                        ndx = dbp->log_filename->id;
                                        #endif
                                        #ifdef BDB_3
                                        ndx = dbp->log_fileid;
                                        #endif
                                        /* log the "od_pgcnt" modification */
                                        _stp_incr_od_pgcnt_log(dbenv,
                                                                btnp->tt->txnid, ndx,
                                                                pgno, &lsn, 0);
                                /* Huilong */

				}

				/* update the read_state and write_state --
				   set read_state to UNSTAMPED and write_state to UNSTAMEPD
				 */
				F_CLR(btnp, STP_READ_NONE);
				F_CLR(btnp, STP_WRITE_STAMPED);

				/* We don't have to log the od_pgcnt update here, because
				 *  this txn hasn't committed yet. Even if the system
				 *  crashes, we will undo this txnid, which means getting
				 *  rid of all the id of this txn in the database data
				 *  records. So we do not need the TT entry of this txn
				 *  when we next restart the system.
				 */
			}

			/* In this case we should decrease the od_pgcnt */
			else if (!F_ISSET(btnp, STP_READ_NONE) && F_ISSET(btnp, STP_WRITE_STAMPED) ){
				/* Do not change the od_pgcnt if it is during recovery */
				if( !F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
					STP_DEBUG( printf("\t(========)      We need to decrease this entry's od_pgcnt\n") );
					if (btnp->tt->od_pgcnt > 0) {
						btnp->tt->od_pgcnt --;
						MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
						/* find the dbmfp's DB */
						for (dbp = LIST_FIRST(&dbenv->dblist);
							 dbp != NULL;
							 dbp = LIST_NEXT(dbp, dblistlinks))
							if (dbp->mpf == dbmfp)
								break;
						MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
						if (dbp == NULL) {
							fprintf(stderr, "STP (__stp_page): DB not found\b");
							STP_ERROR();
						}

						/* find dbp's int32_t fileid */
			#ifdef BDB_4
						ndx = dbp->log_filename->id;
			#endif
			#ifdef BDB_3
						ndx = dbp->log_fileid;
			#endif

						/* log the "od_pgcnt" modification */
						_stp_decr_od_pgcnt_log(dbenv,
								btnp->tt->txnid, ndx,
								pgno, &lsn, 0);
					}
				}

				/* update the read_state and write_state so the tt entry can be reclaimed */
				F_SET(btnp, STP_READ_NONE);
				F_SET(btnp, STP_WRITE_STAMPED);
			}
			/* In the last case we do nothing -- read_state=UNSTAMPED and write_state=UNSTAMPED */
			else {

			}
	    }

/* Huilong 7/29/06 */

/* Huilong*/

		/* Scan through BTN nodes one more time, remove all the BTN nodes
		   with read_state=NONE and write_state=STAMPED, so the corresponding
		   TT entry can be reclaimed after the page write
		 */
		temp = NULL;
		for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
			 btnp != NULL;
			 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node))
		{
			if(temp != NULL){
				/* remove temp from the BT list and TT list */
				SH_TAILQ_REMOVE(&(bt->btn), temp, hlinks, __stp_bt_node);
				SH_TAILQ_REMOVE(&(temp->tt->btn), temp, vlinks, __stp_bt_node);
				/* insert it into the free_bt_node list */
				SH_TAILQ_INSERT_HEAD(&(stp->free_bt_node), temp, hlinks, __stp_bt_node);
				/* __tt_reclaim(dbstp, temp->tt); */
				temp = NULL;
			}
			if(F_ISSET(btnp, STP_READ_NONE) && F_ISSET(btnp, STP_WRITE_STAMPED)){
				temp = btnp;
			}
			/* also remove the btn node for a committed txn with some other states -- might be added
			   during the overflow page insertion -- needed?
			 */
			else if(!IS_INVALID_TIME(btnp->tt->time)){
				temp = btnp;
			}
		}

		if(temp != NULL){
			/* remove temp from the BT list and TT list */
			SH_TAILQ_REMOVE(&(bt->btn), temp, hlinks, __stp_bt_node);
			SH_TAILQ_REMOVE(&(temp->tt->btn), temp, vlinks, __stp_bt_node);
			/* insert it into the gree_bt_node list */
			SH_TAILQ_INSERT_HEAD(&(stp->free_bt_node), temp, hlinks, __stp_bt_node);
			__tt_reclaim(dbstp, temp->tt);
			temp = NULL;
		}

		/* remove BT entry if it has no btn node */
		if((btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node)) == NULL){
			STP_DEBUG( printf("  *** BT removed!! bt=%d\n", bt) );
			STP_BT_DEL(dbstp, bt);
		}

	}
/* Huilong 7/24/2006 */
    R_UNLOCK(dbenv, &dbstp->reginfo);
}
  done:
/* Huilong 7/24/06*/
//    R_UNLOCK(dbenv, &dbstp->reginfo);
/* Huilong */
    STP_EXP( __stp_timestop() );
    STP_DEBUG( printf("  Exit __stp_page action %d\n", action) );
    return (0);

  err:
    printf("ERROR in __stp_page: ret=%d\n", ret);
    R_UNLOCK(dbenv, &dbstp->reginfo);
    STP_EXP( __stp_timestop() );
    __db_err(dbenv, "__stp_page\n");
    exit(1);

}

/* __tt_addbtn --
 * 	Add a page (pgno+fileid), containing txnid's unstamped records, into txnid's
 * 	TT entry's BT, if not already there.
 *  It is called by __stp_addbt() (in 'stp_stamper.c')
 */
int
__tt_addbtn(dbenv, txnid, pgno, fileid)
    DB_ENV          *dbenv;
    STP_ID		txnid;
    db_pgno_t	pgno;
    u_int8_t	fileid[DB_FILE_ID_LEN];
{
    DB_STP          *dbstp;
    STP_TT_ENTRY	*tt;
    STP_BT_ENTRY	*bt;
    u_int32_t	tt_index, tt_prev;
    u_int32_t	bt_head, bt_index, bt_prev;
    STP *stp;
    u_int32_t btn_prev, btn_index;
    STP_BTN *btnp;

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;


    STP_DEBUG(printf("\tENTER tt_addbtn, pgno=%d, tt=%d, txnid=(%d,%d)\n",
	   pgno, tt, txnid.envid, txnid.txnid));

    /* Find the txnid in the TT */
    STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);

    /* If the txnid is not in TT, there must be something wrong */
    if ( tt == NULL )
	STP_ERROR();

    /* If the txnid is not committed yet, we do nothing until it's committed */

    /* Find the (pgno+fileid) in the BT */
    STP_BT_FIND(dbstp, pgno, fileid, bt);

    /* By reaching here, the txn is committed.
     * If the page is not in BT, add it to BT.
     * If (pgno, fileid) is not in BT, add it.
     */
    if (bt == NULL)
	STP_BT_INSERT(dbstp, pgno, fileid, bt);

    if (bt == NULL)
	STP_ERROR();

    /* Add the page (pgno+fileid) into the 2-D list */


    stp = dbstp->reginfo.primary;

    /* See if already there, otherwise find the point to insert */
    for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
	 btnp != NULL;
	 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node)) {
	if (btnp->tt == tt)
	    break;
    }

    /* If the BTN is already there, we are done. */
    if (btnp != NULL){
		goto done;
	}

    /* Allocate a new BTN */
    if ((btnp = SH_TAILQ_FIRST(&(stp->free_bt_node), __stp_bt_node)) != NULL)
	SH_TAILQ_REMOVE(&(stp->free_bt_node), btnp, hlinks, __stp_bt_node);
    else
	STP_ERROR();

    /* Init */
	F_SET(btnp, STP_READ_NONE);
	F_CLR(btnp, STP_WRITE_STAMPED);

    /* Vertical insert */
    btnp->tt = tt;
    SH_TAILQ_INSERT_HEAD(&(btnp->tt->btn), btnp, vlinks, __stp_bt_node);

    /* Horizontal insert */
    SH_TAILQ_INSERT_HEAD(&(bt->btn), btnp, hlinks, __stp_bt_node);

    /*STP_DEBUG(print_tt(dbstp);printf("\tEXIT tt_addbtn, pgno=%d\n", pgno)); */

  done:
    return (0);
  err:
    printf(" ** ERROR in tt_addbtn, pgno=%d\n", pgno);
    exit(1);
}

void __stp_TT_reclaim(DB_ENV *dbenv){
    DB_STP	*dbstp;
    STP_BTN  	*btn;
    u_int32_t tt_index;
    STP_TT_ENTRY *tt;

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    if (dbstp == NULL)
	return;
STP_DEBUG( printf("========= ENTER __stp_TT_reclaim\n"));
    for (tt_index=0; tt_index < dbstp->tt_htab_size; tt_index++) {
	for (tt = SH_TAILQ_FIRST(&(dbstp->tt_htab[tt_index]), __stp_tt_entry);
	     tt != NULL;
	     tt = SH_TAILQ_NEXT(tt, links, __stp_tt_entry)) {
         __tt_reclaim(dbstp, tt);
	}
    }

}

/* __tt_reclaim --
 *	If this TT entry's od_pgcnt is 0 and also its BTN list
 *	is empty, we reclaim this entry.
 *  it is called by __stp_page with PAGE_WRITE action.
 */
void
__tt_reclaim(dbstp, tt)
    DB_STP *dbstp;
    STP_TT_ENTRY *tt;
{
    STP *stp;
    u_int32_t	tt_idx;
    STP_BTN *btnp;


   STP_DEBUG( printf("\tENTER __tt_reclaim\n"));

    stp = dbstp->reginfo.primary;

    /* Find the btn head of this TT entry */
    btnp = SH_TAILQ_FIRST(&(tt->btn), __stp_bt_node);

    /* Check if this entry needs to be claimed */
    if ( (tt->od_pgcnt > 0) || (btnp != NULL) )
	return;

    /* If this entry hasn't committed, we shouldn't reclaim it */
    if (IS_INVALID_TIME(tt->time))
	return;

    /* At this point, the TT entry has at no BTN node. Reclaim the TT entry. */
    tt_idx = STP_TT_HASH(dbstp, tt->txnid);
#ifdef STAMPER_DEBUG
    printf("\t\tTT entry in bucket #%d txnid=(%d,%d) is deleted!\n",
	   tt_idx, tt->txnid.envid, tt->txnid.txnid);
#endif
    SH_TAILQ_REMOVE(&(dbstp->tt_htab[tt_idx]), tt, links, __stp_tt_entry);
    SH_TAILQ_INSERT_HEAD(&(stp->free_tt_ent), tt, links, __stp_tt_entry);
STP_DEBUG( printf("\tExit __tt_reclaim\n"));
    return;
}


/*
 * __stp_overflow --
 *	Scan the overflow records and stamp them
 *  It is called by: __stp_pgread(), __stp_pgwrite(), __stp_mpfput(), and __stp_dbret()
 *  with action as PAGE_READ, PAGE_WRITE, MPF_PUT, and MPF_PUT, accordingly.
 *  By Huilong 7/28/06
 */
/*int
__stp_overflow(dbenv, dbmfp, action)
    DB_ENV			*dbenv;
    DB_MPOOLFILE		*dbmfp;
    u_int32_t       action;
{
    DB_MPOOL		*dbmp;
    u_int32_t		index;
    PAGE		*h;
    void		*test;
    db_pgno_t		pgno;
    db_indx_t		indx;
    BKEYDATA		*bk;
    DBT			pg_dbt;
    char		*filename;
    TTDATA		ttdata;
    DB_STP		*dbstp;
    STP			*stp;
    STP_TT_ENTRY	*tt;
    DB_HASHTAB          *tt_htab;
    STP_BT_ENTRY	*bt;
    STP_BTN 		*btn, *btnp, *temp;
    BH			*bhp;
    u_int32_t		result, tl_file_head, tmp;
    u_int32_t           tl_index, tl_head_index, bt_prev, bt_index;
    u_int32_t		bt_head_index1, bt_tmp1, tt_prev, tt_index, tt_index1, tt_tmp1;
    u_int32_t           bt_index1, tt_tmp;
    u_int8_t 		fileid[DB_FILE_ID_LEN];
    int			unstamped, ret, i, bt_found, bt_empty, fileid_equal, changed, increment;
    MPOOL		*c_mp;
    DB_LSN		lsn;
    struct __db		*dbp;
    int32_t		ndx;
    DB_LOG		*logp;
    DB_ENTRY		*dbe;
    db_indx_t		step;
    MPOOLFILE   *mfp;

    int key_value;
//    BOVERFLOW *bo;

    int  ov_rec_start=-1, ov_rec_stop=-1;
    char *buf;
    int  len, num_ov_rec;

    STP_STOP_WORKING();
    if (stp_pause)
	return 0;

	STP_DEBUG( printf("  ENTER __stp_page action %d\n", action) );

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
/*    if (dbstp == NULL)
	return 0;

    /* If the database is NON-TEMPORAL, the stamper do nothing. */
/*    if (!__stp_istemporal(dbenv, dbmfp))
	return 0;

    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

    dbmp = dbmfp->dbmp;
    stp = (dbstp->reginfo).primary;
    logp = dbenv->lg_handle;

	bt = NULL;

    if(stp->ov_count==0) return 0;

    R_LOCK(dbenv, &dbstp->reginfo);

    tt_htab = dbstp->tt_htab;

#ifdef BDB_4
    memcpy(fileid, dbmfp->fileid, DB_FILE_ID_LEN);
#endif
#ifdef BDB_3
    memcpy(fileid, R_ADDR(dbmp->reginfo, dbmfp->mfp->fileid_off), DB_FILE_ID_LEN);
#endif

/* check for the start, stop point in OV table, indentify the rec_pgno */
/*    pgno = -1;
    for(i=0; i<stp->ov_count; i++){
	    if(dbstp->ov_table[i].dbmfp == dbmfp && ov_rec_start == -1){
	       ov_rec_start = i;
	       pgno = dbstp->ov_table[i].rec_pgno;
	    }
	    if(ov_rec_start>=0 &&
	       (dbstp->ov_table[i].dbmfp != dbmfp ||
	       pgno != dbstp->ov_table[i].rec_pgno)){
		   ov_rec_stop = i;
		   break;
		}
	}
	if(i==stp->ov_count)
	   ov_rec_stop = stp->ov_count;

	if(ov_rec_start==-1 || ov_rec_stop==-1){
        R_UNLOCK(dbenv, &dbstp->reginfo);
        return 0;
	}

/* */

	/* Find the page(pgno+fileid) in BT */
/*	STP_BT_FIND(dbstp, pgno, fileid, bt);
	if (bt == NULL) {
			/* Allocate a new BT entry for this page */
/*			STP_BT_INSERT(dbstp, pgno, fileid, bt);
			if (bt == NULL) {
				fprintf(stderr, "STP (__stp_page): Run out of BT entries\b");
				STP_ERROR();
			}
			/* Initialize the new BT entry */
/*			bt->pgno = pgno;
			SH_TAILQ_INIT(&(bt->btn));
			memcpy(bt->fileid, fileid, DB_FILE_ID_LEN);
	}

	/* Walk through the overflow records
	 * For each records, we do the stamp
	 */
/*	for (indx = ov_rec_start; indx < ov_rec_stop; indx++) {
        /* find the dbmfp's DB */
/*        for(dbp = LIST_FIRST(&dbenv->dblist);
            dbp != NULL;
            dbp = LIST_NEXT(dbp, dblistlinks))
                if (dbp->mpf == dbmfp)  break;
        if(dbp->mpf!=dbmfp) goto err;
        F_SET(&pg_dbt, DB_DBT_USERMEM | DB_DBT_PARTIAL);
        pg_dbt.doff = 0;
        pg_dbt.data = (char *)&ttdata;
        pg_dbt.dlen = sizeof(TTDATA);
        pg_dbt.ulen = sizeof(TTDATA);
        /*if(__db_goff(dbp, &pg_dbt, sizeof(TTDATA),
                     dbstp->ov_table[indx].ov_pgno, &buf, &len)
           !=0){
            goto err;
        }*/
/*        memcpy (&ttdata, (char *)pg_dbt.data, sizeof (TTDATA));

	    /* Stamp the "transaction start time" of this data record */
/*	    switch (TTDB_DBT_STATE(ttdata.state)) {
	    case ID_ID:
		if (__stp_record(dbstp, bt, &(ttdata.start), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, TIME_ID);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    case ID_UC:
		if (__stp_record(dbstp, bt, &(ttdata.start), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, TIME_UC);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    case ID_TIME:
		if (__stp_record(dbstp, bt, &(ttdata.start), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, TIME_TIME);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    }

	    /* Stamp the "transaction stop time" of this data record */
/*	    switch (TTDB_DBT_STATE(ttdata.state)) {
	    case ID_ID:
		if (__stp_record(dbstp, bt, &(ttdata.stop), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, ID_TIME);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    case TIME_ID:
		if (__stp_record(dbstp, bt, &(ttdata.stop), action) == CHANGED) {
		    TTDB_DBT_SETSTATE(ttdata.state, TIME_TIME);
		    changed = 1;
		} else
		    unstamped = 1;
		break;
	    }
        
        if ((ret = memp_fget(dbmfp, &(dbstp->ov_table[indx].ov_pgno), 0, &h)) != 0) {
              goto err;
        }
        memcpy((u_int8_t *)h + P_OVERHEAD, (u_int8_t *)&ttdata, sizeof(ttdata));
        (void)memp_fput(dbp->mpf, h, DB_MPOOL_DIRTY);
	} /* Finished walking through this page */

	/* Remove the records from ov table
	 */
/*	num_ov_rec = ov_rec_stop - ov_rec_start;
	for(indx = ov_rec_stop; indx<stp->ov_count; indx++) {
        dbstp->ov_table[indx-num_ov_rec] =  dbstp->ov_table[indx];
    }
    stp->ov_count -= num_ov_rec;

	/* If we have just written a page to the disk, we should
	 *  Update TT according to TL.
	 */
/*update:
	if ((action == PAGE_WRITE)) {
	    /* Traverse the TL to update TT. For each TL entry, there
	     *  are several casess: some requires to increase the od_pccnt
	     *  in the TT, some requires to decrease it, some requires doing
	     *  nothing and the rest are impossible cases.
	     */
/*	    for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
		 btnp != NULL;
		 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node)) {

			/* Set up the write_state for each btn node */
/*			if(!IS_INVALID_TIME(btnp->tt->time) || unstamped == 0)
				F_SET(btnp, STP_WRITE_STAMPED);

			else
				F_CLR(btnp, STP_WRITE_STAMPED);


			/* In this case we should increase the od_pgcnt */
/*			if (F_ISSET(btnp, STP_READ_NONE) && !F_ISSET(btnp, STP_WRITE_STAMPED)) {
				/* Do not change the od_pgcnt if it is during recovery */
/*				if( !F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
					STP_DEBUG( printf("\t(=======) We need to increase this entry's od_pgcnt\n") );
					btnp->tt->od_pgcnt ++;
                                /* Huilong 11/24/2006 */
  /*                                      MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);
                                        /* find the dbmfp's DB */
/*                                        for (dbp = LIST_FIRST(&dbenv->dblist);
                                                 dbp != NULL;
                                                 dbp = LIST_NEXT(dbp, dblistlinks))
                                                if (dbp->mpf == dbmfp)
                                                        break;
                                        MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
                                        if (dbp == NULL) {
                                                fprintf(stderr, "STP (__stp_page): DB not found\b");
                                                STP_ERROR();
                                        }

                                        /* find dbp's int32_t fileid */
/*                                        #ifdef BDB_4
                                        ndx = dbp->log_filename->id;
                                        #endif
                                        #ifdef BDB_3
                                        ndx = dbp->log_fileid;
                                        #endif
                                        /* log the "od_pgcnt" modification */
/*                                        _stp_incr_od_pgcnt_log(dbenv,
                                                                btnp->tt->txnid, ndx,
                                                                pgno, &lsn, 0);
                                /* Huilong */
/*				}

				/* update the read_state and write_state --
				   set read_state to UNSTAMPED and write_state to UNSTAMEPD
				 */
/*				F_CLR(btnp, STP_READ_NONE);
				F_CLR(btnp, STP_WRITE_STAMPED);

			}

			/* In this case we should decrease the od_pgcnt */
/*			else if (!F_ISSET(btnp, STP_READ_NONE) && F_ISSET(btnp, STP_WRITE_STAMPED) ){
				/* Do not change the od_pgcnt if it is during recovery */
/*				if( !F_ISSET((DB_LOG *)dbenv->lg_handle, DBLOG_RECOVER) ){
					STP_DEBUG( printf("\t(========)      We need to decrease this entry's od_pgcnt\n") );
					if (btnp->tt->od_pgcnt > 0) {
						btnp->tt->od_pgcnt --;

						MUTEX_THREAD_LOCK(dbenv, dbenv->dblist_mutexp);

						/* find the dbmfp's DB */
/*						for (dbp = LIST_FIRST(&dbenv->dblist);
							 dbp != NULL;
							 dbp = LIST_NEXT(dbp, dblistlinks))
							if (dbp->mpf == dbmfp)
								break;
						MUTEX_THREAD_UNLOCK(dbenv, dbenv->dblist_mutexp);
						if (dbp == NULL) {
							fprintf(stderr, "STP (__stp_page): DB not found\b");
							STP_ERROR();
						}
						/* find dbp's int32_t fileid */
/*			#ifdef BDB_4
						ndx = dbp->log_filename->id;
			#endif
			#ifdef BDB_3
						ndx = dbp->log_fileid;
			#endif
						/* log the "od_pgcnt" modification */
/*						_stp_decr_od_pgcnt_log(dbenv,
								btnp->tt->txnid, ndx,
								pgno, &lsn, 0);
					}
				}

				/* update the read_state and write_state so the tt entry can be reclaimed */
/*				F_SET(btnp, STP_READ_NONE);
				F_SET(btnp, STP_WRITE_STAMPED);
			}
			/* In the last case we do nothing -- read_state=UNSTAMPED and write_state=UNSTAMPED */
/*			else {

			}
	 	}

     /* Scan through BTN nodes one more time, remove all the BTN nodes
		   with read_state=NONE and write_state=STAMPED, so the corresponding
		   TT entry can be reclaimed after the page write
	 */
/*	    temp = NULL;
	 	for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
			 btnp != NULL;
			 btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node))
	 	{
			if(temp != NULL){
				/* remove temp from the BT list and TT list */
/*				SH_TAILQ_REMOVE(&(bt->btn), temp, hlinks, __stp_bt_node);
				SH_TAILQ_REMOVE(&(temp->tt->btn), temp, vlinks, __stp_bt_node);
				/* insert it into the free_bt_node list */
/*				SH_TAILQ_INSERT_HEAD(&(stp->free_bt_node), temp, hlinks, __stp_bt_node);
				/* __tt_reclaim(dbstp, temp->tt); */
/*				temp = NULL;
			}
			if(F_ISSET(btnp, STP_READ_NONE) && F_ISSET(btnp, STP_WRITE_STAMPED)){
				temp = btnp;
			}
			/* also remove the btn node for a committed txn with some other states -- might be added
			   during the overflow page insertion -- needed?
			 */
/*			else if(!IS_INVALID_TIME(btnp->tt->time)){
				temp = btnp;
			}
		 }

		 if(temp != NULL){
			/* remove temp from the BT list and TT list */
/*			SH_TAILQ_REMOVE(&(bt->btn), temp, hlinks, __stp_bt_node);
			SH_TAILQ_REMOVE(&(temp->tt->btn), temp, vlinks, __stp_bt_node);
			/* insert it into the gree_bt_node list */
/*			SH_TAILQ_INSERT_HEAD(&(stp->free_bt_node), temp, hlinks, __stp_bt_node);
			__tt_reclaim(dbstp, temp->tt);
			temp = NULL;
		 }

		/* remove BT entry if it has no btn node */
/*		if((btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node)) == NULL){
			STP_DEBUG( printf("  *** BT removed!! bt=%d\n", bt) );
			STP_BT_DEL(dbstp, bt);
		}
	}
    R_UNLOCK(dbenv, &dbstp->reginfo);

  done:
    STP_EXP( __stp_timestop() );
    STP_DEBUG( printf("  Exit __stp_page action %d\n", action) );
    return (0);

  err:
    printf("ERROR in __stp_page: ret=%d\n", ret);
    R_UNLOCK(dbenv, &dbstp->reginfo);
    STP_EXP( __stp_timestop() );
    __db_err(dbenv, "__stp_page\n");
    exit(1);
}*/
