/*-
 * Developed (2002-2003) by:
 *     TAU Project
 *     Computer Science Dept., University of Arizona
 *     Tucson, AZ 85721, USA
 *     URL  : http://www.cs.arizona.edu/tau
 *     Email: tau@cs.arizona.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */
/*-
 * Transaction Time Support to BDB - Stamper Subsystem
 *
 * By     : Shilong Yao           (yao@CS.Arizona.EDU)
 *          Yuhong Liu            (yuhongl@CS.Ariozna.EDU)
 * Advisor: Dr. Richard Snodgrass (rts@CS.Arizona.EDU)
 * Computer Science Dept., The University of Arizona
 *
 * 04/08/02: Create
 * 06/20/02: Version 2 Begin Coding
 * 12/27/02: Version 2 Finished
 * 04/09/05: version 3 Finished
 */

 /***************************************************************************
  *                                                                         *
  * STP (Transaction Time Stamper) module is a lower level module embedded  *
  * in the BerkeleyDB system to timestamp the data records of the           *
  * transaction-time supported database. This file implements all the       *
  * supported functions needed in STP for BerkeleyDB transactions:          *
  *    int __stp_txn_begin(DB_ENV *dbenv, u_int32_t txnid_arg)              *
  *    int __stp_txn_end(DB_ENV *dbenv, u_int32_t txnid_arg)                *
  *    int __stp_txn_commit(DB_ENV *dbenv, u_int32_t txnid_arg, TV *timep)  *
  *    int __stp_txn_abort(DB_ENV *dbenv, u_int32_t txnid_arg)              *
  *                                                                         *
  ***************************************************************************/

#include "db_config.h"

#include <sys/types.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>

/* Original BDB system */
#include "db_int.h"
#include "db_shash.h"
#include "lock.h"
#include "db_page.h"
#include "mp.h"
#include "qam.h"
#include "log.h"

/* TT support */
#include "db.h"
#include "TTDB.h"
#include "clk_txn.h"
#include "stp_int.h"
#include "stp_tt.h"

/* Huilong 7/14/06 */
#ifdef TBDB_AUDIT
#include "aud_int.h"
#endif
/* end 7/14/06 */

#include "temporal.h"

extern DB_ENV *gdbenv;
extern int stp_pause;

/*
 * A summary of the sequence during the txn's life time.
 * (1) Transaction begin:
 *   __txn_begin()
 *   |
 *   +--> Original BDB stuff
 *   +--> __stp_txn_begin()
 *
 * (2) Transaction commit;
 *   __txn_commit()
 *   |
 *   +--> clk_txn_commit()
 *   |    |
 *   |    +--> __stp_txn_commit()
 *   |    |    |
 *   |    |    +--> log the insertion of the committing txn into TT
 *   |    |    +--> insert the committing txn into TT
 *   |    +--> write commit record into log
 *   |    +--> flush log if necessary
 *   +--> __txn_end()
 *   |    |
 *   |    +--> __stp_txn_end()
 *   |    |    |
 *   |    |    +--> stamp all in-memory pages that contains the committing txn's id
 *   |    |    +--> reclaim the TT entry of the committing txn if possible
 *   |    +--> release all locks of the committing txn
 *   |
 *
 * (3) Transaction abort:
 *   (Since CLK doesn't handle abort, STP's abort routine will just wait for
 *    calls from CLK module.  When CLK adds the abort, STP will automatically support it.)
 *   __txn_abort()
 *   |
 *   +--> __txn_undo()
 *   +--> __txn_end() (similar as in 2)
 *
 * (4) Checkpoint:
 *   txn_checkpoint()
 *   |
 *   +--> memp_sync()
 *   +--> STP log the TT table
 *   +--> write the checkpoint log record & __log_flush()
 */
/* Modification by Huilong at 4:51pm 6/21/2006:
 * Below code was added to fix the bug
 */
/* __btn_insert --
 *      Given the BT, TT entry and a page (pgno+fileid), if the page
 *      is not already in the 2-D list, add it, otherwise do thing.
 */
int
__btn_insert(dbstp, bt, tt, pgno, fileid)
    DB_STP *dbstp;
    STP_BT_ENTRY *bt;
    STP_TT_ENTRY *tt;
    db_pgno_t pgno;
    u_int8_t fileid[DB_FILE_ID_LEN];
{
    STP *stp;
    u_int32_t btn_prev, btn_index;
    STP_BTN *btnp;

#ifdef STAMPER_DEBUG
    printf("\tENTER __btn_insert\n");
#endif
    stp = dbstp->reginfo.primary;

    /* See if already there, otherwise find the point to insert */
    for (btnp = SH_TAILQ_FIRST(&(bt->btn), __stp_bt_node);
         btnp != NULL;
         btnp = SH_TAILQ_NEXT(btnp, hlinks, __stp_bt_node)) {
        if (btnp->tt == tt)
            break;
    }

   /* If the BTN is already there, we are done. */
    if (btnp != NULL)
        goto done;

    /* Allocate a new BTN */
    if ((btnp = SH_TAILQ_FIRST(&(stp->free_bt_node), __stp_bt_node)) != NULL)
        SH_TAILQ_REMOVE(&(stp->free_bt_node), btnp, hlinks, __stp_bt_node);
    else
        STP_ERROR();

    /* Init */
    btnp->flags = 0;
    btnp->flags |= (STP_READ_NONE);

    /* Vertical insert */
    btnp->tt = tt;
    SH_TAILQ_INSERT_HEAD(&(btnp->tt->btn), btnp, vlinks, __stp_bt_node);

#ifdef STAMPER_DEBUG
    btnp->pgno = pgno;
#endif

    /* Horizontal insert */
    SH_TAILQ_INSERT_HEAD(&(bt->btn), btnp, hlinks, __stp_bt_node);

  done:

#ifdef STAMPER_DEBUG
    printf("\tEXIT __btn_insert\n");
#endif
    return (0);
  err:
    printf("ERROR in __btn_insert:\n");
    exit(1);
}

/* end of modification by Huilong
 */

/*
 * __stp_txn_begin --
 *	Add an entry in the TT table.
 */
int
__stp_txn_begin(dbenv, txnid_arg)
    DB_ENV		*dbenv;
    u_int32_t	txnid_arg;
{
    STP *stp;
    u_int32_t	btn_head_index, tt_prev, tt_index, tt_new;
    STP_TT_ENTRY	*tt;
    STP_BTN	        *btn;
    DB_STP		*dbstp;
    STP_ID		txnid;

    STP_STOP_WORKING();

    txnid.envid = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid;
    txnid.txnid = txnid_arg;

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    STP_DEBUG( printf("\tENTER __stp_txn_begin, dbstp=0x%x, (ID=%d,%d)\n", \
	   dbstp, txnid.envid, txnid.txnid);fflush(stdout) );

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    stp = dbstp->reginfo.primary;

    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

    R_LOCK(dbenv, &dbstp->reginfo);

    /* Locate txnid in TT */
    STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);

    /* If txnid is already in TT, error */
    if (tt != NULL)
	STP_ERROR();

    /* Insert the txnid into TT */
    if ((tt = SH_TAILQ_FIRST(&(stp->free_tt_ent), __stp_tt_entry)) != NULL)
	SH_TAILQ_REMOVE(&(stp->free_tt_ent), tt, links, __stp_tt_entry);
    else
	STP_ERROR();
    tt->txnid = txnid;
    INVALID_TIME(tt->time);
    SH_TAILQ_INIT(&(tt->btn));
    tt->od_pgcnt = 0;
    tt_index = STP_TT_HASH(dbstp, txnid);
    SH_TAILQ_INSERT_HEAD(&(dbstp->tt_htab[tt_index]), tt, links, __stp_tt_entry);

    STP_DEBUG( print_tt(dbstp) );
    STP_DEBUG( printf("\tEXIT __stp_txn_begin: dbenv=0x%x\n", dbenv); fflush(stdout) );

    R_UNLOCK(dbenv, &dbstp->reginfo);

    STP_EXP( __stp_timestop() );

/* Huilong 7/14/06 */
#ifdef TBDB_AUDIT
    __aud_txn_begin(dbenv, txnid_arg);
#endif
/* end 7/14/06 */
    return (0);

  err:	__db_err(dbenv, "__stp_txn_begin\n");
    R_UNLOCK(dbenv, &dbstp->reginfo);
    printf("ERROR in stp_txn_begin\n");fflush(stdout);
    exit(1);
    return (STP_ERROR_TXN_BEGIN);
}

/*
 * __stp_txn_end --
 *	At committing, before the locker's locks are all released or inherited,
 *	we need to count the number of pages with WRITE lock and put this number
 *	in the in_pgcnt of the TT entry.
 *
 *      When a txn commits, __stp_txn_end() is called after __stp_txn_commit().
 */
int
__stp_txn_end(dbenv, txnid_arg)
    DB_ENV		*dbenv;
    u_int32_t	txnid_arg;
{
    DB_STP		*dbstp;
    int 		ret, i;
    STP_TT_ENTRY	*tt;
    STP_BT_ENTRY	*bt;
    STP_BTN	        *btn;
    DB_LOCKTAB 	*lt;
    DB_LOCKREGION 	*region;
    u_int32_t	tt_index, tt_prev, btn_head_index, bt_prev, bt_index;
    u_int32_t	bt_tmp, bt_result, btn_index;
    u_int32_t 	ndx;
    DB_LOCKER 	*sh_locker, *sh_parent;
    DB_LOCKOBJ 	*obj, *sh_obj;
    db_pgno_t	pgno;
    u_int8_t 	fileid[DB_FILE_ID_LEN];
    u_int32_t	locker;
    PAGE		*h;
    struct __db	*dbp;
    struct __db_lock *lp, *next_lock;
    BH		*bhp;
    DB_MPOOL	*dbmp;
    DB_HASHTAB	*dbht;
    MPOOL		*c_mp, *mp;
    MPOOLFILE	*mfp;
    DB_MPOOLFILE	*dbmfp;
    int		record_it, first;
    size_t		n_bucket, n_cache, mf_offset;
    STP_ID		txnid;
	STP			*stp;
	STP_BTN  	*btnp;

    STP_STOP_WORKING();

    STP_DEBUG( printf("\tENTER __stp_txn_end (ID=%ld)\n", txnid_arg); fflush(stdout) );

    txnid.envid = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid;
    txnid.txnid = txnid_arg;

    PANIC_CHECK(dbenv);

#ifdef BDB_4
    ENV_REQUIRES_CONFIG(dbenv,
			dbenv->lk_handle,
			"DB_ENV->lock_vec",
			DB_INIT_LOCK);
#endif
#ifdef BDB_3
    ENV_REQUIRES_CONFIG(dbenv,
			dbenv->lk_handle,
			DB_INIT_LOCK);
#endif

    locker = txnid_arg;
    lt = dbenv->lk_handle;
    region = lt->reginfo.primary;
    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;


    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

    LOCKREGION(dbenv, (DB_LOCKTAB *)dbenv->lk_handle);
    R_LOCK(dbenv, &dbstp->reginfo);

    /* Find the txn's entry in TT */
    STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);

    /* No matter a txn is committed or aborted, this "end" function
     * will be called.
     * Therefore this txn may have been aborted. So if it is not
     * 	in TT, we simply return without doing anyting :)
     */
    if (tt == NULL)
		goto out1;
	else if (IS_INVALID_TIME(tt->time)){
		/* If txnid's BTN list is not empty, there must be something wrong*/
		btnp = SH_TAILQ_FIRST(&(tt->btn), __stp_bt_node);
		if (btnp != NULL)
		STP_ERROR();
		tt_index = STP_TT_HASH(dbstp, txnid);
		SH_TAILQ_REMOVE(&(dbstp->tt_htab[tt_index]), tt, links, __stp_tt_entry);
		stp = dbstp->reginfo.primary;
		SH_TAILQ_INSERT_HEAD(&(stp->free_tt_ent), tt, links, __stp_tt_entry);
		goto out1;

	}
    /* Find the BTN head node of the TT entry */
    LOCKER_LOCK(lt, region, locker, ndx);
    if ((ret = __lock_getlocker(lt, locker, ndx, 0, &sh_locker)) != 0 ||
	sh_locker == NULL || F_ISSET(sh_locker, DB_LOCKER_DELETED)) {
	if (ret == 0 && sh_locker == NULL)
	    goto out;
	__db_err(dbenv, "Locker error.");
    }

/* Added by Huilong on 5/23/07 for dead lock problem*/
    R_UNLOCK(dbenv, &dbstp->reginfo);
/* End of modification */

    /* Walk through all the locks held by the commiting txn.
     * 	Record all pages WRITE locked by the txn in its BT.
     * 	After the traverse, we add all the recorded pages
     * 	into the im_pgcnt of the txn.
     */
    for (lp = SH_LIST_FIRST(&sh_locker->heldby, __db_lock);
	 lp != NULL;
	 lp = SH_LIST_NEXT(lp, locker_links, __db_lock)) {

	if ((!IS_WRITELOCK(lp->mode))
#ifdef BDB_4
	    && (lp->mode != DB_LOCK_WWRITE)
#endif
	    )
	    continue;

	/* Get the object associated with this lock. */
	obj = (DB_LOCKOBJ *)((u_int8_t *)lp + lp->obj);

	/* Get the (fileid, pgno) of the locked page */
	pgno = ((DB_LOCK_ILOCK *)(obj->objdata))->pgno;
	memcpy(fileid, ((DB_LOCK_ILOCK *)(obj->objdata))->fileid, DB_FILE_ID_LEN);

	/* We need to find the BH pointer of the page. If the page is not
	 * 	in the memory or if it is not a data page, we don't have
	 * 	to record it as the in-memory unstamped page.
	 * In order to get the BH pointer, We get the DB of the page, then
	 * 	MPOOLFILE, then search in the mpool to fine it.  The following
	 *      bulk of code does this.
	 */
	for (dbp = LIST_FIRST(&dbenv->dblist);
	     dbp != NULL; dbp = LIST_NEXT(dbp, dblistlinks)) {
	    if (memcmp(dbp->fileid, fileid, DB_FILE_ID_LEN) == 0)
		break;
	}
	if (dbp == NULL)
	    STP_ERROR();

	/* Get the underlying data structures from DB */
	dbmfp = dbp->mpf;
	dbmp = dbmfp->dbmp;
	mp = dbmp->reginfo[0].primary;
	mfp = dbmfp->mfp;

	/* If the DB is not temporal we don't add this page to TL
	 * (Stanley Yao, 07/24/03) The dbp found above is not accurate,
         *   because there may be more than one DB pointing to the same
         *   file (with the same fileid) because they are all subdatabases
         *   in the same file.
         *   Since the BTN list (in-mem unstamped page list) of a transaction
         *   is just a hint, we can put it on the list first and remove it
         *   when bh_free() the page.
         *
	 if (!__stp_istemporal(dbenv, dbmfp))
	 continue;
	 *
	 */

	/* Stanley Yao (08/01/03)
	 * Since we add the __stp_reg_newpage() function, we can identify
	 * if a page belongs to a temporal database by finding it in BT.
	 * If we find it, it is temporal. Otherwise non-temporal
	 */
	/* 
	STP_BT_FIND(dbstp, pgno, fileid, bt);
	if (bt == NULL)
	    continue;*/
	
	mf_offset = R_OFFSET(dbmp->reginfo, mfp);
	bhp = NULL;
	R_LOCK(dbenv, dbmp->reginfo);
	/*
	 * Determine the hash bucket where this page will live, and get local
	 * pointers to the cache and its hash table.
	 */
#ifdef BDB_4
	n_cache = NCACHE(mp, mf_offset, pgno);
#endif
#ifdef BDB_3
	n_cache = NCACHE(mp, pgno);
#endif
	c_mp = dbmp->reginfo[n_cache].primary;
	n_bucket = NBUCKET(c_mp, mf_offset, pgno);
	dbht = R_ADDR(&dbmp->reginfo[n_cache], c_mp->htab);

	if (dbmfp->addr != NULL && F_ISSET(mfp, MP_CAN_MMAP))
	    STP_ERROR();

	record_it = 0;
	/* Search the hash chain for the page. */
	for (bhp = SH_TAILQ_FIRST(&dbht[n_bucket], __bh);
	     bhp != NULL; bhp = SH_TAILQ_NEXT(bhp, hq, __bh)) {

	    if (bhp->pgno != pgno || bhp->mf_offset != mf_offset)
		continue;

	    /* Any buffer we find might be trouble. */
	    for (first = 1; F_ISSET(bhp, BH_LOCKED)
#ifdef BDB_4
		     && !F_ISSET(dbenv, DB_ENV_NOLOCKING)
#endif
		     ; first = 0) {
		R_UNLOCK(dbenv, dbmp->reginfo);
		/*
		 * Explicitly yield the processor if it's not the first
		 * pass through this loop -- if we don't, we might end
		 * up running to the end of our CPU quantum as we will
		 * simply be swapping between the two locks.
		 */
		if (!first)
		    __os_yield(dbenv, 1);

		MUTEX_LOCK(dbenv, &bhp->mutex, dbenv->lockfhp);

		/* Wait for I/O to finish... */
		MUTEX_UNLOCK(dbenv, &bhp->mutex);
		R_LOCK(dbenv, dbmp->reginfo);
	    }

	    /*
	     * BH_TRASH --
	     * The contents of the buffer are garbage.  Shouldn't happen,
	     * and this read is likely to fail, but might as well try.
	     */
	    if (F_ISSET(bhp, BH_TRASH))
		break;

	    h = (PAGE *)bhp->buf;

	    /* If the page is not a data page, we don't have to stamp it */
	    /* TODO: add other data page types later! */
	    if ( (TYPE(h) == P_LBTREE) || (TYPE(h) == P_LRECNO) ||
		 (TYPE(h) == P_LDUP) ) {
		record_it = 1;
		break; /* 03/23/03 Stanley */
	    } else
		break;
	}

	R_UNLOCK(dbenv, dbmp->reginfo);

	/* If this page is not in mem, or it is not a data page that
	 *	actually need stamping, we don't have to record it in the
	 *	bh list of this txn.
	 */
	if (record_it != 1)
	    continue;

/* Added by Huilong on 5/23/07 for dead lock problem*/
        R_LOCK(dbenv, &dbstp->reginfo);;
/* End of modification */

#ifdef STAMP_AT_COMMIT
	/* Scan the page and stamp it so we don't have to store this
	 * page into the bh list of this transaction.
	 */
	STP_DEBUG( printf("Stamp pgno=%d at commit\n", PGNO(h)) );
	R_UNLOCK(dbenv, &dbstp->reginfo);

	STP_EXP( __stp_timestop() );
	__stp_mpfput(dbmfp, h);
	STP_EXP( __stp_timestart() );

	R_LOCK(dbenv, &dbstp->reginfo);
#else
	/* Our tBDB doesn't support AM other than BTree !!! */
	/* QAM use record based lock, so we have to get the pgno
	 *	first.
	 if (((DB_LOCK_ILOCK *)(obj->objdata))->type == DB_RECORD_LOCK) {
	 pgno = QAM_RECNO_PAGE(dbp, pgno);
	 }
	*/

	/* Search and put the (pgno, fileid) pair on the BT
	 *	if not already there.
	 */
	/* Find the (pgno+fileid) in the BT */
	STP_BT_FIND(dbstp, pgno, fileid, bt);

	/* By reaching here, the txn is committed.
	 * If the page is not in BT, add it to BT.
	 * If (pgno, fileid) is not in BT, add it.
	 */
/*The below code was changed by Huilong at 4:41pm 6/22/2006 to fix error */
	if (bt == NULL){
	    STP_BT_INSERT(dbstp, pgno, fileid, bt);
		STP_DEBUG( printf("  *** new BT inserted in stp_txn_end!! bt=%d\n", bt) );
	}
	/* Add a BTN in the 2-D list, if not already there */
        __btn_insert(dbstp, bt, tt, pgno, fileid);
/* end of change by Huilong */
#endif

/* Added by Huilong on 5/23/07 for dead lock problem*/
        R_UNLOCK(dbenv, &dbstp->reginfo);;
/* End of modification */

    }

/* Added by Huilong on 5/23/07 for dead lock problem*/
  R_LOCK(dbenv, &dbstp->reginfo);;
/* End of modification */

  out:


/* Huilong 7/14/06 */
#ifdef TBDB_AUDIT
    __aud_txn_end(dbenv, txnid);
#endif
/* end 7/14/06 */

    /* If the BTN of this txn is empty and also od_pgcnt of this txn is 0
     * we can reclaim this TT entry.
     */
   __tt_reclaim(dbstp, tt);

  out1:
    R_UNLOCK(dbenv, &dbstp->reginfo);
    UNLOCKREGION(dbenv, (DB_LOCKTAB *)dbenv->lk_handle);

    STP_DEBUG( print_tt(dbstp) );
    /*STP_DEBUG( Fprintf(stderr, "\tEXIT __stp_txn_end: dbenv=0x%x\n", dbenv); fflush(stdout) ); */
    /** modify here by yuhong -- 12/1/04*/
	STP_DEBUG( fprintf(stderr, "\tEXIT __stp_txn_end: dbenv=0x%x\n", dbenv); fflush(stdout) );
    STP_EXP( __stp_timestop() );

    return (0);

  err:
    R_UNLOCK(dbenv, &dbstp->reginfo);
    UNLOCKREGION(dbenv, (DB_LOCKTAB *)dbenv->lk_handle);
    exit(1);
}

/*
 * __stp_txn_commit --
 *	Add the commit time for a newly committed txn.
 */
int
__stp_txn_commit(dbenv, txnid_arg, timep)
    DB_ENV		*dbenv;
    u_int32_t	txnid_arg;
    TV		*timep;
{
    u_int32_t	tt_index, tt_prev, bt_head_index;
    STP_TT_ENTRY	*tt;
    STP_BT_ENTRY	*bt;
    STP_BTN	        *btnp;
    DB_STP		*dbstp;
    DB_MPOOL	*dbmp;
    int		ret;
    DB_LSN 		tmp_lsn;
    STP_ID		txnid;

    STP_STOP_WORKING();

    txnid.envid = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid;
    txnid.txnid = txnid_arg;

    dbmp = (DB_MPOOL *)dbenv->mp_handle;
    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

    STP_DEBUG( printf("\tENTER __stp_txn_commit: ID=(%ld,%ld), timep=(%d,%d)\n", \
		      txnid.envid, txnid.txnid, timep->tv_sec, timep->tv_usec); fflush(stdout) );

    /* MUTEX_LOCK(dbenv, &mp->sync_mutex, dbenv->lockfhp); */
/* Huilong: 2/1/2007  
 * Remove following code for concurrency control with notarizor/validator 
 */
//    R_LOCK(dbenv, &dbstp->reginfo);
//    R_LOCK(dbenv, dbmp->reginfo);
/* Huilong */


    /* Locate txnid in TT */
    STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);

    /* If txnid is not in TT, error */
    if (tt == NULL)
	STP_ERROR();

    /* !! think about if it is correct to put log here??? */
    /* if ( (tt[tt_index].od_pgcnt > 0) || (tt[tt_index].im_pgcnt > 0) ) */
    btnp = SH_TAILQ_FIRST(&(tt->btn), __stp_bt_node);
	if(tt->od_pgcnt > 0 || btnp != NULL)
    	_stp_txn_commit_log(dbenv, txnid, timep, tt->od_pgcnt, &tmp_lsn, 0);

    /* Fill in the commit time */
    tt->time.tv_sec = timep->tv_sec;
    tt->time.tv_usec = timep->tv_usec;


/* Huilong: 2/1/2007
 * Remove following code for concurrency control with notarizor/validator
 */
//    printf("!!!!!!!!!!!!!!!! Transaction commit time (%d, %d)\n", timep->tv_sec, timep->tv_usec);
//    R_UNLOCK(dbenv, dbmp->reginfo);
//    R_UNLOCK(dbenv, &dbstp->reginfo);
/* Huilong */

    /* MUTEX_UNLOCK(dbenv, &mp->sync_mutex); */

    STP_EXP( __stp_timestop() );

    return (0);

  err:	__db_err(dbenv, "__stp_txn_commit\n");
    printf("ERROR in __stp_txn_commit\n");
    exit(1);
    return (STP_ERROR_TXN_COMMIT);
}

/*
 * __stp_txn_abort --
 *	Add the commit time for a newly committed txn.
 *  PS: CLK abort routine should call me.  But seems CLK doesn't
 *      handle abort.  Need to ask them to do so!
 */
int
__stp_txn_abort(dbenv, txnid_arg)
    DB_ENV		*dbenv;
    u_int32_t	txnid_arg;
{
    STP *stp;
    u_int32_t	tt_index, tt_prev;
    STP_TT_ENTRY	*tt;
    STP_BTN  	*btnp;
    DB_STP		*dbstp;
    DB_MPOOL	*dbmp;
    int		ret;
    STP_ID		txnid;

    STP_STOP_WORKING();

    txnid.envid = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->envid;
    txnid.txnid = txnid_arg;

    STP_DEBUG( printf("\tEnter _stp_txn_abort:dbenv=0x%x, txnid=(%d,%d)\n", \
	   dbenv, txnid.envid, txnid.txnid) );

    dbstp = ((TEMPORAL_HANDLE *)(dbenv->temporal_handle))->stp_handle;

    /* If the environment is NON-TEMPORAL, the stamper do nothing. */
    if (dbstp == NULL)
	return 0;

    stp = dbstp->reginfo.primary;

    STP_EXP( __stp_timestart() );
    STP_EXP( addcall(0) );

    dbmp = (DB_MPOOL *)dbenv->mp_handle;

    R_LOCK(dbenv, &dbstp->reginfo);
    R_LOCK(dbenv, dbmp->reginfo);

    /* Locate txnid in TT */
    STP_TT_FIND(dbstp, dbstp->tt_htab, txnid, tt);

    /* If txnid is not in TT, error */
    if (tt == NULL)
	STP_ERROR();

    /* If txnid's BTN list is not empty, there must be something wrong */
    btnp = SH_TAILQ_FIRST(&(tt->btn), __stp_bt_node);
    if (btnp != NULL)
	STP_ERROR();

    /* When txn_abort() is called, the transaction must haven't committed yet. That
     *  means the BTN list of the TT is still empty. Reclaim the TT entry.
     */
    tt_index = STP_TT_HASH(dbstp, txnid);
    SH_TAILQ_REMOVE(&(dbstp->tt_htab[tt_index]), tt, links, __stp_tt_entry);

	/* added by yuhong */
	SH_TAILQ_INSERT_HEAD(&(stp->free_tt_ent), tt, links, __stp_tt_entry);

    STP_DEBUG( printf("\tABORT: delete TT[%d]. The updated TT is:\n", tt_prev) );
    STP_DEBUG( print_tt(dbstp) );

    R_UNLOCK(dbenv, dbmp->reginfo);
    R_UNLOCK(dbenv, &dbstp->reginfo);

    STP_EXP( __stp_timestop() );

    return (0);

  err:	__db_err(dbenv, "__stp_txn_abort\n");
    exit(1);
    return (STP_ERROR_TXN_ABORT);
}
